#include "main.h"
#include "mstdlib.h"
#include "buf.h"

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define M_BUF_INITIAL_SIZE 1024

struct M_buf {
	char       *data;
	size_t      data_size;
	size_t      data_length;
};

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static size_t ensure_space(Ml_buf_t buf, size_t length);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

Ml_buf_t Ml_buf_create(void)
{
	Ml_buf_t buf;

	buf = Ml_malloc(sizeof(*buf));

	buf->data        = NULL;
	buf->data_size   = 0;
	buf->data_length = 0;

	return buf;
}

size_t Ml_buf_len(Ml_buf_t buf)
{
	if (buf == NULL)
		return 0;

	return buf->data_length;
}

const char *Ml_buf_peek(Ml_buf_t buf)
{
	if (buf == NULL)
		return NULL;

	return buf->data;
}

void Ml_buf_truncate(Ml_buf_t buf, size_t length)
{
	if (buf == NULL || buf->data_length <= length || buf->data == NULL)
		return;
	buf->data_length            = length;
	buf->data[buf->data_length] = 0;
}

void Ml_buf_drop(Ml_buf_t buf, size_t num)
{
	if (buf == NULL || num == 0)
		return;

	if (num < buf->data_length) {
		memmove(buf->data, buf->data + num, buf->data_length - num);
		buf->data_length           -= num;
	} else if (num >= buf->data_length) {
		buf->data_length            = 0;
	}

	/* Null-terminate for safety */
	buf->data[buf->data_length] = 0;
}

void Ml_buf_cancel(Ml_buf_t buf)
{
	if (buf == NULL)
		return;

	Ml_free(buf->data);
	Ml_free(buf);
}


void Ml_buf_finish(Ml_buf_t buf, char **out, size_t *out_length)
{
	if (buf == NULL)
		return;

	*out        = buf->data;
	*out_length = buf->data_length;

	Ml_free(buf);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void Ml_buf_merge(Ml_buf_t dest, Ml_buf_t source)
{
	if (dest == NULL || source == NULL)
		return;

	Ml_buf_add_bytes(dest, source->data, source->data_length);
	Ml_buf_cancel(source);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void Ml_buf_add_byte(Ml_buf_t buf, char byte)
{
	size_t new_data_length;

	if (buf == NULL)
		return;

	new_data_length = ensure_space(buf, 1);
	buf->data[buf->data_length] = byte;
	buf->data_length = new_data_length;
}

void Ml_buf_add_bytes(Ml_buf_t buf, const void *bytes, size_t bytes_length)
{
	size_t new_data_length;

	if (buf == NULL || bytes == NULL || bytes_length == 0) {
		return;
	}

	new_data_length = ensure_space(buf, bytes_length);
	memcpy(buf->data + buf->data_length, bytes, bytes_length);
	buf->data_length = new_data_length;
}

void Ml_buf_add_str(Ml_buf_t buf, const char *str)
{
	size_t str_length;

	if (str == NULL) {
		return;
	}

	str_length = strlen(str);
	if (str_length == 0) {
		return;
	}

	Ml_buf_add_bytes(buf, str, str_length);
}

void Ml_buf_add_uint(Ml_buf_t buf, M_uint64 n)
{
	char space[32];

	Ml_snprintf(space, sizeof(space), "%llu", n);
	Ml_buf_add_str(buf, space);
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static size_t next_multiple_of_block_size(size_t needed_size, size_t curr_size)
{
	size_t n;

	n = curr_size;
	if (n == 0)
		n = M_BUF_INITIAL_SIZE;

	while (needed_size > n) {
		n <<= 1;

		/* Detect wrap */
		if (n < curr_size)
			return 0;
	}

	return n;
}

static size_t ensure_space(Ml_buf_t buf, size_t length)
{
	size_t new_data_length;
	size_t new_data_size;
	char *new_data;

	new_data_length = buf->data_length + length;

	if (new_data_length + 1/* NULL term */ > buf->data_size) {
		new_data_size  = next_multiple_of_block_size(new_data_length + 1/* NULL term */, buf->data_size);
		new_data       = Ml_malloc(new_data_size);
		memcpy(new_data, buf->data, buf->data_length);
		Ml_free(buf->data);
		buf->data      = new_data;
		buf->data_size = new_data_size;
	}

	/* NULL term */
	buf->data[new_data_length] = 0;

	return new_data_length;
}



