#ifndef __BUF_H__
#define __BUF_H__

/* make Ml_buf_t opaque */
struct M_buf;
typedef struct M_buf * Ml_buf_t;

/*! Create a new buffer.
 * \return allocated buffer which should be freed with Ml_buf_cancel() or
 *         Ml_buf_finish().
 */
Ml_buf_t Ml_buf_create(void);

/*! Free a buffer, discarding its data.
 * \param[in] buf buffer
 */
void Ml_buf_cancel(Ml_buf_t buf);

/*! Free a buffer, saving its data. The client is responsible for freeing the
 * data.
 * \param[in] buf        buffer
 * \param[in] out        data
 * \param[in] out_length data length
 */
void Ml_buf_finish(Ml_buf_t buf, char **out, size_t *out_length);

/*! Return the length of the data held by a buffer.
 * \param[in] buf buffer
 * \return data length
 */
size_t Ml_buf_len(Ml_buf_t buf);

/*! Take a sneak peek at the buffer (returns const pointer)
 * \param[in] buf buffer
 * \return current beginning data pointer
 */
const char *Ml_buf_peek(Ml_buf_t buf);

/*! Truncate the length of the data to the specified size
 * \param[in] buf buffer
 * \param[in] length Length to truncate buffer to
 */
void Ml_buf_truncate(Ml_buf_t buf, size_t length);

/*! Drop the specified number of bytes from the beginning of the buf
 * \param[in] buf buffer
 * \param[in] num Number of bytes to drop
 */
void Ml_buf_drop(Ml_buf_t buf, size_t num);

/*! Merge two buffers. The data in the source buffer is appended to the
 * destination buffer. The source buffer is freed.
 * \param[in] dest   buffer
 * \param[in] source buffer
 */
void Ml_buf_merge(Ml_buf_t dest, Ml_buf_t source);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*! Append one byte to a buffer.
 * \param[in] buf  buffer
 * \param[in] byte byte to append
 */
void Ml_buf_add_byte(Ml_buf_t buf, char byte);

/*! Append zero or more bytes to a buffer.
 * \param[in] buf          buffer
 * \param[in] bytes        bytes to append
 * \param[in] bytes_length number of bytes to append
 */
void Ml_buf_add_bytes(Ml_buf_t buf, const void *bytes, size_t bytes_length);

/*! Append a C string (zero or more bytes terminated with a NUL) to a buffer.
 * The NUL is not appended.
 * \param[in] buf buffer
 * \param[in] str string to append
 */
void Ml_buf_add_str(Ml_buf_t buf, const char *str);

/*! Append the character decimal representation ("%llu") of an unsigned integer
 * to a buffer.
 * \param[in] buf buffer
 * \param[in] n   unsigned integer to append
 */
void Ml_buf_add_uint(Ml_buf_t buf, M_uint64 n);


#endif

