#include "main.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void Ml_csv_parse_count(const char *str, size_t len, char delim, char quote, int *num_rows_out, int *num_cols_out)
{
	size_t i, row_len = 0;
	int on_quote = 0;

	/* Count rows, and for first row, count columns */
	(*num_cols_out) = 0;
	(*num_rows_out) = 0;
	for (i=0; i<len; i++) {
		if (str[i] == quote) {
			row_len++;
			if (str[i+1] == quote) {
				/* Double Quote, skip next char */
				i++;
			} else if (on_quote) {
				on_quote = 0;
			} else {
				on_quote = 1;
			}
		} else if (!on_quote && str[i] == delim && (*num_rows_out) == 0) {
			row_len++;
			/* If still on first row, increment column count */
			(*num_cols_out)++;
		} else if (!on_quote && str[i] == '\n') {
			if ((*num_rows_out) == 0)
				(*num_cols_out)++;
			row_len = 0;
			(*num_rows_out)++;
		} else if (on_quote || str[i] != '\r') {
			row_len++;
		}
	}

	if ((*num_rows_out) == 0 && row_len)
		(*num_cols_out)++;

	if (row_len)
		(*num_rows_out)++;
}

void Ml_csv_free(Ml_csv_data *data)
{
	if (data == NULL)
		return;
	Ml_free(data->data_arr);
	Ml_free(data->data_ptr);
	Ml_free(data);
}

static void Ml_csv_remove_quotes(char *str, char quote)
{
	size_t len, i, cnt = 0;
	
	if (str == NULL)
		return;
	
	len = strlen(str);

	for (i=0; i<len; i++) {
		if (str[i] == quote) {
			if (str[i+1] == quote) {
				str[cnt++] = quote;
				i+=1;
			}
		} else {
			str[cnt++] = str[i];
		}
	}
	str[cnt] = 0;

	/* If this was an empty string, "", it may leave
	 * a single quote behind because it looks like a
	 * quoted quote.  Lets detect that and remove it */
	if (cnt == 1 && str[cnt-1] == quote)
		str[cnt-1] = 0;
}


Ml_csv_data *Ml_csv_parse_inplace(char *data, size_t len, char delim, char quote)
{
	int num_rows = 0, num_cols = 0;
	size_t i, outlen;
	int on_quote = 0, had_quote = 0;
	char ***out = NULL ,*buf = NULL;
	int row = 0, col = 0;
	Ml_csv_data *csv = NULL;

	Ml_csv_parse_count(data, len, delim, quote, &num_rows, &num_cols);

	/* Allocate as one large matrix of pointers */
	outlen = (num_rows * sizeof(*out)) + (num_rows * (num_cols * sizeof(**out)));
	buf = Ml_malloc(outlen);
	memset(buf, 0, outlen);
	out = (char ***)buf;
	for (i=0; i<(size_t)num_rows; i++) {
		/* Calc offset of allocated block to assign pointers */
		out[i] = (char **)(buf + ((num_rows * sizeof(*out)) + (i * (num_cols * sizeof(**out)))));
	}

	/* Set pointers to positions in data for column starts */
	row       = 0;
	col       = 0;
	on_quote  = 0;
	had_quote = 0;
	out[row][col] = data;
	for (i=0; i<len; i++) {
		if (data[i] == quote) {
			had_quote = 1;
			if (on_quote && data[i+1] == quote) {
				/* Double Quote, skip next char */
				i++;
			} else if (on_quote) {
				on_quote = 0;
			} else {
				on_quote = 1;
			}
		} else if (!on_quote && data[i] == delim) {
			data[i] = 0;

			/* If we know the previous column had quotes, lets
			 * sanitize that data now */
			if (had_quote && col < num_cols) {
				Ml_csv_remove_quotes(out[row][col], quote);
			}
			had_quote = 0;

			col++;

			/* Ensure we don't overflow by one row containing too
			 * many columns.  We will effectively truncate every
			 * row that contains more columns than our first row */
			if (col < num_cols) {
				out[row][col] = data+(i+1);
			}
		} else if (!on_quote && data[i] == '\n') {
			data[i] = 0;

			/* If we know the previous column had quotes, lets
			 * sanitize that data now */
			if (had_quote && col < num_cols) {
				Ml_csv_remove_quotes(out[row][col], quote);
			}

			row++;
			col = 0;

			if (row == num_rows)
				break;

			out[row][col] = data+(i+1);

			had_quote = 0;
		} else if (!on_quote && data[i] == '\r') {
			data[i] = 0;
		}
	}

	/* If we know the previous column had quotes, lets
	 * sanitize that data now */
	if (row < num_rows && col < num_cols && had_quote) {
		
		Ml_csv_remove_quotes(out[row][col], quote);
	}
	csv = Ml_malloc(sizeof(*csv));
	csv->data_arr = out;
	csv->data_ptr = data;
	csv->num_rows = num_rows;
	csv->num_cols = num_cols;

	return csv;
}


Ml_csv_data *Ml_csv_parse(const char *str, size_t len, char delim, char quote)
{
	char *data;

	if (str == NULL || len == 0)
		return NULL;

	/* Duplicate data for modification */
	data = Ml_malloc(len+1);
	memcpy(data, str, len);
	data[len] = 0;
	
	return Ml_csv_parse_inplace(data, len, delim, quote);
}


const char *Ml_csv_get_cell(Ml_csv_data *csv, int row, const char *colname)
{
	int i;
	const char *headername;

	if (csv == NULL)
		return NULL;

	if (row >= Ml_csv_get_numrows(csv))
		return NULL;

	for (i=0; i<Ml_csv_get_numcols(csv); i++) {
		headername = Ml_csv_get_header(csv, i);
		if (strcasecmp(colname, headername) == 0)
			return Ml_csv_get_cellbynum(csv, row, i);
	}
	return NULL;
}
