#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "http_p.h"

#include "http.h"

#include "snprintf.h"
#include "sockets.h"
#include "mstring.h"
#include "mstdlib.h"

#if defined(_WIN32) && !defined(_WIN32_WCE)
#    define strcasecmp(a,b) _stricmp(a,b)
#    define strncasecmp(a,b,c) _strnicmp(a,b,c)
#endif

static int LIBMONETRA_http_del_header(M_http_t http, const char *header);
static int LIBMONETRA_http_append_last_header(M_http_t http, const char *value);

static int LIBMONETRA_http_del_header(M_http_t http, const char *header)
{
	int i, found;

	if ((http == NULL) || (header == NULL) || (strlen(header) == 0)) {
		return 0;
	}

	found = -1;
	for (i = 0; found == -1 && i < http->num_headers; i++) {
		if (strcasecmp(http->headers[i]->name, header) == 0) {
			found = i;
		}
	}

	if (found == -1) {
		return 0;
	}

	Ml_free(http->headers[found]->name);
	Ml_free(http->headers[found]->value);
	Ml_free(http->headers[found]);

	for (i = found; i < http->num_headers - 1; i++) {
		http->headers[i] = http->headers[i + 1];
	}

	http->num_headers--;

	http->headers[http->num_headers] = NULL;

	return 1;
}

void LIBMONETRA_http_free(M_http_t http)
{
	int i;

	if (http == NULL) {
		return;
	}

	Ml_free(http->path);
	Ml_free(http->host);

	for (i = 0; i < http->num_headers; i++) {
		Ml_free(http->headers[i]->name);
		Ml_free(http->headers[i]->value);
		Ml_free(http->headers[i]);
	}

	Ml_free(http->code_msg);
	Ml_free(http->headers);
	Ml_free(http->content);
	Ml_free(http);
}

#define M_HTTP_ALLOC 1024
#define M_HTTP_APPEND(dest, destlen, destalloc, src, srclen) \
        do { \
		while (destlen + srclen + 1 >= destalloc) { \
			dest = Ml_realloc(dest, destalloc + M_HTTP_ALLOC); \
			memset(dest + destalloc, 0, M_HTTP_ALLOC); \
			destalloc += M_HTTP_ALLOC; \
		} \
		if (srclen) {memcpy(dest + destlen, src, srclen); } \
		destlen += srclen; \
	} while (0)

char *LIBMONETRA_http_gen_request(M_http_t http)
{
	char *buf, *var;
	int i, len, alloclen, templen;
	char temp[255];

	if (http == NULL) {
		return NULL;
	}

	if (http->path == NULL) {
		return NULL;
	}

	if ((http->version == MHTTP_HTTP_VER_1_1) && (http->host == NULL)) {
		return NULL;
	}

	buf = NULL;
	alloclen = len = 0;

	switch (http->hosttype) {
		case MHTTP_POST:
			M_HTTP_APPEND(buf, len, alloclen, "POST ", 5);
			break;
		case MHTTP_GET:
			M_HTTP_APPEND(buf, len, alloclen, "GET ", 4);
			break;
		case MHTTP_CONNECT:
			M_HTTP_APPEND(buf, len, alloclen, "CONNECT ", 8);
			break;
	}

	templen = 0;
	if (http->path != NULL) {
		templen = strlen(http->path);
	}

	if (http->hosttype == MHTTP_CONNECT) {
		if (http->port) {
			templen = Ml_snprintf(temp, sizeof(temp), "%s:%d", http->host, (int)http->port);
		} else {
			templen = Ml_snprintf(temp, sizeof(temp), "%s", http->host);
		}

		M_HTTP_APPEND(buf, len, alloclen, temp, templen);
	} else if (templen) {
		M_HTTP_APPEND(buf, len, alloclen, http->path, templen);
	}

	if (http->version == MHTTP_HTTP_VER_1_1) {
		M_HTTP_APPEND(buf, len, alloclen, " HTTP/1.1\r\n", 11);

		var = LIBMONETRA_http_get_header_str(http, "Connection", 0);
		if (var == NULL) {
			LIBMONETRA_http_set_header(http, "Connection", "Keep-Alive", 1);
		} else {
			Ml_free(var);
		}
	} else {
		M_HTTP_APPEND(buf, len, alloclen, " HTTP/1.0\r\n", 11);
	}

	if ((http->version == MHTTP_HTTP_VER_1_1) || (http->host != NULL)) {
		M_HTTP_APPEND(buf, len, alloclen, "Host: ", 6);
		if (http->port) {
			templen = Ml_snprintf(temp, sizeof(temp), "%s:%d", http->host, (int)http->port);
		} else {
			templen = Ml_snprintf(temp, sizeof(temp), "%s", http->host);
		}

		M_HTTP_APPEND(buf, len, alloclen, temp, templen);
	}

	for (i = 0; i < http->num_headers; i++) {
		M_HTTP_APPEND(buf, len, alloclen, "\r\n", 2);
		templen = strlen(http->headers[i]->name);
		M_HTTP_APPEND(buf, len, alloclen, http->headers[i]->name, templen);
		M_HTTP_APPEND(buf, len, alloclen, ": ", 2);
		templen = strlen(http->headers[i]->value);
		M_HTTP_APPEND(buf, len, alloclen, http->headers[i]->value, templen);
	}

	M_HTTP_APPEND(buf, len, alloclen, "\r\n\r\n", 4);
	http->header_len = len;

	M_HTTP_APPEND(buf, len, alloclen, http->content, http->content_len);

	return buf;
}

char *LIBMONETRA_http_gen_resp(M_http_t http)
{
	char *buf;
	const char *ver;
	int i, len, alloclen, templen;
	char temp[255];
	int has_contenttype = 0;

	if (http == NULL) {
		return NULL;
	}

	if (http->code_msg == NULL) {
		return NULL;
	}

	if (http->version == MHTTP_HTTP_VER_1_1) {
		ver = "1.1";
	} else {
		ver = "1.0";
	}

	Ml_snprintf(
		temp,
		sizeof(temp),
		"HTTP/%s %i %s\r\n" "Server: %s v%d.%d.%d\r\n",
		ver,
		http->code,
		http->code_msg,
		"MHTTP",
		MHTTP_MAJOR,
		MHTTP_MINOR,
		MHTTP_RELEASE
	);

	buf = NULL;
	alloclen = len = 0;

	templen = strlen(temp);
	M_HTTP_APPEND(buf, len, alloclen, temp, templen);

	for (i = 0; i < http->num_headers; i++) {
		if (strcasecmp(http->headers[i]->name, "Content-Type") == 0) {
			has_contenttype = 1;
		}

		templen = strlen(http->headers[i]->name);
		M_HTTP_APPEND(buf, len, alloclen, http->headers[i]->name, templen);
		M_HTTP_APPEND(buf, len, alloclen, ": ", 2);
		templen = strlen(http->headers[i]->value);
		M_HTTP_APPEND(buf, len, alloclen, http->headers[i]->value, templen);
		M_HTTP_APPEND(buf, len, alloclen, "\r\n", 2);
	}

	if (!has_contenttype) {
		/* Always make sure there is a content type */
		M_HTTP_APPEND(buf, len, alloclen, "Content-Type: text/plain\r\n", 26);
	}

	M_HTTP_APPEND(buf, len, alloclen, "\r\n", 2);

	http->header_len = len;

	M_HTTP_APPEND(buf, len, alloclen, http->content, http->content_len);

	return buf;
}

int LIBMONETRA_http_get_code(M_http_t http, char **code_msg)
{
	if ((http == NULL) || (http->code == 0)) {
		return 0;
	}

	*code_msg = Ml_strdup(M_Safe(http->code_msg));

	return http->code;
}

int LIBMONETRA_http_get_content(M_http_t http, char **content)
{
	if ((http == NULL) || (http->content == NULL)) {
		return 0;
	}

	*content = Ml_malloc(http->content_len + 1);
	memcpy(*content, http->content, http->content_len);
	(*content)[http->content_len] = '\0';

	return http->content_len;
}

int LIBMONETRA_http_get_content_len(M_http_t http)
{
	if (http == NULL) {
		return 0;
	}

	return http->content_len;
}

int LIBMONETRA_http_get_gettype(M_http_t http)
{
	if (http == NULL) {
		return -1;
	}

	return http->hosttype;
}

int LIBMONETRA_http_get_request_len(M_http_t http)
{
	if (http == NULL) {
		return 0;
	}

	return http->header_len + http->content_len;
}

int LIBMONETRA_http_get_resp_len(M_http_t http)
{
	return LIBMONETRA_http_get_request_len(http);
}

int LIBMONETRA_http_get_header_len(M_http_t http)
{
	if (http == NULL) {
		return 0;
	}

	return http->header_len;
}

int LIBMONETRA_http_get_header_int(M_http_t http, const char *header, int myindex)
{
	char *buf;
	int ret;

	if ((http == NULL) || (header == NULL) || (strlen(header) == 0)) {
		return 0;
	}

	ret = 0;

	buf = LIBMONETRA_http_get_header_str(http, header, myindex);
	if (buf != NULL) {
		ret = atoi(buf);
		Ml_free(buf);
	}

	return ret;
}

char *LIBMONETRA_http_get_header_str(M_http_t http, const char *header, int myindex)
{
	char *ret;
	int count, i;

	if ((http == NULL) || (header == NULL) || (strlen(header) == 0)) {
		return NULL;
	}

	ret = NULL;

	count = 0;

	for (i = 0; ret == NULL && i < http->num_headers; i++) {
		if (strcasecmp(http->headers[i]->name, header) == 0) {
			if (count == myindex) {
				ret = Ml_strdup(http->headers[i]->value);
			} else {
				count++;
			}
		}
	}

	return ret;
}

int LIBMONETRA_http_get_version(M_http_t http)
{
	if (http == NULL) {
		return -1;
	}

	return http->version;
}

int LIBMONETRA_http_get_complete(M_http_t http)
{
	if (http == NULL) {
		return 0;
	}

	return http->complete;
}

M_http_t LIBMONETRA_http_new()
{
	M_http_t http;

	http = Ml_malloc(sizeof(struct M_http_s));
	memset(http, 0, sizeof(struct M_http_s));

	http->version = MHTTP_HTTP_VER_1_0;
	http->hosttype = MHTTP_POST;
	http->path = NULL;
	http->host = NULL;
	http->headers = NULL;
	http->num_headers = 0;
	http->content = NULL;
	http->content_len = 0;
	http->header_len = 0;
	http->code = 0;
	http->code_msg = NULL;
	http->complete = 0;

	return http;
}

/* Returns position of end of parse on success,
 * -1 if not enough data to parse next section,
 * 0 if complete.  The input data may be modified
 * for efficiency */
int LIBMONETRA_http_process_chunksection(char *data, int len, char **out_unchunked, int *out_len)
{
	int myoffset = 0;
	char *dptr = NULL, *ptr = NULL;
	int mylen = 0;
	long chunk_part_len = 0;
	int i;
	int found_cr = 0, found_lf = 0;
	int cr_pos = -1, lf_pos = -1;

	/* No data yet */
	if (len <= 0) {
		return -1;
	}

	dptr = data;
	mylen = len;

	myoffset = 0;
	/* Strip any leading whitespace */
	while ((*dptr == ' ' || *dptr == '\r' || *dptr == '\n') && myoffset < mylen) {
		dptr++;
		myoffset++;
	}

	if (myoffset >= len) {
		return -1;
	}

	/* Find CRLF */
	found_cr = found_lf = 0;
	for (i = 0; i < mylen - myoffset; i++) {
		if (dptr[i] == '\r') {
			found_cr = 1;
			dptr[i] = 0;
			cr_pos = i;
		} else if ((dptr[i] == '\n') && found_cr) {
			dptr[i] = 0;
			found_lf = 1;
			lf_pos = i;
			break;
		}
	}

	if (!found_lf) {
		if (cr_pos > -1) {
			dptr[cr_pos] = '\r';
		}

		return -1;
	}

	myoffset += i + 1;
	/* See if there is a semi-colon, if so, null terminate */
	ptr = strchr(dptr, ';');
	if (ptr != NULL) {
		*ptr = 0;
	}

	/* fetch length, it's in hex */
	chunk_part_len = strtol(dptr, NULL, 16);

	/* Last chunk is always 0 bytes */
	if (chunk_part_len == 0) {
		return 0;
	}

	/* Not enough bytes read yet */
	if (chunk_part_len > mylen - myoffset) {
		/* We rely on CRLF for parsing, make sure we put them back */
		if (cr_pos > -1) {
			dptr[cr_pos] = '\r';
		}

		if (lf_pos > -1) {
			dptr[lf_pos] = '\n';
		}

		return -1;
	}

	dptr += myoffset;
	*out_unchunked = (char *)Ml_realloc(*out_unchunked, (*out_len) + chunk_part_len + 1);
	memcpy((*out_unchunked) + (*out_len), dptr, chunk_part_len);
	(*out_len) += chunk_part_len;
	(*out_unchunked)[*out_len] = 0;

	dptr += chunk_part_len;

	return (int)(dptr - data);
}

/* Returns length of processed data on success (0 is valid), and
 * -1 on incomplete data.  The passed data may be modified for
 * efficiency */
static int LIBMONETRA_http_process_chunkeddata(char *data, int len, char **out_unchunked)
{
	char *dptr = NULL;
	int mylen = 0;
	int ret = 0;
	int outlen = 0;

	/* No data yet */
	if (len <= 0) {
		return -1;
	}

	*out_unchunked = NULL;

	dptr = data;
	mylen = len;
	while (1) {
		ret = LIBMONETRA_http_process_chunksection(dptr, mylen, out_unchunked, &outlen);
		if (ret == 0) {
			break;
		} else if (ret == -1) {
			Ml_free(*out_unchunked);
			*out_unchunked = NULL;
			return -1;
		} else {
			dptr += ret;
			mylen -= ret;
		}
	}

	return outlen;
}

static char *LIBMONETRA_http_strcasestr(char *haystack, const char *needle)
{
	int h_len;
	int n_len;
	int i;

	if ((haystack == NULL) || (needle == NULL)) {
		return NULL;
	}

	h_len = strlen(haystack);
	n_len = strlen(needle);

	if (!h_len || !n_len) {
		return NULL;
	}

	for (i = 0; i < h_len; i++) {
		if (strncasecmp(haystack + i, needle, n_len) == 0) {
			return haystack + i;
		}
	}

	return NULL;
}

static M_http_t LIBMONETRA_http_parse_internal(const char *buf, int *len, int return_on_incomplete, char *debugmsg,
                                               int msglen)
{
	char code[4], *content, *content_pos, *header, *path, *pos, **rows, *str;
	const char *bufpos;
	char *encoding = NULL;
	int content_len, i, num_rows;
	int num_chars = 0, complete_header = 0;
	int mylen = 0, myoffset = 0;
	M_http_t http;
	char *unchunked = NULL;
	int unchunked_len = 0;

	memset(debugmsg, 0, msglen);

	if ((buf == NULL) || (strlen(buf) == 0) || ((*len) < 1)) {
		return NULL;
	}

	/* skip over any leading spaces */
	bufpos = buf;
	while ((*bufpos == ' ' || *bufpos == '\r' || *bufpos == '\n') && myoffset < (*len)) {
		bufpos++;
		myoffset++;
	}

	if (myoffset >= (*len)) {
		return NULL;
	}

	/* headers are text, content may not be */
	mylen = (*len) - myoffset;
	str = Ml_malloc(mylen + 1);
	memcpy(str, bufpos, mylen);
	str[mylen] = '\0';

	/* split into rows */
	rows = Ml_malloc(sizeof(char *));
	rows[0] = str;
	num_rows = 1;

	num_chars = 0;
	complete_header = 0;
	for (i = 0; i < mylen; i++) {
		switch (str[i]) {
			case '\n':
				str[i] = '\0';

				if (num_chars == 0) {
					content_pos = str + (i + 1);
					complete_header = 1;
					break;
				}

				rows = Ml_realloc(rows, sizeof(char *) * (num_rows + 1));
				rows[num_rows++] = str + (i + 1);

				num_chars = 0;
				break;
			case '\r':
				str[i] = '\0';
				break;
			default:
				num_chars++;
				break;
		}

		if (complete_header) {
			break;
		}
	}

	if (!complete_header) {
		Ml_free(str);
		Ml_free(rows);
		return NULL;
	}

	http = LIBMONETRA_http_new();
	http->header_len = (int)(content_pos - str);
	if (LIBMONETRA_http_strcasestr(rows[0], "HTTP/1.1") != NULL) {
		http->version = MHTTP_HTTP_VER_1_1;
	} else if (LIBMONETRA_http_strcasestr(rows[0], "HTTP/1.0") != NULL) {
		http->version = MHTTP_HTTP_VER_1_0;
	} else {
		LIBMONETRA_http_free(http);
		Ml_free(str);
		Ml_free(rows);
		return NULL;
	}

	if ((strncasecmp(rows[0], "GET", 3) == 0) || (strncasecmp(rows[0], "POST", 4) == 0)) {
		if (rows[0][0] == 'G') {
			pos = rows[0] + 3;
		} else {
			pos = rows[0] + 4;
		}

		path = Ml_strdup(pos);
		pos = LIBMONETRA_http_strcasestr(path, "HTTP");
		if (pos == NULL) {
			LIBMONETRA_http_free(http);
			Ml_free(str);
			Ml_free(rows);
			return NULL;
		}

		pos--;
		*pos = '\0';

		http->path = path;
	} else {
		http->code = 0;
		if (strlen(rows[0]) >= 12) {
			memcpy(code, rows[0] + 9, 3);
			code[3] = '\0';
		}

		http->code = atoi(code);

		if (http->code == 0) {
			LIBMONETRA_http_free(http);
			Ml_free(str);
			Ml_free(rows);
			return NULL;
		}

		if (strlen(rows[0]) > 13) {
			http->code_msg = Ml_strdup(rows[0] + 13);
		}
	}

	/* Parse each header row */
	for (i = 1; i < num_rows; i++) {
		pos = strchr(rows[i], ':');
		if (pos == NULL) {
			/* HTTP/1.1 section 2.2 states that if a row begins with
			 * whitespace, it is a continuation of the previous row */
			if ((rows[i][0] == '\t') || (rows[i][0] == ' ')) {
				pos = rows[i];
				while ((*pos == ' ' || *pos == '\t') && *pos != 0) {pos++; }

				LIBMONETRA_http_append_last_header(http, pos);
			}

			continue;
		}

		header = Ml_malloc(pos - rows[i] + 1);
		memcpy(header, rows[i], pos - rows[i]);
		header[pos - rows[i]] = '\0';

		pos++;

		while (*pos == ' ' && *pos != 0) {
			pos++;
		}

		LIBMONETRA_http_set_header(http, header, pos, 0);

		Ml_free(header);
	}

	content = NULL;
	content_len = 0;

	if (http->code == 100) {
		/* 100 results have no content (any other codes we should ignore content for?) */
		http->content = content;
		http->content_len = content_len;
		/* Set len so we can chop off this part of the content */
		(*len) = myoffset + (content_pos - str);
	} else {
		/* Calculate what we think the content length is */
		if (content_pos - str <= mylen - 1) {
			content_len = mylen - (content_pos - str);
		}

		/* Check the calced len against a specified length */
		i = LIBMONETRA_http_get_header_int(http, "Content-Length", 0);
		if ((i != 0) && (content_len < i)) {
			/* if the specified length is greater than the actual length, error out */
			Ml_snprintf(debugmsg, msglen, "need more bytes, expecting %i have %i",
				i, content_len);
			Ml_free(str);
			Ml_free(rows);
			if (!return_on_incomplete) {
				LIBMONETRA_http_free(http);
				http = NULL;
			}

			return http;
		} else if ((i != 0) && (content_len > i)) {
			/* if the specified length is smaller than the actual length, truncate */
			Ml_snprintf(debugmsg, msglen, "more bytes in buffer, expecting %i have %i",
				i, content_len);
			content = Ml_malloc(content_len + 1);
			memcpy(content, content_pos, content_len);
			content[content_len] = '\0';
			content[i] = '\0';
			content_len = i;
			/* Set len so that we can at least partially support chunked encoding */
			(*len) = myoffset + (content_pos - str) + i;
		} else {
			/* Lengths match, copy it */
			content = Ml_malloc(content_len + 1);
			memcpy(content, content_pos, content_len);
			content[content_len] = '\0';
		}
	}

	Ml_free(str);
	Ml_free(rows);

	encoding = LIBMONETRA_http_get_header_str(http, "Transfer-Encoding", 0);
	if ((encoding != NULL) && (strcasecmp(encoding, "chunked") == 0)) {
		Ml_free(encoding);
		unchunked_len = LIBMONETRA_http_process_chunkeddata(content, content_len, &unchunked);
		Ml_free(content);
		if (unchunked_len == -1) {
			if (!return_on_incomplete) {
				LIBMONETRA_http_free(http);
				http = NULL;
			}

			return http;
		}

		http->content = unchunked;
		http->content_len = unchunked_len;
	} else {
		Ml_free(encoding);
		http->content = content;
		http->content_len = content_len;
	}

	http->complete = 1;

	return http;
}

M_http_t LIBMONETRA_http_parse(const char *buf, int len, int return_on_incomplete, char *debugmsg, int msglen)
{
	int mylen = 0;
	int offset = 0;
	M_http_t http = NULL;

	while (1) {
		mylen = len - offset;
		http = LIBMONETRA_http_parse_internal(buf + offset, &mylen, return_on_incomplete, debugmsg, msglen);
		if (http == NULL) {
			break;
		} else if (http->code == 100) {
			LIBMONETRA_http_free(http); http = NULL;
			offset += mylen;
		} else {
			break;
		}
	}

	return http;
}

int LIBMONETRA_http_set_content(M_http_t http, const char *content, int len)
{
	char lenstr[10];

	if ((http == NULL) || (content == NULL)) {
		return 0;
	}

	Ml_free(http->content);
	http->content = Ml_malloc(len + 1);
	memcpy(http->content, content, len);
	http->content[len] = '\0';

	http->content_len = len;

	Ml_snprintf(lenstr, sizeof(lenstr), "%i", len);

	LIBMONETRA_http_set_header(http, "Content-Length", lenstr, 1);

	return 1;
}

/* Append last header in list */
static int LIBMONETRA_http_append_last_header(M_http_t http, const char *value)
{
	int num, len, oldlen;

	if ((http == NULL) || (value == NULL) || *value == '\0') {
		return 0;
	}

	num = http->num_headers - 1;
	if (num < 0) {
		return 0;
	}

	len = strlen(value);
	if (http->headers[num]->value == NULL) {
		oldlen = 0;
	} else {
		oldlen = strlen(http->headers[num]->value);
	}

	http->headers[num]->value = Ml_realloc(http->headers[num]->value, oldlen + len + 1);
	memcpy(http->headers[num]->value + oldlen, value, len);
	len += oldlen;
	http->headers[num]->value[len] = 0;

	return 1;
}

int LIBMONETRA_http_set_header(M_http_t http, const char *header, const char *value, int overwrite)
{
	if ((http == NULL) || (header == NULL) || (strlen(header) == 0) || (value == NULL)) {
		return 0;
	}

	/* remove any existing headers */
	if (overwrite != 0) {
		while (LIBMONETRA_http_del_header(http, header) == 1) {}
	}

	http->headers = Ml_realloc(http->headers, sizeof(M_http_header_t) * (http->num_headers + 1));
	http->headers[http->num_headers] = Ml_malloc(sizeof(struct M_http_header_s));
	http->headers[http->num_headers]->name = Ml_strdup(header);
	http->headers[http->num_headers]->value = Ml_strdup(value);

	http->num_headers++;

	return 1;
}

int LIBMONETRA_http_set_hosttype(M_http_t http, int hosttype, const char *path, const char *host, unsigned short port)
{
	if ((http == NULL) || ((hosttype != MHTTP_GET) && (hosttype != MHTTP_POST) && (hosttype != MHTTP_CONNECT))) {
		return 0;
	}

	http->hosttype = hosttype;

	Ml_free(http->path);
	if ((path == NULL) || (strlen(path) == 0)) {
		http->path = Ml_strdup("/");
	} else {
		http->path = Ml_strdup(path);
	}

	if (host != NULL) {
		http->host = Ml_strdup(host);
	}

	http->port = port;

	return 1;
}

int LIBMONETRA_http_set_resp_code(M_http_t http, int code, const char *code_msg)
{
	if ((http == NULL) || (code < 0) || (code_msg == NULL) || (strlen(code_msg) == 0)) {
		return 0;
	}

	http->code = code;
	if (http->code_msg == NULL) {
		Ml_free(http->code_msg);
	}

	http->code_msg = Ml_strdup(code_msg);

	return 1;
}

int LIBMONETRA_http_set_version(M_http_t http, int version)
{
	if ((http == NULL) || ((version != MHTTP_HTTP_VER_1_0) && (version != MHTTP_HTTP_VER_1_1))) {
		return 0;
	}

	http->version = version;

	return 1;
}
