#include "main.h"

#ifdef ENABLE_SSL
#  include <openssl/crypto.h>
#  include <openssl/x509.h>
#  include <openssl/pem.h>
#  include <openssl/ssl.h>
#  include <openssl/err.h>
#  include <openssl/rand.h>
#  ifndef OPENSSL_NO_ENGINE
#    include <openssl/engine.h>
#  endif
#  include <openssl/conf.h>
#endif /* ifdef ENABLE_SSL */

#include "atoint64.h"
#include "atouint64.h"
#include "base64.h"
#include "mstring.h"
#include "mstdlib.h"
#include "mlfile.h"

static int M_Initialization_Count = 0;
static char M_CaFilePath[1024] = { 0 };

static M_QUEUE *M_NewQueueSlot(M_CONN *myconn);
static int M_ProcessBuffer(M_CONN *conn);
static int M_IP_GetTransParams(const char *data, int len, int last_parse_attempt_len, M_uintptr *start, M_uintptr *fs, M_uintptr *finish);
static void M_ProcessResponse(M_CONN *conn, M_uint64 identifier, const char *msg, size_t len);
static M_QUEUE *M_FindQueue(M_CONN *conn, M_uint64 identifier);
static void M_add_resp_keyval(M_QUEUE *ptr, const char *key, const char *val);

/* BASE ROUTINES */
#ifdef ENABLE_SSL
static void **M_SSL_LOCKS = NULL;
static int M_NUM_SSL_LOCKS = 0;
static enum m_ssllock_style M_SSLLOCK_STYLE = M_SSLLOCK_NONE;
static M_Register_Mutex M_SSL_mutexreg = NULL;
static M_Unregister_Mutex M_SSL_mutexunreg = NULL;
static M_Mutex_Lock M_SSL_mutexlock = NULL;
static M_Mutex_Unlock M_SSL_mutexunlock = NULL;
static M_ThreadID M_SSL_ThreadID = NULL;

static int M_SSL_load_cadir(SSL_CTX *ctx, const char *path);

static void M_SSL_locking_callback(int mode, int n, const char *file, int line)
{
	UNUSED(file);
	UNUSED(line);

	if (
		(M_SSL_LOCKS == NULL)                   ||
		(n >= M_NUM_SSL_LOCKS)                  ||
		(M_SSLLOCK_STYLE != M_SSLLOCK_INTERNAL) ||
		(M_SSL_mutexlock == NULL)               ||
		(M_SSL_mutexunlock == NULL)
	) {
		return;
	}

	if (mode & CRYPTO_LOCK) {
		(*M_SSL_mutexlock)(&(M_SSL_LOCKS[n]));
	} else {
		(*M_SSL_mutexunlock)(&(M_SSL_LOCKS[n]));
	}
}

static int M_SSL_load_cadir(SSL_CTX *ctx, const char *path)
{
	char       **filenames = NULL;
	int          num_files = 0;
	int          i;
	size_t       len;
	FILE        *fp;
	X509_STORE  *store     = SSL_CTX_get_cert_store(ctx);
	X509        *x509;
	int          numloaded = 0;

	num_files = Ml_dirlist(&filenames, path);
	if (num_files == 0 || filenames == NULL)
		return 0;

	for (i=0; i<num_files; i++) {
		len = strlen(filenames[i]);
		if (len <= 4 || strcmp(&(filenames[i][len-4]), ".pem") != 0) {
			continue;
		}

		fp = fopen(filenames[i], "rb");
		if (fp == NULL) {
			goto cleanup;
		}
		
		x509 = PEM_read_X509(fp, NULL, NULL, NULL);
		if (x509 == NULL) {
			goto cleanup;
		}

		X509_STORE_add_cert(store, x509);
		X509_free(x509);
		numloaded++;

cleanup:
		fclose(fp);
	}

	return numloaded;
}
#endif /* ifdef ENABLE_SSL */

int M_SPEC M_InitEngine_ex(enum m_ssllock_style lockstyle)
{
	M_Initialization_Count++;
	/* XXX: It seems as if the static memory OpenSSL utilizes
	 * has issues with dlopen()'d modules that also use those
	 * libraries not being able to reference that internal data
	 * so the SSL_library_init() is going to be necessary still.
	 * Exhibits itself as this error:
	 * "SSL_CTX_new:unable to load ssl2 md5 routines"
	 * when upgrading Monetra via the Monetra Installer. Should
	 * probably be researched further. */
#if 0
	if (M_Initialization_Count != 1) {
		return 1;
	}
#endif

	if ((M_Initialization_Count == 1) && !M_InitSockets()) {
		return 0;
	}

#ifdef ENABLE_SSL
	if (lockstyle != M_SSLLOCK_EXTERNAL) {
		SSL_load_error_strings();
		SSL_library_init();
	}

	M_SSLLOCK_STYLE = lockstyle;
#else
	UNUSED(lockstyle);
#endif

	return 1;
}

int M_SPEC M_InitEngine(const char *location)
{
	int retval = 0;

	retval = M_InitEngine_ex(M_SSLLOCK_INTERNAL);
	if (!retval) {
		return retval;
	}

	if (
		((M_Initialization_Count == 1) || (*M_CaFilePath == '\0')) &&
		(location != NULL) && (*location != '\0')
	) {
		Ml_snprintf(M_CaFilePath, sizeof(M_CaFilePath), "%s", location);
	}

	return 1;
}

void M_SPEC M_DestroyEngine()
{
#ifdef ENABLE_SSL
	int i;

	M_Initialization_Count--;
	if (M_Initialization_Count != 0) {
		if (M_Initialization_Count < 0) {
			M_Initialization_Count = 0;
		}

		return;
	}

	memset(M_CaFilePath, 0, sizeof(M_CaFilePath));

	if (M_SSLLOCK_STYLE != M_SSLLOCK_EXTERNAL) {
		ERR_remove_state(0);
#  ifndef OPENSSL_NO_ENGINE
		ENGINE_cleanup();
#  endif
		CONF_modules_unload(1);
		ERR_free_strings();
		EVP_cleanup();
		CRYPTO_cleanup_all_ex_data();
	}

	if ((M_SSLLOCK_STYLE == M_SSLLOCK_INTERNAL) && M_NUM_SSL_LOCKS) {
		M_SSLLOCK_STYLE = M_SSLLOCK_NONE;
		for (i = 0; i < M_NUM_SSL_LOCKS; i++) {
			M_SSL_mutexunreg(&(M_SSL_LOCKS[i]));
		}

		M_NUM_SSL_LOCKS = 0;
		Ml_free(M_SSL_LOCKS);
		M_SSL_LOCKS = NULL;
		M_SSL_mutexreg = NULL;
		M_SSL_mutexunreg = NULL;
		M_SSL_mutexlock = NULL;
		M_SSL_mutexunlock = NULL;
		M_SSL_ThreadID = NULL;
	}

	M_SSLLOCK_STYLE = M_SSLLOCK_NONE;
#endif /* ifdef ENABLE_SSL */
}

/* Initializes the MCVE Connection structure */
void M_SPEC M_InitConn(M_CONN *conn)
{
	*conn = Ml_malloc(sizeof(**conn));
	memset(*conn, 0, sizeof(**conn));

	(*conn)->conn_method = -1;
	(*conn)->port = -1;
	(*conn)->ptr = -1;
	(*conn)->timeout = -1;
	(*conn)->verifyconn = 1;
	(*conn)->max_conn_time = 5;

#ifdef ENABLE_SSL
	if (*M_CaFilePath != '\0') {
		Ml_snprintf((*conn)->ca_location, sizeof((*conn)->ca_location), "%s", M_CaFilePath);
	}
#endif

	LIBMONETRA_proxy_config_init(&(*conn)->proxy);
	(*conn)->last_identifier = 0;
}

int M_SPEC M_Register_mutexinit(M_CONN *conn, M_Register_Mutex reg)
{
	(*conn)->mutexreg = reg;
	return 1;
}

int M_SPEC M_Register_mutexdestroy(M_CONN *conn, M_Unregister_Mutex reg)
{
	(*conn)->mutexunreg = reg;
	return 1;
}

int M_SPEC M_Register_mutexlock(M_CONN *conn, M_Mutex_Lock reg)
{
	(*conn)->mutexlock = reg;
	return 1;
}

int M_SPEC M_Register_mutexunlock(M_CONN *conn, M_Mutex_Unlock reg)
{
	(*conn)->mutexunlock = reg;
	return 1;
}

int M_SPEC M_Register_threadid(M_CONN *conn, M_ThreadID reg)
{
	(*conn)->threadid = reg;
	return 1;
}

int M_SPEC M_EnableThreadSafety(M_CONN *conn)
{
	if (
		((*conn)->mutexreg    == NULL) ||
		((*conn)->mutexunreg  == NULL) ||
		((*conn)->mutexlock   == NULL) ||
		((*conn)->mutexunlock == NULL)
	) {
		return 0;
	}

	M_init_locks(conn);
	if ((*conn)->mutex == NULL) {
		return 0;
	}

#ifdef ENABLE_SSL
	/* On first run of 'M_EnableThreadSafety' for the first connection,
	 * see if the OpenSSL needs to be initialized for Thread Safety */
	if (
		(M_SSLLOCK_STYLE == M_SSLLOCK_INTERNAL) &&
	    (M_NUM_SSL_LOCKS == 0)
	) {
		int i;
		M_NUM_SSL_LOCKS = CRYPTO_num_locks();
		M_SSL_LOCKS = Ml_malloc(M_NUM_SSL_LOCKS * sizeof(*M_SSL_LOCKS));
		M_SSL_mutexreg = (*conn)->mutexreg;
		M_SSL_mutexunreg = (*conn)->mutexunreg;
		M_SSL_mutexlock = (*conn)->mutexlock;
		M_SSL_mutexunlock = (*conn)->mutexunlock;
		M_SSL_ThreadID = (*conn)->threadid;

		for (i = 0; i < M_NUM_SSL_LOCKS; i++) {
			M_SSL_LOCKS[i] = M_SSL_mutexreg();
		}

		CRYPTO_set_locking_callback(M_SSL_locking_callback);
		CRYPTO_set_id_callback(M_SSL_ThreadID);
	}
#endif /* ifdef ENABLE_SSL */

	return 1;
}

int M_SPEC M_ValidateIdentifier(M_CONN *conn, int tf)
{
	(*conn)->validate_identifier = tf;
	return 1;
}

int M_SPEC M_SetBlocking(M_CONN *conn, int tf)
{
	(*conn)->blocking = tf;
	return 1;
}

int M_SPEC M_SetTimeout(M_CONN *conn, long timeout)
{
	(*conn)->timeout = timeout;
	return 1;
}

/* 0 = off, 1 = on */
int M_SPEC M_SetLogging(M_CONN *conn, int level)
{
	(*conn)->do_debug = level;
	return 1;
}

/* DropFile support has been removed, so M_SetDropFile has been rigged to fail.
 */
int M_SPEC M_SetDropFile(M_CONN *conn, const char *df_location)
{
	UNUSED(conn);
	UNUSED(df_location);

	return 0;
}

/* Sets method,location, and portnum for IP */
/* Returns 0 on error */
int M_SPEC M_SetIP(M_CONN *conn, const char *host, unsigned short port)
{
	if ((strlen(host) > 249) || (strlen(host) < 1)) {
		return 0;
	}

	M_strlcpy((*conn)->location, host, sizeof((*conn)->location));
	if (port < 1) {
		return 0;
	}

	(*conn)->port = port;
	(*conn)->conn_method = M_SOCKETS;

	return 1;
}

int M_SPEC M_SetSSL_CAfile(M_CONN *conn, const char *path)
{
#ifdef ENABLE_SSL
	M_strlcpy((*conn)->ca_location, path, sizeof((*conn)->ca_location));
	return 1;
#else
	UNUSED(conn);
	UNUSED(path);
	return 0;
#endif
}

int M_SPEC M_SetSSL_CAdir(M_CONN *conn, const char *path)
{
#ifdef ENABLE_SSL
	M_strlcpy((*conn)->ca_dir, path, sizeof((*conn)->ca_dir));
	return 1;
#else
	UNUSED(conn);
	UNUSED(path);
	return 0;
#endif
}

int M_SPEC M_SetSSL_Files(M_CONN *conn, const char *sslkeyfile, const char *sslcertfile)
{
#ifdef ENABLE_SSL
	if (sslkeyfile != NULL) {
		M_strlcpy((*conn)->sslkeyfile, sslkeyfile, sizeof((*conn)->sslkeyfile));
	}

	if (sslcertfile != NULL) {
		M_strlcpy((*conn)->sslcertfile, sslcertfile, sizeof((*conn)->sslcertfile));
	}

	return 1;
#else
	UNUSED(conn);
	UNUSED(sslkeyfile);
	UNUSED(sslcertfile);
	return 0;
#endif
}

int M_SPEC M_SetSSL(M_CONN *conn, const char *host, unsigned short port)
{
#ifdef ENABLE_SSL
	if (
		(host == NULL)       ||
		(strlen(host) > 249) ||
		(strlen(host) < 1)
	) {
		M_Set_Conn_Error(conn, "Invalid Host Name");
		return 0;
	}

	M_strlcpy((*conn)->location, host, sizeof((*conn)->location));
	if (port < 1) {
		M_Set_Conn_Error(conn, "Invalid Port Number");
		return 0;
	}

	(*conn)->port = port;
	(*conn)->conn_method = M_SSL;

	return 1;
#else /* ifdef ENABLE_SSL */
	UNUSED(conn);
	UNUSED(host);
	UNUSED(port);
	return 0;
#endif /* ifdef ENABLE_SSL */
}

void M_SPEC M_VerifyConnection(M_CONN *conn, int tf)
{
	(*conn)->verifyconn = tf;
}

void M_SPEC M_VerifySSLCert(M_CONN *conn, int level)
{
	(*conn)->verifyssl = level;
}



enum M_ssl_host_validate_level {
	M_SSL_HOST_VALIDATE_LEVEL_NONE  = 0, /*!< Perform no validation */
	M_SSL_HOST_VALIDATE_LEVEL_FULL  = 1, /*!< Require a full certificate name match (or wildcard cert) */
	M_SSL_HOST_VALIDATE_LEVEL_FUZZY = 2  /*!< Validate only the for the parent domain meaning this would
	                                      *   match:
	                                      *      gpgw3.globalpay.com == igusproda.globalpay.com
	                                      *   but this would not:
	                                      *      gpgw3.globalpay.com != igusproda.globalpayspoof.com */
};


static int M_ssl_domain_from_host(const char *hostname, char *domain, size_t domain_len)
{
	char       **strs;
	int          num  = 0;
	int          domain_sects;
	char        *out;

	/* Split at the periods */
	strs = Ml_str_explode_str('.', hostname, &num);

	/* If the last section is only 2 characters, then it is a 
	 * country domain and thus we need to include another domain
	 * section */
	if (strs && num && strlen(strs[num-1]) == 2) {
		domain_sects = 3; 
	} else {
		domain_sects = 2;
	}

	/* Either we need the whole thing, or there aren't
	 * enough desired sections, so just return the original
	 * input */
	if (domain_sects >= num) {
		Ml_str_explode_free(strs, num);
		Ml_str_cpy(domain, hostname, domain_len);
		return 1;
	}

	/* Implode the dots */
	out = Ml_str_implode('.', 0, 0, &(strs[num-domain_sects]), domain_sects, 0);
	Ml_str_explode_free(strs, num);

	if (out == NULL || *out == '\0') {
		Ml_str_cpy(domain, hostname, domain_len);
		return 0;
	}

	Ml_str_cpy(domain, out, domain_len);
	Ml_free(out);
	return 1;
}


static int M_ssl_host_validate(const char *hostname, const char *common_name, enum M_ssl_host_validate_level level)
{
	char common_domain[1024];
	char host_domain[1024];

	if (hostname == NULL || *hostname == '\0' || common_name == NULL || *common_name == '\0')
		return 0;

	if (level == M_SSL_HOST_VALIDATE_LEVEL_NONE)
		return 1;

	/* Lowercase the hostnames for matching purposes */
	Ml_str_cpy(common_domain, common_name, sizeof(common_domain));
	Ml_str_lower(common_domain);
	Ml_str_cpy(host_domain, hostname, sizeof(host_domain));
	Ml_str_lower(host_domain);

	/* Both FULL and FUZZY can try a pattern match.  We use pattern
	 * match for wildcard common name certificates */
	if (Ml_pattern_match(common_domain, host_domain))
		return 1;

	if (level == M_SSL_HOST_VALIDATE_LEVEL_FULL)
		return 0;

	/* Level must be FUZZY */
	memset(host_domain, 0, sizeof(host_domain));
	memset(common_domain, 0, sizeof(common_domain));
	M_ssl_domain_from_host(hostname, host_domain, sizeof(host_domain));
	M_ssl_domain_from_host(common_name, common_domain, sizeof(common_domain));
	if (strcasecmp(host_domain, common_domain) != 0)
		return 0;

	return 1;
}


static int M_SSL_validate_common_name(X509_NAME *subject, const char *hostname, enum M_ssl_host_validate_level level)
{
	int    idx   = -1;
	int    match = 0;

	while ((idx = X509_NAME_get_index_by_NID(subject, NID_commonName, idx)) != -1) {
		X509_NAME_ENTRY *e    = X509_NAME_get_entry(subject, idx);
		ASN1_STRING     *asn1;
		unsigned char   *cn   = NULL;
		int              len;

		asn1 = X509_NAME_ENTRY_get_data(e);
		if (asn1 == NULL)
			continue;

		len  = ASN1_STRING_to_UTF8(&cn, asn1);
		if (len <= 0 || cn == NULL)
			continue;

		if (M_ssl_host_validate(hostname, (const char *)cn, level)) {
			match = 1;
			OPENSSL_free(cn);
			break;
		}
		OPENSSL_free(cn);
	}
	return match;
}



int M_SPEC M_Connect(M_CONN *conn)
{
#ifdef ENABLE_SSL
	int connected;
	long err;
#endif

	if ((*conn)->conn_method == M_SOCKETS) {
		(*conn)->ptr = M_ip_connect(conn);
		if ((*conn)->ptr == -1) {
			return 0;       /* Connection Failed */
		}
	}

#ifdef ENABLE_SSL
	if ((*conn)->conn_method == M_SSL) {
		connected = 0;
		/* due to problems with simultaneous SSL connections on some client/
		 * server combinations, retry up to M_SSL_RETRIES times to connect
		 * if only SSL_connect() has failed
		 */
		do {
			(*conn)->client_ctx = SSL_CTX_new(SSLv23_client_method());
			if ((*conn)->client_ctx == NULL) {
				M_Set_Conn_Error(conn, "SSL_CTX_new() failed");
				return 0;
			}

			if (*((*conn)->ca_dir) != '\0') {
				if (M_SSL_load_cadir((*conn)->client_ctx, (*conn)->ca_dir) == 0) {
					M_Set_Conn_Error(conn, "Loading CA files for directory failed: No CA files could be loaded");
					SSL_CTX_free((*conn)->client_ctx);
					(*conn)->client_ctx = NULL;
					return 0;
				}
			} else {
				SSL_CTX_load_verify_locations((*conn)->client_ctx, (*conn)->ca_location, NULL);
			}

			if (*((*conn)->sslkeyfile) != '\0') {
				if (SSL_CTX_use_PrivateKey_file((*conn)->client_ctx, (*conn)->sslkeyfile, SSL_FILETYPE_PEM) != 1) {
					M_Set_Conn_Error(conn, "SSL_CTX_use_PrivateKey_file() failed");
					SSL_CTX_free((*conn)->client_ctx);
					(*conn)->client_ctx = NULL;
					return 0;
				}
			}

			if (*((*conn)->sslcertfile) != '\0') {
				if (SSL_CTX_use_certificate_file((*conn)->client_ctx, (*conn)->sslcertfile, SSL_FILETYPE_PEM) != 1) {
					M_Set_Conn_Error(conn, "SSL_CTX_use_certificate_file() failed");
					SSL_CTX_free((*conn)->client_ctx);
					(*conn)->client_ctx = NULL;
					return 0;
				}
			}

			(*conn)->ptr = M_ip_connect(conn);
			if ((*conn)->ptr == -1) {
				if ((*conn)->client_ctx != NULL) {
					SSL_CTX_free((*conn)->client_ctx);
					(*conn)->client_ctx = NULL;
				}

				return 0;
			}

			(*conn)->ssl = SSL_new((*conn)->client_ctx);
			if ((*conn)->ssl == NULL) {
				M_Set_Conn_Error(conn, "SSL_new() failed");
				return 0;
			}

			err = SSL_set_fd((*conn)->ssl, (*conn)->ptr);
			if (err == 0) {
				M_Set_Conn_Error(conn, "SSL_set_fd() failed");
				return 0;
			}

			/* Support the TLS SNI extension which advertises the hostname we are connecting
 			 * to as part of the negotiation.  This allows the server to share multiple domains
			 * via a single IP address and negotiate using different certificates. */
			SSL_set_tlsext_host_name((*conn)->ssl, (char *)(*conn)->location);

			if (SSL_connect((*conn)->ssl) != 1) {
				SSL_free((*conn)->ssl);
				SSL_CTX_free((*conn)->client_ctx);
				(*conn)->ssl = NULL;
				(*conn)->client_ctx = NULL;
				M_closesocket((*conn)->ptr, 1);
				(*conn)->ptr = -1;
				connected--;
				M_uwait(100000);
			} else {
				connected = 1;
			}
		} while (connected >= -M_SSL_RETRIES && connected != 1);

		if (connected != 1) {
			M_Set_Conn_Error(conn, "SSL Negotiation Failed");
			return 0;
		}

		(*conn)->server_cert = SSL_get_peer_certificate((*conn)->ssl);
		if ((*conn)->verifyssl == 1 || (*conn)->verifyssl == 2) {
			X509_NAME *x509_name = X509_get_subject_name((*conn)->server_cert);
			if (x509_name == NULL) {
				M_Set_Conn_Error(conn, "server peer certificate did not have a subject");
				M_closesocket((*conn)->ptr, 1);
				X509_free((*conn)->server_cert);
				SSL_free((SSL *)(*conn)->ssl);
				SSL_CTX_free((*conn)->client_ctx);
				(*conn)->server_cert = NULL;
				(*conn)->ssl         = NULL;
				(*conn)->client_ctx  = NULL;
				return 0;
			}

			if (!M_SSL_validate_common_name(x509_name, (*conn)->location, (*conn)->verifyssl == 1?M_SSL_HOST_VALIDATE_LEVEL_FULL:M_SSL_HOST_VALIDATE_LEVEL_FUZZY)) {
				M_Set_Conn_Error(conn, "SSL certificate common name mismatch");
				M_closesocket((*conn)->ptr, 1);
				X509_free((*conn)->server_cert);
				SSL_free((SSL *)(*conn)->ssl);
				SSL_CTX_free((*conn)->client_ctx);
				(*conn)->server_cert = NULL;
				(*conn)->ssl         = NULL;
				(*conn)->client_ctx  = NULL;
				return 0;
			}
		}
		/* printf("Server Certificate subject: %s\r\n", */
		/* X509_NAME_oneline(X509_get_subject_name((X509 */
		/* *)(*conn)->server_cert),0,0)); */
		/* printf("Server Certificate issuer: %s\r\n", */
		/* X509_NAME_oneline(X509_get_issuer_name((X509 */
		/* *)(*conn)->server_cert),0,0)); */

		if (!(*conn)->server_cert || ((*conn)->verifyssl && (SSL_get_verify_result((*conn)->ssl) != X509_V_OK))) {
			M_Set_Conn_Error(conn, "SSL Certificate verification failed");
			if ((*conn)->ptr != -1) {
				M_closesocket((*conn)->ptr, 1);
			}

			(*conn)->ptr = -1;
			if ((*conn)->server_cert) 
				X509_free((*conn)->server_cert);
			SSL_free((SSL *)(*conn)->ssl);
			SSL_CTX_free((*conn)->client_ctx);
			(*conn)->server_cert = NULL;
			(*conn)->ssl         = NULL;
			(*conn)->client_ctx  = NULL;
			return 0;
		}
	}
#endif /* ENABLE_SSL */

	if ((*conn)->conn_method == -1) {
		M_Set_Conn_Error(conn, "Must set connectivity method before calling M_Connect");
		return 0;
	}

	(*conn)->status = M_CONNECTED;
	if ((((*conn)->conn_method == M_SOCKETS) || ((*conn)->conn_method == M_SSL))
	    && (*conn)->verifyconn) {
		if (!M_VerifyPing(conn)) {
			M_Set_Conn_Error(conn, "Monetra did not respond to a PING request. Ensure proper port number and MCVE v2.1 or greater.");
			if ((*conn)->ptr != -1) {
				M_closesocket((*conn)->ptr, 1);
			}

			(*conn)->ptr = -1;
#ifdef ENABLE_SSL
			if ((*conn)->conn_method == M_SSL) {
				X509_free((*conn)->server_cert);
				SSL_free((*conn)->ssl);
				SSL_CTX_free((*conn)->client_ctx);
			}

			(*conn)->ssl = NULL;
			(*conn)->client_ctx = NULL;
			(*conn)->server_cert = NULL;
#endif
			(*conn)->status = M_DISCONNECTED;
			return 0;
		}
	}

	(*conn)->status = M_CONNECTED;
	return 1;
}

void M_SPEC M_MaxConnTimeout(M_CONN *conn, int maxtime)
{
	(*conn)->max_conn_time = maxtime;
}

const char *M_SPEC M_ConnectionError(M_CONN *conn)
{
	return (*conn)->error_text;
}

void M_SPEC M_DestroyConn(M_CONN *conn)
{
	while ((*conn)->queue != NULL) {
		M_DeleteTrans(conn, (M_uintptr)(*conn)->queue);
	}

#ifdef ENABLE_SSL
	if ((*conn)->server_cert != NULL) {
		X509_free((*conn)->server_cert);
	}

	if ((*conn)->conn_method == M_SSL) {
		if ((*conn)->ssl != NULL) {
			SSL_shutdown((*conn)->ssl);
			SSL_free((SSL *)(*conn)->ssl);
		}

		if ((*conn)->client_ctx != NULL) {
			SSL_CTX_free((*conn)->client_ctx);
		}
	}
#endif /* ifdef ENABLE_SSL */

	M_destroy_locks(conn);

	if ((*conn)->ptr != -1) {
		M_closesocket((*conn)->ptr, 1);
	}

	if ((*conn)->inbuf != NULL) {
		Ml_buf_cancel((*conn)->inbuf);
	}

	if ((*conn)->outbuf != NULL) {
		Ml_buf_cancel((*conn)->outbuf);
	}

	if ((*conn)->error_text != NULL) {
		Ml_free((*conn)->error_text);
	}

	if ((*conn)->queue != NULL) {
		Ml_free((*conn)->queue);
	}

	LIBMONETRA_proxy_config_free(&(*conn)->proxy);

	memset(*conn, 0, sizeof(**conn));
	(*conn)->conn_method = -1;
	(*conn)->port = -1;
	(*conn)->ptr = -1;
	(*conn)->max_conn_time = 5;
	Ml_free(*conn);
	*conn = NULL;
}

int M_SPEC M_Monitor_ex(M_CONN *conn, long wait_us)
{
	int ret = 0;

	if ((*conn)->status == M_DISCONNECTED) {
		M_Set_Conn_Error(conn, "Not Connected to Monetra");
		return 0;
	}

	if (
		((*conn)->conn_method == M_SOCKETS) ||
		((*conn)->conn_method == M_SSL)
	) {
		ret = M_Monitor_IP(conn, wait_us);

		if (!ret) {
			M_lock(conn);
			if ((*conn)->ptr != -1) {
				M_closesocket((*conn)->ptr, 0);
				(*conn)->ptr = -1;
#ifdef ENABLE_SSL
				if ((*conn)->conn_method == M_SSL) {
					SSL_free((SSL *)(*conn)->ssl);
					(*conn)->ssl = NULL;
				}
#endif /* ENABLE_SSL */

				M_Set_Conn_Error(conn, "Unexpected disconnection from Monetra");
				(*conn)->status = M_DISCONNECTED;
			}

			M_unlock(conn);
		} else if (!M_ProcessBuffer(conn)) {
			return 1;                      /* Not a disconnect, but parse error */
		}
	}

	return ret;
}

int M_SPEC M_Monitor(M_CONN *conn)
{
	return M_Monitor_ex(conn, 0);
}

int M_SPEC M_TransactionsSent(M_CONN *conn)
{
	int ret = 0;

	M_lock(conn);
	if ((*conn)->outbuf == NULL) {
		ret = 1;
	}

	M_unlock(conn);
	return ret;
}

void M_SPEC M_DeleteTrans(M_CONN *conn, M_uintptr identifier)
{
	int i;
	M_QUEUE *ptr = (M_QUEUE *)identifier;
	M_QUEUE *next = NULL, *prev = NULL, *base = NULL;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return;
	}

	M_lock(conn);
	if (ptr->transaction != NULL) {
		for (i = 0; i < ptr->transaction_fields; i++) {
			Ml_free(ptr->transaction[i].key);
			Ml_free(ptr->transaction[i].value);
		}

		Ml_free(ptr->transaction);
		ptr->transaction = NULL;
	}

	ptr->transaction_fields = 0;
	if (ptr->parsed_resp != NULL) {
		for (i = 0; i < ptr->resp_fields; i++) {
			Ml_free(ptr->parsed_resp[i].key);
			Ml_free(ptr->parsed_resp[i].value);
		}

		Ml_free(ptr->parsed_resp);
		ptr->parsed_resp = NULL;
	}

	ptr->resp_fields = 0;

	ptr->identifier = 0;
	ptr->status = M_UNUSED;
	if (ptr->response != NULL) {
		Ml_free(ptr->response);
	}

	ptr->response = NULL;
	ptr->iscommadelimited = 0;
	Ml_csv_free(ptr->csv);
	ptr->csv = NULL;

	(*conn)->outstanding_auths--;

	prev = ptr->prev;
	next = ptr->next;
	base = (*conn)->queue;

	/* We're only trans in queue */
	if ((base == ptr) && (ptr->next == ptr)) {
		(*conn)->queue = NULL;
		/* we're first trans in queue */
	} else if (base == ptr) {
		(*conn)->queue = next;
		next->prev = prev;
		prev->next = next;
		/* we're in the middle */
	} else {
		next->prev = prev;
		prev->next = next;
	}

	Ml_free(ptr);
	M_unlock(conn);
}

/* Add a custom Transaction Parameter (modules, etc) */
static int M_TransParam_Add(M_CONN *conn, M_uintptr identifier, const char *key, const char *value)
{
	int field;
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	field = ptr->transaction_fields;

	if (field % 10 == 0) {
		ptr->transaction = Ml_realloc(ptr->transaction, (field + 10) * sizeof(M_TRANS));
	}

	ptr->transaction[field].key = Ml_strdup(key);
	ptr->transaction[field].value = Ml_strdup(value);
	ptr->transaction_fields++;

	return 1;
}

/* Get a new Queue Number, return the location */
M_uintptr M_SPEC M_TransNew(M_CONN *conn)
{
	M_QUEUE *ptr = NULL;
	char timeout[20];

	M_lock(conn);
	ptr = M_NewQueueSlot(conn);
	ptr->status = M_NEW;
	ptr->identifier = (*conn)->last_identifier++;
	M_unlock(conn);

	if ((*conn)->timeout > 0) {
		Ml_snprintf(timeout, sizeof(timeout), "%ld", (*conn)->timeout);
		M_TransParam_Add(conn, (M_uintptr)ptr, "timeout", timeout);
	}

	return (M_uintptr)ptr;
}

static int M_VerifyTrans(M_CONN *conn, M_uintptr identifier)
{
	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	return 1;
}

int M_SPEC M_TransSend(M_CONN *conn, M_uintptr identifier)
{
	int ret;
	M_QUEUE *ptr = NULL;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	ptr = (M_QUEUE *)identifier;

	/* Can only send transactions in the M_NEW state */
	if (ptr->status != M_NEW) {
		return 0;
	}

	if (!M_VerifyTrans(conn, identifier)) {
		return 0;
	}

	ret = M_SendTransaction(conn, ptr);

	if (ret) {
		(*conn)->outstanding_auths++;

		/* Blocking mode is set, wait until trans is complete */
		if ((*conn)->blocking) {
			while (M_CheckStatus(conn, identifier) == M_PENDING) {
				if (!M_Monitor_ex(conn, 50000)) {
					return 0;
				}
			}
		}
	}

	return ret;
}

const char *M_SPEC M_ResponseParam(M_CONN *conn, M_uintptr identifier, const char *key)
{
	int i;
	M_QUEUE *ptr = NULL;
	const char *val = NULL;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	ptr = (M_QUEUE *)identifier;
	M_lock(conn);
	for (i = 0; i < ptr->resp_fields; i++) {
		if (strcasecmp(key, ptr->parsed_resp[i].key) == 0) {
			val = ptr->parsed_resp[i].value;
			break;
		}
	}

	M_unlock(conn);
	return val;
}

unsigned char *M_SPEC M_ResponseBinaryParam(M_CONN *conn, M_uintptr identifier, const char *key, int *out_length)
{
	const char *param = NULL;
	unsigned char *ret = NULL;
	size_t src_len;

	*out_length = 0;
	param = M_ResponseParam(conn, identifier, key);
	if (param == NULL) {
		return NULL;
	}

	/* base64 decode */
	src_len = strlen(param);
	ret = malloc(src_len);  /* can't use Ml_malloc */
	(*out_length) = base64_decode(param, src_len, (char *)ret, src_len);
	if (*out_length <= 0) {
		free(ret); ret = NULL;  /* can't use Ml_free */
		*out_length = 0;
	}

	return ret;
}

char **M_SPEC M_ResponseKeys(M_CONN *conn, M_uintptr identifier, int *num_keys)
{
	char **ret = NULL;
	int i;
	M_QUEUE *ptr = NULL;

	(*num_keys) = 0;
	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	ptr = (M_QUEUE *)identifier;
	M_lock(conn);
	(*num_keys) = ptr->resp_fields;
	ret = malloc((*num_keys) * sizeof(char *));  /* can't use Ml_malloc */

	for (i = 0; i < (*num_keys); i++) {
		ret[i] = strdup(M_Safe(ptr->parsed_resp[i].key));  /* can't use Ml_strdup */
	}

	M_unlock(conn);
	return ret;
}

const char *M_SPEC M_ResponseKeys_index(char **keys, int num_keys, int idx)
{
	if (keys == NULL) {
		return NULL;
	}

	if (idx < 0) {
		return NULL;
	}

	if (idx >= num_keys) {
		return NULL;
	}

	return keys[idx];
}

int M_SPEC M_FreeResponseKeys(char **keys, int num_keys)
{
	int i;

	for (i = 0; i < num_keys; i++) {
		free(keys[i]);  /* can't use Ml_free */
	}

	free(keys);  /* can't use Ml_free */

	return 1;
}

/* Instead of using MC_CUSTOM for M_TransParam, use this ... infact, deprecate
 * M_TransParam ! */
int M_SPEC M_TransKeyVal(M_CONN *conn, M_uintptr identifier, const char *key, const char *value)
{
	return M_TransParam_Add(conn, identifier, key, value);
}

static int base64_remove_returns(char *source)
{
	int cnt = 0, i;

	for (i = 0; source[i] != 0; i++) {
		if ((source[i] != '\n') && (source[i] != '\r')) {
			source[cnt++] = source[i];
		}
	}

	source[cnt] = 0;

	return cnt;
}

int M_SPEC M_TransBinaryKeyVal(M_CONN *conn, M_uintptr identifier, const char *key, const char *value, int value_len)
{
	char *b64 = NULL;
	int len = 0;
	int ret;

	if ((value_len == 0) || (value == NULL)) {
		return 0;
	}

	b64 = Ml_malloc(B64_TARGSIZE(value_len));
	len = base64_encode(value, value_len, b64, B64_TARGSIZE(value_len));
	if (len <= 0) {
		return 0;
	}

	base64_remove_returns(b64);
	ret = M_TransParam_Add(conn, identifier, key, b64);
	Ml_free(b64);

	return ret;
}

/* __stdcall calling convention does not support varargs */
#ifndef LIBMONETRA_SPEC_STDCALL
/* Add a new Transaction Parameter */
int M_SPEC M_TransParam(M_CONN *conn, M_uintptr identifier, int key, ...)
{
	char key_string[100];
	char value_string[1024];
	int trans_type;
	int excharges_type;
	int temp_tf;
	char *temp_str = NULL;
	va_list ap;
	const char *temp = NULL;
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	/* Make sure people don't add params to structures not in the M_NEW
	 * state */
	if (ptr->status != M_NEW) {
		return 0;
	}

	key_string[0] = 0;
	value_string[0] = 0;
	va_start(ap, key);

	switch (key) {
		case MC_TRANTYPE:
			M_strlcpy(key_string, "action", sizeof(key_string));
			trans_type = va_arg(ap, int);
			if (trans_type < 1000) {
				M_strlcpy(value_string, M_GetTypeString(trans_type), sizeof(value_string));
			} else {
				M_strlcpy(value_string, M_GetEngineAdminTypeString(trans_type), sizeof(value_string));
			}

			break;

		case MC_ADMIN:
			M_strlcpy(key_string, "admin", sizeof(key_string));
			trans_type = va_arg(ap, int);
			M_strlcpy(value_string, M_GetAdminTypeString(trans_type), sizeof(value_string));
			break;

		case MC_USERNAME:
			M_strlcpy(key_string, "username", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_PASSWORD:
			M_strlcpy(key_string, "password", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_ACCOUNT:
			M_strlcpy(key_string, "account", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_TRACKDATA:
			M_strlcpy(key_string, "trackdata", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_EXPDATE:
			M_strlcpy(key_string, "expdate", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_STREET:
			M_strlcpy(key_string, "street", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_ZIP:
			M_strlcpy(key_string, "zip", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_CV:
			M_strlcpy(key_string, "cvv2", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_COMMENTS:
			M_strlcpy(key_string, "comments", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_CLERKID:
			M_strlcpy(key_string, "clerkid", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_STATIONID:
			M_strlcpy(key_string, "stationid", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_APPRCODE:
			M_strlcpy(key_string, "apprcode", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_AMOUNT:
			M_strlcpy(key_string, "amount", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%.2f", va_arg(ap, double));
			break;

		case MC_PTRANNUM:
			M_strlcpy(key_string, "ptrannum", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%ld", va_arg(ap, long));
			break;

		case MC_TIMESTAMP:
			M_strlcpy(key_string, "timestamp", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%ld", va_arg(ap, long));
			break;

		case MC_TTID:
			M_strlcpy(key_string, "ttid", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%lld", va_arg(ap, M_int64));
			break;

		case MC_USER:
			M_strlcpy(key_string, "user", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_PWD:
			M_strlcpy(key_string, "pwd", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_ACCT:
			M_strlcpy(key_string, "acct", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_BDATE:
			M_strlcpy(key_string, "bdate", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_EDATE:
			M_strlcpy(key_string, "edate", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_BATCH:
			M_strlcpy(key_string, "batch", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_FILE:
			M_strlcpy(key_string, "file", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_AUDITTYPE:
			M_strlcpy(key_string, "type", sizeof(key_string));
			trans_type = va_arg(ap, int);
			M_strlcpy(value_string, M_GetTypeString(trans_type), sizeof(value_string));
			break;

		case MC_CUSTOM:
			M_strlcpy(key_string, va_arg(ap, char *), sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		/* Additional args for restaurant, lodging and auto rental */
		case MC_EXAMOUNT:
			M_strlcpy(key_string, "examount", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%.2f", va_arg(ap, double));
			break;

		case MC_EXCHARGES:
			M_strlcpy(key_string, "excharges", sizeof(key_string));
			excharges_type = va_arg(ap, int);
			M_strlcpy(value_string, M_GetExchargesArgString(excharges_type), sizeof(value_string));
			break;

		case MC_RATE:
			M_strlcpy(key_string, "rate", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%.2f", va_arg(ap, double));
			break;

		case MC_RENTERNAME:
			M_strlcpy(key_string, "rentername", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_RETURNCITY:
			M_strlcpy(key_string, "returncity", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_RETURNSTATE:
			M_strlcpy(key_string, "returnstate", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_RETURNLOCATION:
			M_strlcpy(key_string, "returnlocation", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_INQUIRY:
			M_strlcpy(key_string, "inquiry", sizeof(key_string));
			temp_tf = va_arg(ap, int);
			if (temp_tf) {
				M_strlcpy(value_string, "yes", sizeof(value_string));
			} else {
				M_strlcpy(value_string, "no", sizeof(value_string));
			}
			break;

		case MC_PRIORITY:
			M_strlcpy(key_string, "priority", sizeof(key_string));
			M_strlcpy(value_string, M_GetPriorityString(va_arg(ap, int)), sizeof(value_string));
			break;

		case MC_CARDTYPES:
			M_strlcpy(key_string, "cardtypes", sizeof(key_string));
			M_strlcpy(value_string, M_GetCardTypesString(va_arg(ap, int)), sizeof(value_string));
			break;

		case MC_SUB:
			M_strlcpy(key_string, "sub", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%d", va_arg(ap, int));
			break;

		case MC_MARKER:
			M_strlcpy(key_string, "marker", sizeof(key_string));
			Ml_snprintf(value_string, sizeof(value_string), "%ld", va_arg(ap, long));
			break;

		case MC_DEVICETYPE:
			M_strlcpy(key_string, "devicetype", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_ERRORCODE:
			M_strlcpy(key_string, "errorcode", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_NEWBATCH:
			M_strlcpy(key_string, "newbatch", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_CURR:
			M_strlcpy(key_string, "curr", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_DESCMERCH:
			M_strlcpy(key_string, "descmerch", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_DESCLOC:
			M_strlcpy(key_string, "descloc", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		case MC_ORIGTYPE:
			M_strlcpy(key_string, "origtype", sizeof(key_string));
			trans_type = va_arg(ap, int);
			if (trans_type < 1000) {
				M_strlcpy(value_string, M_GetTypeString(trans_type), sizeof(value_string));
			} else {
				M_strlcpy(value_string, M_GetEngineAdminTypeString(trans_type), sizeof(value_string));
			}

			break;

		case MC_VOIDORIGTYPE:
			M_strlcpy(key_string, "voidorigtype", sizeof(key_string));
			trans_type = va_arg(ap, int);
			if (trans_type < 1000) {
				M_strlcpy(value_string, M_GetTypeString(trans_type), sizeof(value_string));
			} else {
				M_strlcpy(value_string, M_GetEngineAdminTypeString(trans_type), sizeof(value_string));
			}

			break;

		case MC_PIN:
			M_strlcpy(key_string, "pin", sizeof(key_string));
			M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
			break;

		default:
			if (key >= 2000) {
				temp = M_GetUserArgString(key);
				if ((temp != NULL) && strlen(temp)) {
					M_strlcpy(key_string, temp, sizeof(key_string));
					if (key == MC_USER_SUB) {
						Ml_snprintf(value_string, sizeof(value_string), "%d", va_arg(ap, int));
					} else if (key == MC_USER_CARDTYPES) {
						temp_str = M_GetCardTypesString(va_arg(ap, int));
						M_strlcpy(value_string, temp_str, sizeof(value_string));
						Ml_free(temp_str);
					} else if (key == MC_USER_MODE) {
						temp_str = M_GetModeString(va_arg(ap, int));
						M_strlcpy(value_string, temp_str, sizeof(value_string));
						Ml_free(temp_str);
					} else {
						M_strlcpy(value_string, va_arg(ap, char *), sizeof(value_string));
					}
				}
			}

			break;
	}

	va_end(ap);
	/* should allow blank value */
	if (key_string != NULL && *key_string != '\0') {
		/* Maintain compatability with MCVE 2.0-2.1 */
		if (strcasecmp(key_string, "ttid") == 0) {
			M_TransParam_Add(conn, identifier, "sid", value_string);
		}

		return M_TransParam_Add(conn, identifier, key_string, value_string);
	}

	return 0;
}
#endif /* ifndef LIBMONETRA_SPEC_STDCALL */

/* RESPONSE CHECKING ROUTINES */

int M_SPEC M_ReturnStatus(M_CONN *conn, M_uintptr identifier)
{
	int code;
	int status;
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	code = M_ReturnCode(conn, identifier);

	M_lock(conn);
	if (ptr->status != M_DONE) {
		status = M_ERROR;
	} else if (code == -1) {
		status = M_SUCCESS;
	} else if (code == M_ERROR) {
		status = M_ERROR;
	} else if (code == M_SUCCESS) {
		status = M_SUCCESS;
	} else if (code == M_AUTH) {
		status = M_SUCCESS;
	} else { status = M_FAIL; }

	M_unlock(conn);

	return status;
}

int M_SPEC M_ReturnCode(M_CONN *conn, M_uintptr identifier)
{
	const char *v;
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	if (ptr->iscommadelimited) {
		return M_SUCCESS;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "code");
	M_unlock(conn);

	if (v == NULL) {
		return M_ERROR;
	}

	if (strcmp(v, "AUTH") == 0) {
		return M_AUTH;
	}

	if (strcmp(v, "DENY") == 0) {
		return M_DENY;
	}

	if (strcmp(v, "DUPL") == 0) {
		return M_DUPL;
	}

	if (strcmp(v, "CALL") == 0) {
		return M_CALL;
	}

	if (strcmp(v, "PKUP") == 0) {
		return M_PKUP;
	}

	if (strcmp(v, "RETRY") == 0) {
		return M_RETRY;
	}

	if (strcmp(v, "SETUP") == 0) {
		return M_SETUP;
	}

	if (strcmp(v, "SUCCESS") == 0) {
		return M_SUCCESS;
	}

	if (strcmp(v, "FAIL") == 0) {
		return M_FAIL;
	}

	if (strcmp(v, "TIMEOUT") == 0) {
		return M_TIMEOUT;
	}

	return M_DENY;
}

long M_SPEC M_TransactionItem(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "item");
	M_unlock(conn);

	return (v == NULL)
		? -1
		: atol(v);
}

long M_SPEC M_TransactionBatch(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "batch");
	M_unlock(conn);

	return (v == NULL)
		? -1
		: atol(v);
}

M_int64 M_SPEC M_TransactionID(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "ttid");
	M_unlock(conn);

	return (v == NULL)
		? -1
		: M_atoll(v);
}

const char *M_SPEC M_TransactionAuth(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "auth");
	M_unlock(conn);

	return v;
}

const char *M_SPEC M_TransactionText(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "verbiage");
	M_unlock(conn);

	return v;
}

int M_SPEC M_TransactionAVS(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "avs");
	M_unlock(conn);

	if (v == NULL) {
		return -1;
	}

	if (strcmp(v, "GOOD") == 0) {
		return M_GOOD;
	}

	if (strcmp(v, "BAD") == 0) {
		return M_BAD;
	}

	if (strcmp(v, "STREET") == 0) {
		return M_STREET;
	}

	if (strcmp(v, "ZIP") == 0) {
		return M_ZIP;
	}

	return M_UNKNOWN;
}

int M_SPEC M_TransactionCV(M_CONN *conn, M_uintptr identifier)
{
	const char *v;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	v = M_ResponseParam(conn, identifier, "cv");
	M_unlock(conn);

	if (v == NULL) {
		return -1;
	}

	if (strcmp(v, "GOOD") == 0) {
		return M_GOOD;
	}

	if (strcmp(v, "BAD") == 0) {
		return M_BAD;
	}

	return M_UNKNOWN;
}

long M_SPEC M_TransInQueue(M_CONN *conn)
{
	long auths;

	M_lock(conn);
	auths = (*conn)->outstanding_auths;
	M_unlock(conn);
	return auths;
}

int M_SPEC M_CheckStatus(M_CONN *conn, M_uintptr identifier)
{
	int status;
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	M_lock(conn);
	if ((*conn)->status == M_DISCONNECTED) {
		status = M_ERROR;
	} else {
		status = ptr->status;
	}

	M_unlock(conn);
	return status;
}

long M_SPEC M_CompleteAuthorizations(M_CONN *conn, M_uintptr **listings)
{
	long num = 0;
	M_QUEUE *ptr = NULL;

	M_lock(conn);
	ptr = (*conn)->queue;

	while (ptr != NULL) {
		if (ptr->status == M_DONE) {
			num++;
		}

		ptr = ptr->next;
		if (ptr == (*conn)->queue) {
			break;
		}
	}

	if ((num != 0) && (listings != NULL)) {
		(*listings) = malloc((num) * sizeof(M_uintptr));  /* can't use Ml_malloc */
		num = 0;
		ptr = (*conn)->queue;
		while (ptr != NULL) {
			if (ptr->status == M_DONE) {
				(*listings)[num++] = (M_uintptr)ptr;
			}

			ptr = ptr->next;
			if (ptr == (*conn)->queue) {
				break;
			}
		}
	}

	M_unlock(conn);

	return num;
}

/* For non-C compatible languages (ones that don't understand arrays) */
M_uintptr M_SPEC M_CompleteAuthorizations_index(M_uintptr *listings, int num_listings, int idx)
{
	if (idx < 0) {
		return 0;
	}

	if (idx >= num_listings) {
		return 0;
	}

	if (listings == NULL) {
		return 0;
	}

	return listings[idx];
}

/* For non-C compatible languages (ones that don't understand arrays) */
int M_SPEC M_FreeCompleteAuthorizations(M_uintptr *listings, int num_listings)
{
	if (listings == NULL) {
		return 0;
	}

	if (num_listings <= 0) {
		return 0;
	}

	free(listings);  /* can't use Ml_free */

	return 1;
}

/* COMMA DELIMITED ROUTINES */

int M_SPEC M_IsCommaDelimited(M_CONN *conn, M_uintptr identifier)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	return ptr->iscommadelimited;
}

int M_SPEC M_ParseCommaDelimited(M_CONN *conn, M_uintptr identifier)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier))
		return 0;

	if (!ptr->response)
		return 0;

	ptr->csv = Ml_csv_parse_inplace(ptr->response, strlen(ptr->response), ',', '"');
	if (!ptr->csv)
		return 0;

	/* If CSV parsing succeeded, it took ownership of the 'response' buffer,
	 * note that here */
	ptr->response = NULL;
	return 1;
}

const char *M_SPEC M_GetCommaDelimited(M_CONN *conn, M_uintptr identifier)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	return ptr->response;
}

const char *M_SPEC M_GetCell(M_CONN *conn, M_uintptr identifier, const char *column, long row)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;
	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}
	return Ml_csv_get_cell(ptr->csv, row, column);
}

char *M_SPEC M_GetBinaryCell(M_CONN *conn, M_uintptr identifier, const char *column, long row, int *out_length)
{
	const char *cell = NULL;
	char *ret = NULL;
	int src_len = 0;

	(*out_length) = 0;
	cell = M_GetCell(conn, identifier, column, row);

	if (cell == NULL) {
		return NULL;
	}

	/* Base64 decode cell */
	src_len = strlen(cell);
	ret = malloc(src_len);  /* can't use Ml_malloc */
	(*out_length) = base64_decode(cell, src_len, ret, src_len);
	if (*out_length <= 0) {
		free(ret);  /* can't use Ml_free */
		ret = NULL;
		*out_length = 0;
	}

	return ret;
}

void M_SPEC M_FreeBinaryCell(char *cell)
{
	free(cell);  /* can't use Ml_free */
}

const char *M_SPEC M_GetCellByNum(M_CONN *conn, M_uintptr identifier, int column, long row)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	/* checks */
	if (column < 0) {
		return NULL;
	}

	if (row < 0) {
		return NULL;
	}

	if (column >= Ml_csv_get_numcols(ptr->csv)) {
		return NULL;
	}

	if (row >= Ml_csv_get_numrows(ptr->csv)) {
		return NULL;
	}

	return Ml_csv_get_cellbynum(ptr->csv, row, column);
}

int M_SPEC M_NumColumns(M_CONN *conn, M_uintptr identifier)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	return Ml_csv_get_numcols(ptr->csv);
}

long M_SPEC M_NumRows(M_CONN *conn, M_uintptr identifier)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return 0;
	}

	return Ml_csv_get_numrows(ptr->csv);
}

const char *M_SPEC M_GetHeader(M_CONN *conn, M_uintptr identifier, int column_num)
{
	M_QUEUE *ptr = (M_QUEUE *)identifier;

	if (!Ml_verify_trans_in_queue(conn, identifier)) {
		return NULL;
	}

	return Ml_csv_get_header(ptr->csv, column_num);
}

/* TEXT RESPONSE CONVERSIONS */

const char *M_SPEC M_TEXT_Code(int code)
{
	const char *text = NULL;

	switch (code) {
		case M_AUTH:
			text = "AUTH";
			break;
		case M_DENY:
			text = "DENY";
			break;
		case M_DUPL:
			text = "DUPLICATE TRANS";
			break;
		case M_CALL:
			text = "CALL PROCESSOR";
			break;
		case M_PKUP:
			text = "PICKUP CARD";
			break;
		case M_RETRY:
			text = "RETRY";
			break;
		case M_SETUP:
			text = "SETUP";
			break;
		case M_TIMEOUT:
			text = "TIMEOUT";
			break;
		case M_SUCCESS:
			text = "SUCCESS";
			break;
		case M_FAIL:
			text = "FAIL";
			break;
		case M_ERROR:
			text = "ERROR";
			break;
		default:
			text = "UNKNOWN";
			break;
	}

	return text;
}

const char *M_SPEC M_TEXT_AVS(int code)
{
	const char *text = NULL;

	switch (code) {
		case M_STREET:
			text = "STREET FAILED";
			break;
		case M_ZIP:
			text = "ZIP FAILED";
			break;
		case M_GOOD:
			text = "GOOD";
			break;
		case M_BAD:
			text = "BAD";
			break;
		case M_UNKNOWN:
			text = "UNKNOWN";
			break;
		default:
			text = "UNKNOWN";
			break;
	}

	return text;
}

const char *M_SPEC M_TEXT_CV(int code)
{
	return M_TEXT_AVS(code);
}

#ifndef O_BINARY
#    define O_BINARY 0
#endif

char *M_SPEC M_SSLCert_gen_hash(const char *filename)
{
#ifdef ENABLE_SSL
	BIO *bio;
	X509 *x509;
	char *buf, *fingerprint;
	int retval;
	size_t size;
	unsigned char md[EVP_MAX_MD_SIZE];
	unsigned int i, md_size;

	buf = M_readfile(filename, &size);
	if (buf == NULL)
		return 0;

	bio = BIO_new_mem_buf(buf, strlen(buf));
	if (bio == NULL) {
		Ml_free(buf);
		return 0;
	}

	x509 = PEM_read_bio_X509_AUX(bio, NULL, NULL, NULL);
	BIO_free(bio);
	Ml_free(buf);

	md_size = sizeof(md);

	retval = X509_digest(x509, EVP_sha1(), md, &md_size);
	X509_free(x509);
	if ((retval == 0) || (md_size == 0)) {
		return 0;
	}

	fingerprint = malloc(md_size * 3);  /* can't use Ml_malloc */
	if (fingerprint == NULL) {
		return 0;
	}

	memset(fingerprint, 0, md_size * 3);

	for (i = 0; i < md_size; i++) {
		Ml_snprintf(fingerprint + i * 3, md_size * 3, "%02x%c", md[i], (i + 1 == md_size) ? '\0' : ':');
	}

	return fingerprint;
#else /* ENABLE_SSL */
	UNUSED(filename);
	return NULL;
#endif /* ENABLE_SSL */
}

int M_SPEC M_SetProxyHost(M_CONN *conn, const char *host, unsigned short port, const char *type)
{
	int proxy_type = 0;

	if (type == NULL) {
		return 0;
	}

	if (
		(strcasecmp(type, "WEB") == 0) ||
		(strcasecmp(type, "HTTP") == 0)
	) {
		proxy_type = PROXY_TYPE_HTTP;
	} else if (
		(strcasecmp(type, "SOCKS") == 0) ||
		(strcasecmp(type, "SOCKS5") == 0) ||
		(strcasecmp(type, "SOCKSv5") == 0)
	) {
		proxy_type = PROXY_TYPE_SOCKS5;
	} else {
		return 0;
	}

	return LIBMONETRA_proxy_config_host(&(*conn)->proxy, host, port, proxy_type);
}

int M_SPEC M_SetProxyTimeout(M_CONN *conn, int timeout)
{
	return LIBMONETRA_proxy_config_timeout(&(*conn)->proxy, timeout);
}

int M_SPEC M_SetProxyUser(M_CONN *conn, const char *username, const char *password)
{
	return LIBMONETRA_proxy_config_user(&(*conn)->proxy, username, password);
}

int M_SPEC M_SetProxyLocalNets(M_CONN *conn, const char *localnets, char *error, int errlen)
{
	char myerror[255];

	if (error == NULL) {
		error = myerror;
		errlen = sizeof(myerror);
	}

	return LIBMONETRA_proxy_config_noproxy(&(*conn)->proxy, localnets, error, errlen);
}

static M_QUEUE *M_NewQueueSlot(M_CONN *conn)
{
	M_QUEUE *ptr = NULL, *base = NULL, *last = NULL;

	ptr = Ml_malloc(sizeof(M_QUEUE));
	ptr->next = NULL;
	ptr->prev = NULL;
	ptr->status = M_UNUSED;
	ptr->identifier = 0;
	ptr->status = M_NEW;
	ptr->transaction_fields = 0;
	ptr->transaction = NULL;

	ptr->resp_fields = 0;
	ptr->parsed_resp = NULL;
	ptr->response = NULL;
	ptr->iscommadelimited = 0;

	ptr->csv = NULL;

	base = (*conn)->queue;
	/* Circular queue! */
	if (base == NULL) {
		(*conn)->queue = ptr;
		ptr->next = ptr;
		ptr->prev = ptr;
	} else {
		last = base->prev;
		last->next = ptr;
		ptr->prev = last;
		ptr->next = base;
		base->prev = ptr;
	}

	(*conn)->queue_length++;

	return ptr;
}

/* return 1 on success, return 0 on critical failure */
static int M_ProcessBuffer(M_CONN *conn)
{
	size_t parse_offset = 0;
	M_uintptr parse_start = 0, parse_fs = 0, parse_finish = 0;
	M_uint64 identifier;
	const char *message = NULL;
	char temp[32];
	char error[64];
	const char *inbuf;

	M_lock(conn);

	if ((*conn)->inbuf != NULL && (size_t)(*conn)->last_parse_attempt_len < Ml_buf_len((*conn)->inbuf)) {
		parse_offset = 0;
		while (
			parse_offset < Ml_buf_len((*conn)->inbuf) &&
			M_IP_GetTransParams(
				Ml_buf_peek((*conn)->inbuf) + parse_offset,
			    Ml_buf_len((*conn)->inbuf) - parse_offset,
			    (*conn)->last_parse_attempt_len,
			    &parse_start, &parse_fs,
			    &parse_finish
			)
		) {
			/* Reset this so that it will rescan the entire
			 * buffer, rather than the partial buffer, as this
			 * flag is an optimization for _large_ reports */
			(*conn)->last_parse_attempt_len = 0;

			if (parse_start != 0) {
				M_unlock(conn);
				Ml_snprintf(error, sizeof(error), "byte %d of %d, first character not STX", (int)parse_offset, (int)Ml_buf_len((*conn)->inbuf));
				M_Set_Conn_Error(conn, error);
				return 0;
			}

			inbuf = Ml_buf_peek((*conn)->inbuf) + parse_offset;

			/* Get identifier */
			memcpy(temp, inbuf + parse_start + 1, parse_fs - parse_start - 1);
			temp[parse_fs - parse_start - 1] = 0;
			identifier = M_atouint64(temp);

			/* Get message */
			message = inbuf + (parse_fs + 1);
			M_ProcessResponse(conn, identifier, message, parse_finish - parse_fs - 1);

			/* Track place */
			parse_offset += (parse_finish + 1);
		}

		if (parse_offset) {
			if (parse_offset < Ml_buf_len((*conn)->inbuf)) {
				Ml_buf_drop((*conn)->inbuf, parse_offset);
			} else {
				Ml_buf_cancel((*conn)->inbuf);
				(*conn)->inbuf                  = NULL;
				(*conn)->last_parse_attempt_len = 0;
			}
		}

		if ((*conn)->inbuf != NULL)
			(*conn)->last_parse_attempt_len = Ml_buf_len((*conn)->inbuf);

	}

	M_unlock(conn);
	return 1;
}

static int M_IP_GetTransParams(const char *data, int len, int last_parse_attempt_len, M_uintptr *start, M_uintptr *fs, M_uintptr *finish)
{
	const char *end_ptr, *fs_ptr = NULL;

	(*start) = 0;
	(*fs) = 0;
	(*finish) = 0;

	if ((data == NULL) || (len < 3)) {
		return 0;
	}

	if (last_parse_attempt_len > len) {
		last_parse_attempt_len = 0;
	}

	/* Find end first, as fast as possible, as it may not be
	 * there yet */
	end_ptr = memchr(data + last_parse_attempt_len, 0x03, len - last_parse_attempt_len);
	if (end_ptr == NULL) {
		return 0;
	}

	(*finish) = (M_uintptr)(end_ptr - data);

	if (data[0] != 0x02) {
		(*start) = 1;
		return -1;
	}

	(*start) = 0;

	fs_ptr = memchr(data + 1, 0x1c, (*finish) - 2);
	if (fs_ptr == NULL) {
		/* Umm, should never get here, invalid message */
		return 0;
	}

	(*fs) = (M_uintptr)(fs_ptr - data);

	return 1;
}

static void M_ProcessResponse(M_CONN *conn, M_uint64 identifier, const char *msg, size_t len)
{
	size_t  i;
	char *key = NULL, *val = NULL;
	int on_quote = 0; /* 0 = not on quote, 1 = on quote, 2 = quote done (wait till end of line
	                   *to store) */
	int all_whitespace = 0;
	int multi_quote = 0;
	int vallen, j;
	char *message;
	M_QUEUE *ptr = NULL;

	/* Must already be holding a lock on conn!! */
	ptr = M_FindQueue(conn, identifier);
	if (ptr == NULL) {
		return;
	}

	ptr->status = M_DONE;

	message = Ml_malloc(len + 1);
	memcpy(message, msg, len);
	message[len] = 0;

	/* See if it's comma delimited */
	ptr->iscommadelimited = 1;
	for (i = 0; i < len; i++) {
		if (message[i] == '=') {
			ptr->iscommadelimited = 0;
			break;
		}

		if ((i >= 25) || (message[i] == '\r') || (message[i] == '\n')) {
			break;
		}
	}

	if (ptr->iscommadelimited) {
		ptr->response = message;
		return;
	} else {
		/* Parse out key/value pairs from each line */
		key = message;
		for (i = 0; i < len; i++) {
			if ((on_quote != 1) && ((message[i] == '\r') || (message[i] == '\n'))) {
				message[i] = 0;
				if ((key != NULL) && (val != NULL)) {
					if (on_quote) {
						/* Strip leading and trailing quotes */
						val++;
						val[strlen(val) - 1] = 0;
						if (multi_quote) {
							UnescapeQuotes(val);
						}
					}

					M_add_resp_keyval(ptr, key, val);
					key = NULL;
					val = NULL;
					on_quote = 0;
					multi_quote = 0;
				}
			} else if ((on_quote == 0) && (val != NULL) && (message[i] == '"')) {
				/* If on value, and we hit a quote, lets see if it comes before
				 * any non-whitespace values */
				all_whitespace = 1;
				/* Current length of value up to this point */
				vallen = val - (message + i) - 1;
				for (j = 0; j < vallen; j++) {
					if ((val[j] != ' ') && (val[j] != '\t')) {
						all_whitespace = 0;
						break;
					}

					/* Ok, properly quoted, treat as such */
				}

				if (all_whitespace) {
					on_quote = 1;
					val = message + i; /* New val start is at quote */
				}
			} else if ((on_quote == 1) && (message[i] == '"')) {
				if (message[i + 1] == '"') {
					/* Quoted quotes in this value */
					multi_quote = 1;
					i++; /* Skip over next quote */
				} else {
					/* Found ending quote */
					on_quote = 2;

					/* Null terminate after ending quote, unless next
					 * character is a newline */
					if ((message[i + 1] != '\r') && (message[i + 1] != '\n')) {
						message[i + 1] = 0;
					}
				}
			} else if ((message[i] == '=') && (val == NULL)) {
				message[i] = 0;
				val = message + (i + 1);
			} else if (key == NULL) {
				key = message + i;
			}
		}

		if ((key != NULL) && (val != NULL)) {
			if (on_quote) {
				/* Strip leading and trailing quotes */
				val++;
				val[strlen(val) - 1] = 0;
				if (multi_quote) {
					UnescapeQuotes(val);
				}
			}

			M_add_resp_keyval(ptr, key, val);
		}
	}

	Ml_free(message);
}

static M_QUEUE *M_FindQueue(M_CONN *conn, M_uint64 identifier)
{
	M_QUEUE *ptr = NULL;

	ptr = (*conn)->queue;

	while (ptr != NULL) {
		if (
			(ptr->status == M_PENDING) &&
			(ptr->identifier == identifier)
	    ) {
			return ptr;
		}

		ptr = (M_QUEUE *)ptr->next;
		if (ptr == (*conn)->queue) {
			break;
		}
	}

	return NULL;
}

static void M_add_resp_keyval(M_QUEUE *ptr, const char *key, const char *val)
{
	if (ptr->resp_fields % 10 == 0) {
		ptr->parsed_resp = Ml_realloc(ptr->parsed_resp, (ptr->resp_fields + 10) * sizeof(M_TRANS));
	}

	ptr->parsed_resp[ptr->resp_fields].key = Ml_strdup(key);
	ptr->parsed_resp[ptr->resp_fields].value = Ml_strdup(val);
	ptr->resp_fields++;
}
