#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifndef _WIN32
#    include <unistd.h>
#else
#    include <winsock2.h>
#    include <windows.h>
#endif
#include <string.h>
#ifdef HAVE_STRINGS_H
#    include <strings.h>
#endif
#include <time.h>
#include <errno.h>
#ifdef HAVE_SYS_TYPES_H
#    include <sys/types.h>
#endif
#ifdef HAVE_SYS_TIME_H
#    include <sys/time.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#    include <sys/select.h>
#endif
#ifdef HAVE_SYS_FCNTL_H
#    include <sys/fcntl.h>
#endif
#ifdef HAVE_FCNTL_H
#    include <fcntl.h>
#endif

#include "proxy.h"
#include "md5.h"

/* Build-dependent includes and defines */
#include "base64.h"

#ifndef _WIN32
#    include <netdb.h>
#    include <sys/ioctl.h>
#else
#  define ioctl(a,b,c)       ioctlsocket(a,b,(u_long *)c)
#  ifndef _WIN32_WCE
#    define strcasecmp(a,b)    _stricmp(a,b)
#    define strncasecmp(a,b,c) _strnicmp(a,b,c)
#  endif
#endif

#include "http.h"

#include "snprintf.h"
#include "sockets.h"
#include "mstring.h"
#include "mstdlib.h"

#ifndef INADDR_NONE
#    define INADDR_NONE 0xffffffff
#endif

static void LIBMONETRA_proxy_authentication_digest(
	const char *algorithm,
	const char *username,
	const char *realm,
	const char *password,
	const char *nonce,
	int NonceCount,
	const char *cnonce,
	const char *qop,
	const char *method,
	const char *URL,
	const char *HEntity,
	char *digest_response
);

/** ghbn
 *  Description: GetHostByName DNS lookup.  Call native if possible.
 *  Result     : -1 on failure
 */
int LIBMONETRA_ghbn(const char *host, char **addr, char *error, int errlen)
{
#if defined (M_INTERNAL)
	return M_gethostbyname(host, addr, error, errlen);
#else
#    if !defined (HAVE_GETHOSTBYNAME_1)
	struct hostent hostbuf;
#    endif
	struct hostent *hp = NULL;
	size_t hstbuflen;
	char *tmphstbuf = NULL;
	int len;
	int ret;
#    if defined (HAVE_GETHOSTBYNAME_3)
	struct hostent_data hostbuf_data;
#    elif !defined (HAVE_GETHOSTBYNAME_1)
	int herr;
#    endif

	memset(error, 0, errlen);

	*addr = NULL;
	hstbuflen = 10000;
	tmphstbuf = Ml_malloc(hstbuflen);
#    if defined (HAVE_GETHOSTBYNAME_6)
	gethostbyname_r(host, &hostbuf, tmphstbuf, hstbuflen, &hp, &herr);
#    elif defined (HAVE_GETHOSTBYNAME_5)
	hp = gethostbyname_r(host, &hostbuf, tmphstbuf, hstbuflen, &herr);
#    elif defined (HAVE_GETHOSTBYNAME_3)
	if (gethostbyname_r(host, &hostbuf, &hostbuf_data) == 0) {
		hp = &hostbuf;
	} else {
		hp = NULL;
	}

#    elif defined (HAVE_GETHOSTBYNAME_1) || defined (_WIN32)
	/* Windows uses gethostbyname() */
	hp = gethostbyname(host);
#    else /* if defined (HAVE_GETHOSTBYNAME_6) */
#        error NO GETHOSTBYNAME DEFINITION SPECIFIED
#    endif /* if defined (HAVE_GETHOSTBYNAME_6) */
	if (hp == NULL) {
		Ml_free(tmphstbuf);
		ret = -1;
	} else {
		len = hp->h_length;
		*addr = Ml_malloc((len + 1) * sizeof(char));
		memset(*addr, 0, len + 1);
		memcpy(*addr, hp->h_addr_list[0], len);
		Ml_free(tmphstbuf);
		ret = len;
	}

	return ret;
#endif /* if defined (M_INTERNAL) */
}

/** checkread
 *  Description: Check to see if there are bytes available
 *  Result     : 0 on timeout, >0 if bytes available
 */
int LIBMONETRA_checkread(int fd, long timeout_s)
{
	fd_set readfs;
	struct timeval timeout;
	int res;

	FD_ZERO(&readfs);
	FD_SET(fd, &readfs);
	timeout.tv_sec = 0;
	timeout.tv_usec = timeout_s;

	res = select(fd + 1, &readfs, NULL, NULL, &timeout);
	if (res > 0) {
		if (FD_ISSET(fd, &readfs)) {
			return 1;
		}
	}

	return 0;
}

/** setnonblock
 *  Description: set the ip connection to non-blocking mode
 *  Result     : 0 on failure, 1 on success
 */
static int LIBMONETRA_setnonblock(int fd, int tf)
{
#ifdef O_NONBLOCK
	int flags;

	if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
		return 0;
	}

	if (tf) {
		flags |= O_NONBLOCK;
	} else {
		flags &= ~(O_NONBLOCK);
	}

	if (fcntl(fd, F_SETFL, flags) == -1) {
		return 0;
	}

#else /* ifdef O_NONBLOCK */
#    ifndef __WATCOMC__
	if (ioctl(fd, FIONBIO, &tf) == -1)
#    else
	if (ioctl(fd, FIONBIO, (char *)&tf) == -1)
#    endif
	{ return 0; }

#endif /* ifdef O_NONBLOCK */
	return 1;
}

#ifdef _WIN32
/** socket_strerror
 *  Description: Determine error on win32 sockets
 *  Result     : constant character string describing error
 */
static const char *LIBMONETRA_socket_strerror(int err)
{
	switch (err) {
		case WSANOTINITIALISED:
			return "NOT INITIALIZED";

		case WSAENETDOWN:
			return "NET DOWN";

		case WSAEFAULT:
			return "FAULT";

		case WSAENOTCONN:
			return "NOT CONNECTED";

		case WSAEINTR:
			return "CALL INTERRUPTED";

		case WSAEINPROGRESS:
			return "IN PROGRESS";

		case WSAENETRESET:
			return "CONNECTION RESET";

		case WSAENOTSOCK:
			return "NOT A SOCKET";

		case WSAEOPNOTSUPP:
			return "NOT SUPPORTED";

		case WSAESHUTDOWN:
			return "SHUTDOWN";

		case WSAEWOULDBLOCK:
			return "WOULD BLOCK";

		case WSAEMSGSIZE:
			return "MESSAGE SIZE";

		case WSAEINVAL:
			return "INVALID";

		case WSAECONNABORTED:
			return "CONNECTION ABORTED";

		case WSAETIMEDOUT:
			return "TIMED OUT";

		case WSAECONNRESET:
			return "CONNECTION RESET";

		case WSAECONNREFUSED:
			return "CONNECTION REFUSED";
	}

	return "UNKNOWN";
}

#endif /* ifdef _WIN32 */

/** connect_nonblock
 *  Description: Connect to address in a non-blocking fashion with timeout
 *  Returns    : 0 on failure, 1 on success
 */
int LIBMONETRA_connect_nonblock(
	int sockfd,
	const struct sockaddr *serv_addr,
	socklen_t addrlen,
	int maxtime,
	char *error,
	int errlen
) {
	int rc;

	socklen_t arglen;
	int err;
	fd_set fds;
	struct timeval timeout;
	time_t t, starttime;

	if (!maxtime) { /* Blocking mode, let OS decide when to stop */
		if (connect(sockfd, serv_addr, addrlen) == -1) {
#ifdef _WIN32
			err = WSAGetLastError();
			Ml_snprintf(error, errlen, "Connection Failed: %s", LIBMONETRA_socket_strerror(err));
#else
			err = errno;
			Ml_snprintf(error, errlen, "Connection Failed: %s", strerror(err));
#endif
			return 0;
		} else {
			return 1;
		}
	} else { /* Non-blocking mode, stop after maxtime (seconds) */
		time(&starttime);
		if (!LIBMONETRA_setnonblock(sockfd, 1)) { /* Set non-block mode */
			Ml_snprintf(error, errlen, "Could not set Non-Blocking mode");
			return 0;
		}

		rc  = connect(sockfd, serv_addr, addrlen);
#ifdef _WIN32
		err = WSAGetLastError();
#else
		err = errno;
#endif
		if (rc >= 0) {   /*  connect() completed immediately */
			LIBMONETRA_setnonblock(sockfd, 0);  /* Set blocking mode */
			return 1;
#ifdef _WIN32
		} else if (err != WSAEWOULDBLOCK) {
			Ml_snprintf(error, errlen, "Connection Failed: %d : %s", err, LIBMONETRA_socket_strerror(err));
			return 0;
#else
		} else if ((err != 0) && (err != EINPROGRESS)) { /* Error occurred */
			Ml_snprintf(error, errlen, "Connection Failed: %d : %s", err, strerror(err));
			return 0;
#endif
		} else {  /* Connect in progress */
			while (1) {
				time(&t);
				FD_ZERO(&fds);
				FD_SET((unsigned int)sockfd, &fds);
#ifdef _WIN32
				timeout.tv_sec  = (long)(maxtime - (t - starttime));
#else
				timeout.tv_sec  = (time_t)(maxtime - (t - starttime));
#endif
				timeout.tv_usec = 0;

				rc = select(sockfd+1, NULL, &fds,
#ifdef _WIN32
				            /* Windows uses exceptfds to return connect errors */
				            &fds,
#else
				            NULL,
#endif
				            &timeout);

				if (rc == 0) { /* Connection taking too long */
					time(&t);
					Ml_snprintf(error, errlen, "Connection Timed Out (%d secs)", (int)(t - starttime));
					return 0;
#ifdef EINTR
				} else if ((rc < 0) && (errno == EINTR)) {
					/* Interrupted by signal, we need a re-do */
#endif
				} else if (rc < 0) {  /* Error Occurred */
					Ml_snprintf(error, errlen, "Unknown select error occurred");
					return 0;
				} else {  /* rc > 0 .. good so far... */
					err    = 0;
					arglen = (socklen_t)sizeof(int);
#ifdef _WIN32
					if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (char *)&err, &arglen) < 0)
						err = WSAGetLastError();
#else
					if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &err, &arglen) < 0)
						err = errno;
#endif

					if (err != 0) {
						Ml_snprintf(error, errlen, "Connection Failed: %d: %s", err,
#ifdef _WIN32
						           LIBMONETRA_socket_strerror(err)
#else
						           strerror(err)
#endif
						);
						return 0;   /* Connect Failed */
					}

					/* Connection established! Let's break out of this loop! */
					break;
				}
			}
		}

		LIBMONETRA_setnonblock(sockfd, 0);  /* Set blocking mode */
		return 1;
	}

	return 0;  /* Should never get here... */
}

/** proxy_check_noproxy()
 *  Description: checks no-proxy list to see if on no-proxy list
 *  Returns    : 0 if on no-proxy list, 1 if not on list
 */
int LIBMONETRA_proxy_check_noproxy(proxy_config_st *config, const char *host)
{
	int i;

#if !defined (HAVE_IN_ADDR_T)
	unsigned long hostaddr, n_host, n_net, n_mask;
#else
	in_addr_t hostaddr, n_host, n_net, n_mask;
#endif
	int hostlen, suffixlen;
	if (host == NULL) {
		return 1;
	}

	hostaddr = inet_addr(host);
	n_host = htonl(hostaddr);
	hostlen = strlen(host);
	/* M_DebugLog(1, "Proxy: check host %s\r\n", host); */
	for (i = 0; i < config->num_noproxy; i++) {
		/* M_DebugLog(1, "Proxy: check %d\r\n", i); */
		if ((config->noproxy[i].host_suffix != NULL) && (hostaddr == INADDR_NONE)) {
			/* Only check if hostname is not an IP and we have a host suffix */
			/* M_DebugLog(1, "Proxy: check host suffix %s\r\n",
			 * config->noproxy[i].host_suffix); */
			suffixlen = strlen(config->noproxy[i].host_suffix);
			if (hostlen >= suffixlen) {
				/* Match the end of the hostname to the suffix */
				if (strcasecmp(host + (hostlen - suffixlen),
					    config->noproxy[i].host_suffix) == 0) {
					/* M_DebugLog(1, "Proxy: %s on no-proxy list #%d (%s)\r\n",
					 * host, i, config->noproxy[i].host_suffix); */
					/* On no-proxy list */
					return 0;
				}
			}
		} else if ((config->noproxy[i].host_suffix == NULL) && (hostaddr != INADDR_NONE)) {
			/* Types both match as IP addresses */
			/* M_DebugLog(1, "Proxy: check address\r\n"); */
			/* Convert to native byte order */
			n_net = htonl(config->noproxy[i].net);
			n_mask = htonl(config->noproxy[i].mask);

			if ((n_net & n_mask) == (n_host & n_mask)) {
				/* M_DebugLog(1, "Proxy: %s on no-proxy list #%d\r\n", host, i); */
				return 0;  /* On no-proxy list */
			}
		} else {
			/* M_DebugLog(1, "Proxy: not checking, address types differ\r\n"); */
		}
	}

	return 1;  /* Not on no-proxy list */
}

/** proxy_tunnel_check()
 *  Description: checks to see if the host/port is something we need to proxy.
 *               should be called prior to attempting establish a proxy connection.
 *  Returns    : 0 if no proxy needed, 1 if proxy needed
 */
int LIBMONETRA_proxy_tunnel_check(proxy_config_st *config, const char *host, unsigned short port)
{
	/* No proxy config, or proxying is disabled */
	if ((config == NULL) || (config->proxy_type == PROXY_TYPE_NONE)) {
		return 0;
	}

	/* Bad param */
	if (host == NULL) {
		return 0;
	}

	/* HTTP requests don't get tunneled, its standard HTTP, just
	 * connecting to a different host */
	if ((config->proxy_type == PROXY_TYPE_HTTP) && (port == 80)) {
		return 0;
	}

	return LIBMONETRA_proxy_check_noproxy(config, host);
}

/** proxy_trim_field()
 *  Description: Removes whitespace and quotes from fields
 *  Returns    : Ml_malloc()'d memory containging string, or NULL
 */
static char *LIBMONETRA_proxy_trim_field(const char *field)
{
	char *temp = NULL, *ptr = NULL, *end = NULL;
	char *ret = NULL;

	if (field == NULL) {
		return NULL;
	}

	temp = Ml_strdup(field);

	/* Chop off leading whitespace */
	ptr = temp;
	while (*ptr == ' ' || *ptr == '\t') {ptr++; }

	/* Chop off trailing whitespace */
	end = ptr + (strlen(ptr) - 1);
	while (*end == ' ' || *end == '\t' || *end == '\n' || *end == '\r') {
		*end = 0;
		end--;
	}

	/* Chop off surrounding quotes */
	if ((*ptr == *end) && ((*ptr == '\'') || (*ptr == '\"'))) {
		ptr++;
		*end = 0;
	}

	ret = Ml_strdup(ptr);
	Ml_free(temp);

	return ret;
}

/** proxy_http_parse_auth_fields()
 *  Description: Parses http fields, separated by commas, and stores
 *               known results to structure (Internal only)
 *  Returns    : 0 on unacceptible qop or algorithm, 1 on success
 */
static int LIBMONETRA_proxy_http_parse_auth_fields(const char *fields, proxy_http_auth_st *auth)
{
	int i, len;
	int j;
	char *qop_temp = NULL, *qop_temp2 = NULL;
	char **vars = NULL, *temp = NULL;
	int num_vars = 0;
	char *key = NULL, *val = NULL, *ptr = NULL;
	char *real_key = NULL;

	temp = Ml_strdup(fields);
	len = strlen(temp);

	/* Split at commas */
	vars = Ml_realloc(vars, (num_vars + 1) * sizeof(char *));
	vars[num_vars++] = temp;
	for (i = 0; i < len; i++) {
		if (temp[i] == ',') {
			temp[i] = 0;
			vars = Ml_realloc(vars, (num_vars + 1) * sizeof(char *));
			vars[num_vars++] = temp + (i + 1);
		}
	}

	for (i = 0; i < num_vars; i++) {
		key = vars[i];
		val = strchr(vars[i], '=');
		if (val == NULL) {
			continue;
		}

		val[0] = 0;
		val++;

		real_key = LIBMONETRA_proxy_trim_field(key);
		if (strcasecmp(real_key, "realm") == 0) {
			auth->realm = LIBMONETRA_proxy_trim_field(val);
		} else if (strcasecmp(real_key, "nonce") == 0) {
			auth->nonce = LIBMONETRA_proxy_trim_field(val);
		} else if (strcasecmp(real_key, "qop") == 0) {
			/* Could be a comma-delimited field. We only support 'auth', so
			 * lets try to see if 'auth' is specified */
			auth->qop = NULL;
			qop_temp = LIBMONETRA_proxy_trim_field(val);
			if (qop_temp != NULL) {
				len = strlen(qop_temp);
				ptr = qop_temp;
				for (j = 0; j < len; j++) {
					if (qop_temp[j] == ',') {
						qop_temp[j] = 0;
						qop_temp2 = LIBMONETRA_proxy_trim_field(ptr);
						if ((qop_temp2 != NULL) && (strcasecmp(qop_temp2, "auth") == 0)) {
							auth->qop = qop_temp2;
							break;
						}

						Ml_free(qop_temp2);
						ptr = qop_temp + (j + 1);
					}
				}

				if (auth->qop == NULL) {
					auth->qop = LIBMONETRA_proxy_trim_field(ptr);
				}

				Ml_free(qop_temp);
			}
		} else if (strcasecmp(real_key, "opaque") == 0) {
			auth->opaque = LIBMONETRA_proxy_trim_field(val);
		} else if (strcasecmp(real_key, "algorithm") == 0) {
			auth->algorithm = LIBMONETRA_proxy_trim_field(val);
		} else if (strcasecmp(real_key, "domain") == 0) {
			auth->domain = LIBMONETRA_proxy_trim_field(val);
		}

		Ml_free(real_key);
	}

	Ml_free(vars);
	Ml_free(temp);

	/* Unrecognized algorithm */
	if (
		(auth->algorithm != NULL)                      &&
		(strcasecmp(auth->algorithm, "md5") != 0)      &&
		(strcasecmp(auth->algorithm, "md5-sess") != 0)
	) {
		LIBMONETRA_proxy_http_clear_authenticate(auth);
		return 0;
	}

	/* Unrecognized qop */
	if ((auth->qop != NULL) && *(auth->qop) && (strcasecmp(auth->qop, "auth") != 0)) {
		/* We don't support auth-int (or anything else besides 'auth') */
		LIBMONETRA_proxy_http_clear_authenticate(auth);
		return 0;
	}

	return 1;
}

/** proxy_http_clear_authenticate()
 *  Description: Clears authentication structure
 *  Returns    : 0 on failure, 1 on success
 */
int LIBMONETRA_proxy_http_clear_authenticate(proxy_http_auth_st *auth)
{
	Ml_free(auth->realm);
	Ml_free(auth->nonce);
	Ml_free(auth->qop);
	Ml_free(auth->opaque);
	Ml_free(auth->algorithm);
	Ml_free(auth->domain);
	memset(auth, 0, sizeof(*auth));
	return 1;
}

/** proxy_http_parse_authenticate()
 *  Description: Parses the http Proxy-Authenticate header.
 *  Returns    : 0 on no acceptable methods, 1 on success
 */
int LIBMONETRA_proxy_http_parse_authenticate(M_http_t http, proxy_http_auth_st *auth)
{
	int hcnt = 0;
	char *temp_str, *pos;

	memset(auth, 0, sizeof(*auth));
	while ((temp_str = LIBMONETRA_http_get_header_str(http, "Proxy-Authenticate", hcnt++)) != NULL) {
		pos = temp_str;

		/* Skip leading whitespace */
		while (*pos == ' ' || *pos == '\t') {pos++; }

		if (strncasecmp(pos, "Basic", 5) == 0) {
			auth->type = PROXY_AUTH_BASIC;

			/* We really don't care about the realm or anything */
			Ml_free(temp_str);
			return 1;
		} else if (strncasecmp(pos, "Digest", 6) == 0) {
			auth->type = PROXY_AUTH_DIGEST;
			pos += 6;
			if (LIBMONETRA_proxy_http_parse_auth_fields(pos, auth)) {
				Ml_free(temp_str);
				return 1;
			}
		} else {
			/* Not a valid authentication type for us */
		}

		Ml_free(temp_str);
	}

	return 0;
}

/** socks5_callback_initial()
 *  Description: Determines when initial packet response is complete
 *  Returns    : -1 on error, 0 on not enough bytes, 1 on no authentication,
 *               2 if using authentication
 */
static int LIBMONETRA_socks5_callback_initial(const char *data, int datalen, char *error, int errlen)
{
	if (datalen == 2) {
		if (data[1] == 0x00) {
			return 1;
		} else if (data[1] == 0x02) {
			return 2;
		} else {
			Ml_snprintf(error, errlen, "Invalid authentication type");
			return -1;
		}
	} else if (datalen > 2) {
		Ml_snprintf(error, errlen, "Too many bytes to initial packet response");
		return -1;
	}

	return 0;  /* Need more bytes */
}

/** socks5_callback_auth()
 *  Description: Determines if response to socks5 auth request is valid
 *  Returns    : -1 on error, 0 on not enough bytes, 1 on success
 */
static int LIBMONETRA_socks5_callback_auth(const char *data, int datalen, char *error, int errlen)
{
	if (datalen == 2) {
		if (data[1] != 0x00) {
			Ml_snprintf(error, errlen, "Authentication failed with error 0x%02x", data[1]);
			return -1;
		} else {
			return 1;
		}
	} else if (datalen > 2) {
		Ml_snprintf(error, errlen, "Too many bytes to auth packet response");
		return -1;
	}

	return 0;  /* Need more bytes */
}

/** socks5_callback_connect()
 *  Description: Determines if response to socks5 connect request is
 *               valid, and if a failure occurs, why.
 *  Returns    : -1 on error, 0 on not enough bytes, 1 on success
 */
static int LIBMONETRA_socks5_callback_connect(const char *data, int datalen, char *error, int errlen)
{
	if (datalen <= 6) {
		/* Need more bytes */
	} else {
		/* If not successful, generate an error message */
		if (data[1] != 0x00) {
			switch (data[1]) {
				case 0x01:
					Ml_snprintf(error, errlen, "0x01: General SOCKS server failure");
					break;
				case 0x02:
					Ml_snprintf(error, errlen, "0x02: Connection not allowed by ruleset");
					break;
				case 0x03:
					Ml_snprintf(error, errlen, "0x03: Network unreachable");
					break;
				case 0x04:
					Ml_snprintf(error, errlen, "0x04: Host unreachable");
					break;
				case 0x05:
					Ml_snprintf(error, errlen, "0x05: Connection refused");
					break;
				case 0x06:
					Ml_snprintf(error, errlen, "0x06: TTL expired");
					break;
				case 0x07:
					Ml_snprintf(error, errlen, "0x07: Command not supported");
					break;
				case 0x08:
					Ml_snprintf(error, errlen, "0x08: Address type not supported");
					break;
				default:
					Ml_snprintf(error, errlen, "0x%02x: Unknown failure", data[1]);
					break;
			}

			return -1;
		}

		/* Lets validate we got the whole message here, and dont need to wait on more bytes
		 **/
		if (data[3] == 0x01) {
			/* IPv4, static size */

			if (datalen == 10) {
				return 1;  /* All bytes read */
			} else if (datalen > 10) {
				Ml_snprintf(error, errlen, "Too many bytes in Proxy connect response");
				return -1;
			} else {
				/* Need more bytes */
			}
		} else if (data[3] == 0x03) {
			/* Domain name */
			if (datalen == 7 + data[4]) {
				return 1;  /* All bytes read */
			} else if (datalen > 7 + data[4]) {
				Ml_snprintf(error, errlen, "Too many bytes in Proxy connect response");
				return -1;
			} else {
				/* Need more bytes */
			}
		} else {
			Ml_snprintf(error, errlen, "unrecognized address type in connect reply");
			return -1;
		}
	}

	return 0;  /* Need more bytes */
}

/** socks5_callback typedef
 *  Description: Determines if data is complete and processes.
 *  Returns    : -1 on error, 0 on need more data, > 0 user-definable
 *               on success
 *  Params:    data_input, data_input_len, error, error_len
 */
typedef int (*socks5_callback)(const char *, int, char *, int);

/** proxy_socks5_readresponse()
 *  Description: Waits on response for designated amount of time and calls callback
 *               to determine if response is complete. On error, this will close
 *               the connection (Internal Only)
 *  Returns    : 0 on error, > 0 on success
 */
static int LIBMONETRA_proxy_socks5_readresponse(
	int fd,
	char socks_ver,
	int timeout,
	socks5_callback callback,
	char *error,
	int errlen
) {
	char *data = NULL;
	int datalen = 0, bytes;
	char temp[1024];
	time_t starttime, t;
	int ret = 0;

	data = NULL; datalen = 0;
	time(&starttime);
	while (1) {
		if (LIBMONETRA_checkread(fd, 20000)) {
			memset(temp, 0, sizeof(temp));
			/* Only read 1 byte at a time here, otherwise when the connection
			 * is established, we might end up reading bytes as part of the
			 * newly opened tunnel, which could affect things negatively */
			bytes = M_read_socket("PROXY", 6, temp, 1, fd);
			if (bytes <= 0) {
				Ml_snprintf(error, errlen, "Lost connection to proxy");
				M_closesocket(fd, 0);
				Ml_free(data);
				return 0;
			}

			data = Ml_realloc(data, datalen + bytes + 1);
			memcpy(data + datalen, temp, bytes);
			datalen += bytes;
			data[datalen] = 0;

			if (datalen && (data[0] != socks_ver)) {
				Ml_snprintf(error, errlen, "Invalid socks5 response");
				M_closesocket(fd, 1);
				Ml_free(data);
				return 0;
			}

			ret = (*callback)(data, datalen, error, errlen);
			if (ret < 0) {
				/* Critical error, close socket */
				Ml_free(data);
				M_closesocket(fd, 1);
				return 0;
			} else if (ret > 0) {
				/* Success */
				Ml_free(data);
				return ret;
			} else {
				/* Wait on more bytes */
			}
		}

		time(&t);
		if (t - starttime > timeout) {
			break;
		}
	}

	Ml_free(data); data = NULL;
	Ml_snprintf(error, errlen, "Timeout Proxy Negotiation (%d secs)", (int)(t - starttime));
	M_closesocket(fd, 1);

	return -1;
}

/** proxy_tunnel_connect_socks5()
 *  Description: Establishes the socks5 tunnel to host and port
 *               specified using authentication if needed. (RFC1928 & 1929)
 *               Internal Function Only.
 *  Returns    : -1 on error, file descriptor for socket connection on success
 */
static int LIBMONETRA_proxy_tunnel_connect_socks5(
	proxy_config_st *config,
	const char *host,
	unsigned short port,
	char *error,
	int errlen
) {
	struct sockaddr_in peer;
	int fd = -1;
	int len;
	char *addr = NULL;
	char temp[1024];
	int bytes;
	int do_authentication = 0;
	unsigned short myport;
	int ret = 0;
	char temperr[255];

#ifdef HAVE_IN_ADDR_T
	in_addr_t myaddr;
#else
	unsigned long myaddr;
#endif

	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_port = htons(config->proxy_port);

	len = LIBMONETRA_ghbn(config->proxy_host, &addr, temperr, sizeof(temperr));
	if (len != -1) {
		memcpy((char *)&peer.sin_addr, addr, len);
		Ml_free(addr);
	} else {
		Ml_free(addr);
		Ml_snprintf(error, errlen, "DNS Lookup Failed: %s", temperr);
		return -1;
	}

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		Ml_snprintf(error, errlen, "socket() failed: %s", strerror(errno));
		return -1;
	}

	if (
		!LIBMONETRA_connect_nonblock(
			fd,
			(struct sockaddr *)&peer,
			sizeof(peer),
			config->proxy_timeout,
			error,
			errlen
		)
	) {
		M_closesocket(fd, 0);
		return -1;
	}

	/* Initial SOCKS5 connection packet */

	temp[0] = 0x05; /* Version 5 */
	temp[1] = 0x02; /* 2 methods specified */
	temp[2] = 0x00; /* 0x00 = No authentication required */
	temp[3] = 0x02; /* 0x02 = Username/Password authentication */
	if (M_write_socket("PROXY", 6, temp, 4, fd) != 4) {
		Ml_snprintf(error, errlen, "Failed to write data to proxy");
		M_closesocket(fd, 0);
		return -1;
	}

	ret = LIBMONETRA_proxy_socks5_readresponse(
		fd,
		0x05,
		config->proxy_timeout,
		LIBMONETRA_socks5_callback_initial,
		error,
		errlen
	);
	if (ret <= 0) {
		return -1;
	} else if (ret == 1) {
		do_authentication = 0;
	} else if (ret == 2) {
		do_authentication = 1;
	}

	if (do_authentication) {
		/* Basic username/password authentication only */
		len = 0;
		temp[len++] = 0x01; /* Authentication version is v1 */

		bytes = strlen(M_Safe(config->authuser));
		temp[len++] = (char)(bytes & 0x000000FF);
		memcpy(temp + len, M_Safe(config->authuser), bytes); /* Copy auth user */
		len += bytes;

		bytes = strlen(M_Safe(config->authpass));
		temp[len++] = (char)(bytes & 0x000000FF);
		memcpy(temp + len, M_Safe(config->authpass), bytes); /* Copy auth user */
		len += bytes;

		if (M_write_socket("PROXY", 6, temp, len, fd) != len) {
			Ml_snprintf(error, errlen, "Failed to write data to proxy");
			M_closesocket(fd, 0);
			return -1;
		}

		ret = LIBMONETRA_proxy_socks5_readresponse(
			fd,
			0x01,
			config->proxy_timeout,
			LIBMONETRA_socks5_callback_auth,
			error,
			errlen
		);
		if (ret <= 0) {
			return -1;
		}
	}

	/* Send request to connect to desired address */
	len = 0;
	temp[len++] = 0x05; /* Socks version 5 */
	temp[len++] = 0x01; /* 0x01 == CONNECT */
	temp[len++] = 0x00; /* RESERVED */

	myaddr = inet_addr(host);
	if (myaddr != INADDR_NONE) {
		temp[len++] = 0x01;  /* Type is IPv4 */
		memcpy(temp + len, (char *)&myaddr, 4);
		len += 4;
	} else {
		temp[len++] = 0x03;  /* Type is DomainName */
		temp[len++] = (char)(strlen(host) & 0x000000FF); /* One byte length of domain */
		memcpy(temp + len, host, strlen(host)); /* Copy domain name */
		len += strlen(host);
	}

	myport = ntohs(port);
	memcpy(temp + len, &myport, 2); /* Copy port in network byte order */
	len += 2;

	if (M_write_socket("PROXY", 6, temp, len, fd) != len) {
		Ml_snprintf(error, errlen, "Failed to write data to proxy");
		M_closesocket(fd, 0);
		return -1;
	}

	ret = LIBMONETRA_proxy_socks5_readresponse(
		fd,
		0x05,
		config->proxy_timeout,
		LIBMONETRA_socks5_callback_connect,
		error,
		errlen
	);
	if (ret <= 0) {
		return -1;
	}

	return fd;
}

/** proxy_tunnel_connect_http()
 *  Description: Establishes the http tunnel to host and port specified,
 *               using authentication if needed.  Internal function only.
 *  Returns    : -1 on error, file descriptor for socket connection on success
 */
static int LIBMONETRA_proxy_tunnel_connect_http(
	proxy_config_st *config,
	const char *host,
	unsigned short port,
	char *error,
	int errlen
) {
	M_http_t http;
	int fd = -1;
	int len;
	char *addr = NULL;
	struct sockaddr_in peer;
	char *data;
	char temp[1024];
	int bytes;
	int datalen;
	time_t starttime, t;
	int http_code;
	char *code_msg = NULL;
	char *temp_str = NULL;
	int tries;
	proxy_http_auth_st auth;
	char temperr[255];

	memset(&auth, 0, sizeof(auth));

	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_port = htons(config->proxy_port);

	len = LIBMONETRA_ghbn(config->proxy_host, &addr, temperr, sizeof(temperr));
	if (len != -1) {
		memcpy((char *)&peer.sin_addr, addr, len);
		Ml_free(addr);
	} else {
		Ml_free(addr);
		Ml_snprintf(error, errlen, "DNS Lookup Failed: %s", temperr);
		return -1;
	}

	for (tries = 0; tries < 2; tries++) {
		if (fd < 0) {
			fd = socket(AF_INET, SOCK_STREAM, 0);
			if (fd < 0) {
				Ml_snprintf(error, errlen, "socket() failed: %s", strerror(errno));
				LIBMONETRA_proxy_http_clear_authenticate(&auth);
				return -1;
			}

			if (!LIBMONETRA_connect_nonblock(fd, (struct sockaddr *)&peer, sizeof(peer),
				    config->proxy_timeout, error, errlen)) {
				M_closesocket(fd, 0);
				LIBMONETRA_proxy_http_clear_authenticate(&auth);
				return -1;
			}
		}

		http = LIBMONETRA_http_new();
		LIBMONETRA_http_set_version(http, MHTTP_HTTP_VER_1_1);
		if (!LIBMONETRA_http_set_hosttype(http, MHTTP_CONNECT, NULL, host, port)) {
			LIBMONETRA_http_free(http);
			Ml_snprintf(error, errlen, "Failed to set CONNECT method");
			LIBMONETRA_proxy_http_clear_authenticate(&auth);
			return -1;
		}

		switch (auth.type) {
			case PROXY_AUTH_NONE:
				break;
			case PROXY_AUTH_BASIC:
			case PROXY_AUTH_DIGEST:
				temp_str = LIBMONETRA_proxy_authorization_response_header(
					&auth,
					config->authuser,
					config->authpass,
					"CONNECT"
				);
				LIBMONETRA_http_set_header(http, "Proxy-Authorization", temp_str, 1);
				Ml_free(temp_str);
				break;
		}

		/* Clear authentication data here */
		LIBMONETRA_proxy_http_clear_authenticate(&auth);

		data = LIBMONETRA_http_gen_request(http);
		datalen = LIBMONETRA_http_get_request_len(http);
		LIBMONETRA_http_free(http);

		if (M_write_socket("PROXY", 6, data, datalen, fd) != datalen) {
			Ml_snprintf(error, errlen, "Failed to write data to proxy");
			Ml_free(data);
			M_closesocket(fd, 0);
			LIBMONETRA_proxy_http_clear_authenticate(&auth);
			return -1;
		}

		Ml_free(data);
		datalen = 0;
		data = NULL;
		time(&starttime);

		while (1) {
			if (LIBMONETRA_checkread(fd, 20000)) {
				memset(temp, 0, sizeof(temp));
				/* Only read 1 byte at a time here, otherwise when the connection
				 * is established, we might end up reading bytes as part of the
				 * newly opened tunnel, which could affect things negatively */
				bytes = M_read_socket("PROXY", 6, temp, 1, fd);
				if (bytes <= 0) {
					Ml_snprintf(error, errlen, "Lost connection to proxy");
					M_closesocket(fd, 0);
					Ml_free(data);
					LIBMONETRA_proxy_http_clear_authenticate(&auth);
					return -1;
				}

				data = Ml_realloc(data, datalen + bytes + 1);
				memcpy(data + datalen, temp, bytes);
				datalen += bytes;
				data[datalen] = 0;

				http = LIBMONETRA_http_parse(data, datalen, 0, error, errlen);
				if (http != NULL) {
					Ml_free(data); data = NULL;
					break;
				}
			}

			time(&t);
			if (t - starttime > config->proxy_timeout) {
				Ml_snprintf(error, errlen, "Timeout Proxy Negotiation (%d secs)", (int)(t - starttime));
				M_closesocket(fd, 1);
				Ml_free(data);
				LIBMONETRA_proxy_http_clear_authenticate(&auth);
				return -1;
			}
		}

		http_code = LIBMONETRA_http_get_code(http, &code_msg);
		if (http_code == 200) {
			/* Connection successful */
			LIBMONETRA_http_free(http);
			Ml_free(code_msg);
			LIBMONETRA_proxy_http_clear_authenticate(&auth);
			return fd;
		} else if (http_code == 407) {
			/* Proxy authentication required */
			Ml_free(code_msg);

			/* See if we should force a disconnect */
			temp_str = LIBMONETRA_http_get_header_str(http, "Connection", 0);
			if (temp_str == NULL) {
				temp_str = LIBMONETRA_http_get_header_str(http, "Proxy-Connection", 0);
			}

			if ((temp_str != NULL) && (strcasecmp(temp_str, "Close") == 0)) {
				M_closesocket(fd, 0);
				fd = -1;
			}

			Ml_free(temp_str);

			if (!LIBMONETRA_proxy_http_parse_authenticate(http, &auth)) {
				LIBMONETRA_http_free(http);
				M_closesocket(fd, 1);
				Ml_snprintf(error, errlen, "No acceptible proxy authentication method specified");
				LIBMONETRA_proxy_http_clear_authenticate(&auth);
				return -1;
			}

			/*
			 * M_DebugLog(1,	"Proxy Settings:\r\n"
			 *              "\ttype=%d\r\n"
			 *              "\trealm=%s\r\n"
			 *              "\tnonce=%s\r\n"
			 *              "\tqop=%s\r\n"
			 *              "\topaque=%s\r\n"
			 *              "\talgorithm=%s\r\n"
			 *              "\tdomain=%s\r\n",
			 *              auth.type, auth.realm, auth.nonce, auth.qop, auth.opaque,
			 *              auth.algorithm, auth.domain);
			 */
		} else {
			/* Unrecognized error code */
			LIBMONETRA_http_free(http);
			M_closesocket(fd, 1);
			Ml_snprintf(error, errlen, "Proxy authentication failed %d: %s", http_code, code_msg);
			Ml_free(code_msg);
			LIBMONETRA_proxy_http_clear_authenticate(&auth);
			return -1;
		}
	}

	/* if we're here, it didn't succeed bad */
	M_closesocket(fd, 1);
	Ml_snprintf(error, errlen, "Proxy authentication failed (max tries)");
	LIBMONETRA_proxy_http_clear_authenticate(&auth);
	return -1;
}

/** proxy_tunnel_connect()
 *  Description: Establishes the proxy tunnel to the host and port specified,
 *               using authentication if needed
 *  Returns    : -1 on error, file descriptor for socket connection on success
 */
int LIBMONETRA_proxy_tunnel_connect(
	proxy_config_st *config,
	const char *host,
	unsigned short port,
	char *error,
	int errlen
) {
	int fd;

	if ((config == NULL) || (config->proxy_type == PROXY_TYPE_NONE)) {
		Ml_snprintf(error, errlen, "Not configured");
		return -1;
	}

	if (config->proxy_type == PROXY_TYPE_HTTP) {
		fd = LIBMONETRA_proxy_tunnel_connect_http(config, host, port, error, errlen);
	} else if (config->proxy_type == PROXY_TYPE_SOCKS5) {
		fd = LIBMONETRA_proxy_tunnel_connect_socks5(config, host, port, error, errlen);
	} else {
		/* Not a known proxy type */
		Ml_snprintf(error, errlen, "Invalid proxy type");
		fd = -1;
	}

	return fd;
}

/** proxy_config_init()
 *  Description: Initializes proxy default settings
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_init(proxy_config_st *config)
{
	if (config == NULL) {
		return 0;
	}

	memset(config, 0, sizeof(*config));
	config->proxy_timeout = 10; /* Default to 10 second timeout */
	return 1;
}

/** proxy_config_host()
 *  Description: Adds host/port/type to proxy config
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_host(proxy_config_st *config, const char *host, unsigned short port, int proxy_type)
{
	if (config == NULL) {
		return 0;
	}

	config->proxy_host = Ml_strdup(host);
	config->proxy_port = port;
	config->proxy_type = proxy_type;
	return 1;
}

/** proxy_config_timeout()
 *  Description: Sets proxy negotiation timeout in seconds
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_timeout(proxy_config_st *config, int timeout)
{
	if (config == NULL) {
		return 0;
	}

	config->proxy_timeout = timeout;
	return 1;
}

/** proxy_config_user()
 *  Description: Sets proxy authentication parameters (username/password)
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_user(proxy_config_st *config, const char *username, const char *password)
{
	if (
		(config == NULL)   ||
		(username == NULL) ||
		(password == NULL) ||
		*username == '\0'  ||
		*password == '\0'
	) {
		return 0;
	}

	config->authuser = Ml_strdup(username);
	config->authpass = Ml_strdup(password);
	return 1;
}

/** proxy_addr_is_ip()
 *  Description: Determine if string is an IP address (internal only)
 *  Returns    : 0 if not IP, 1 if IP
 */
static int LIBMONETRA_proxy_addr_is_ip(const char *addr)
{
	int i, len;
	int dot_cnt = 0, num_cnt = 0;
	int has_slash = 0;

	if (addr == NULL) {
		return 0;
	}

	len = strlen(addr);
	for (i = 0; i < len; i++) {
		if (((addr[i] < '0') || (addr[i] > '9')) && (addr[i] != '.') && (addr[i] != '/')) {
			/* Must be numeric with dots, or have a mask */
			return 0;
		}

		if (addr[i] == '.') {
			dot_cnt++;
			num_cnt = 0;
		} else if (addr[i] == '/') {
			/* can only contain 1 slash */
			if (has_slash) {
				return 0;
			}

			/* Must have 3 dots */
			if (dot_cnt != 3) {
				return 0;
			}

			has_slash = 1;
			dot_cnt = 0;
			num_cnt = 0;
		} else {
			num_cnt++;
		}

		/* Too many digits */
		if (num_cnt > 3) {
			return 0;
		}
	}

	/* Must have 3 dots */
	if (!has_slash && (dot_cnt != 3)) {
		return 0;
	}

	/* Netmask must have either 0 dots, or 3 */
	if (has_slash && (dot_cnt > 0) && (dot_cnt != 3)) {
		return 0;
	}

	return 1;
}

/** proxy_addr_is_host
 *  Description: Determines if characters are valid for hostnames (internal only)
 *  Returns    : 0 if not proper hostname, 1 if proper hostname
 */
static int LIBMONETRA_proxy_addr_is_host(const char *addr)
{
	int i, len;

	if (addr == NULL) {
		return 0;
	}

	len = strlen(addr);
	for (i = 0; i < len; i++) {
		if (
			((addr[i] >= 'a') && (addr[i] <= 'z')) ||
			((addr[i] >= 'A') && (addr[i] <= 'Z')) ||
			((addr[i] >= '0') && (addr[i] <= '9')) ||
			(addr[i] == '.')                       ||
			(addr[i] == '-')                       ||
		    (addr[i] == '_')
		) {
			/* Is a valid character for a hostname */
		} else {
			return 0;
		}
	}

	return 1;
}

/** proxy_config_noproxy_addr()
 *  Description: Parse single address type and fill in no_proxy_st (internal only)
 *  Returns    : 0 on error, 1 on success
 */
static int LIBMONETRA_proxy_config_noproxy_addr(
	no_proxy_st **noproxy,
	int *num_noproxy,
	const char *address,
	char *error,
	int errlen
) {
	char *temp = NULL, *ptr = NULL;
	no_proxy_st mynoproxy;

	if (address == NULL || *address == '\0') {
		Ml_snprintf(error, errlen, "Missing address, invalid format");
		return 0;
	}

	memset(&mynoproxy, 0, sizeof(mynoproxy));

	temp = LIBMONETRA_proxy_trim_field(address);

	if (LIBMONETRA_proxy_addr_is_ip(temp)) {
		ptr = strchr(temp, '/');
		if (ptr != NULL) {
			ptr[0] = 0;
			ptr++;
		}

		mynoproxy.net = inet_addr(temp);
		mynoproxy.mask = 0xFFFFFFFF;
		if (ptr) {
			if (LIBMONETRA_proxy_addr_is_ip(ptr)) {
				mynoproxy.mask = inet_addr(ptr);
			} else {
				if ((atoi(ptr) <= 0) && (atoi(ptr) > 32)) {
					Ml_snprintf(error, errlen, "Bad Mask: %s", ptr);
					Ml_free(temp);
					return 0;
				}

				mynoproxy.mask <<= 32 - atoi(ptr);
				mynoproxy.mask = ntohl(mynoproxy.mask);
			}
		} else {}
	} else {
		if (!LIBMONETRA_proxy_addr_is_host(temp)) {
			Ml_snprintf(error, errlen, "Bad Host: %s", temp);
			Ml_free(temp);
			return 0;
		}

		mynoproxy.host_suffix = Ml_strdup(temp);
	}

	Ml_free(temp);

	*noproxy = Ml_realloc(*noproxy, ((*num_noproxy) + 1) * sizeof(**noproxy));
	memcpy(&((*noproxy)[(*num_noproxy)++]), &mynoproxy, sizeof(mynoproxy));

	return 1;
}

/** proxy_config_noproxy()
 *  Description: Comma delimited list of addresses which are not proxied
 *               Valid formatting:
 *                  .mydomain.com              (domain suffix)
 *                  192.168.1.1                (single ip)
 *                  192.168.1.0/255.255.255.0  (ip with subnet mask)
 *                  192.168.1.0/24             (ip with subnet mask)
 *               Even on success, error may be populated with the last warning.
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_noproxy(proxy_config_st *config, const char *noproxy, char *error, int errlen)
{
	char **addresses = NULL, *temp = NULL;
	int num_addresses = 0, i, len;

	if (
		(config == NULL)  ||
		(noproxy == NULL) ||
		*noproxy == '\0'
	) {
		Ml_snprintf(error, errlen, "parameters not set");
		return 0;
	}

	temp = Ml_strdup(noproxy);
	len = strlen(temp);
	addresses = (char **)Ml_realloc(addresses, (num_addresses + 1) * sizeof(char *));
	addresses[num_addresses++] = temp;

	for (i = 0; i < len; i++) {
		if (temp[i] == ',') {
			temp[i] = 0;
			addresses = (char **)Ml_realloc(addresses, (num_addresses + 1) * sizeof(char *));
			addresses[num_addresses++] = temp + (i + 1);
		}
	}

	config->noproxy = NULL;
	config->num_noproxy = 0;

	memset(error, 0, errlen);

	for (i = 0; i < num_addresses; i++) {
		if (
			!LIBMONETRA_proxy_config_noproxy_addr(
				&config->noproxy,
				&config->num_noproxy,
				addresses[i],
				error,
				errlen
			)
		) {
			/* Do nothing, error should be set though */
		}
	}

	Ml_free(addresses);
	Ml_free(temp);
	return 1;
}

/** proxy_config_free()
 *  Description: Free proxy configuration
 *  Returns    : 0 on error, 1 on success
 */
int LIBMONETRA_proxy_config_free(proxy_config_st *config)
{
	int i;

	if (config == NULL) {
		return 0;
	}

	for (i = 0; i < config->num_noproxy; i++) {
		if (config->noproxy[i].host_suffix) {
			Ml_free(config->noproxy[i].host_suffix);
		}
	}

	Ml_free(config->noproxy);

	Ml_free(config->authuser);
	Ml_free(config->authpass);
	Ml_free(config->proxy_host);

	LIBMONETRA_proxy_config_init(config);
	return 1;
}

/** proxy_bin2hex
 *  Description: convert binary data to hex (ascii) form, must provide
 *               a buffer of bin_len * 2 + 1 (Internal Only)
 *  Returns    : nothing
 */
static void LIBMONETRA_proxy_bin2hex(const unsigned char *bin, int bin_len, char *hex)
{
	unsigned short i;

	for (i = 0; i < bin_len; i++) {
		hex[i * 2] = ("0123456789abcdef")[(bin[i] >> 4) & 0xf];
		hex[i * 2 + 1] = ("0123456789abcdef")[bin[i] & 0xf];
	}

	hex[bin_len * 2] = '\0';
}

/** proxy_authentication_digest
 *  Description: Generates a 32 character digest for use as response in
 *               proxy authentication (Internal Only)
 *  Returns    : nothing
 */
static void LIBMONETRA_proxy_authentication_digest(
        const char *algorithm,          /* algorithm from server */
        const char *username,           /* username to authenticate as */
        const char *realm,              /* realm from server */
        const char *password,           /* plain text password to authenticate as */
        const char *nonce,              /* nonce from server (unique token generated per request) */
        int NonceCount,                 /* specified only if qop directive sent. client-side
                                         * incremented
                                         * for each request count for a given nonce response */
        const char *cnonce,             /* specified only if qop directive sent. client nonce,
                                         *generated */
        const char *qop,                /* Quality of Protection, "", "auth", or "auth-int" */
        const char *method,             /* method from the request */
        const char *URL,                /* Requested URL */
        const char *HEntity,            /* H(entity body) if qop="auth-int"  */
        char *digest_response           /* Response Digest, must be 32bytes + 1 (null-term) in
                                         *length, hex */
        )
{
	md5_state_t Md5Ctx;
	unsigned char HA1[16], HA2[16], RespHash[16];
	char HA1Hex[33], HA2Hex[33];

	char temp[1024];

	/* Calculate H(A1) */
	md5_init(&Md5Ctx);

	Ml_snprintf(temp, sizeof(temp), "%s:%s:%s", username, realm, password);
	md5_append(&Md5Ctx, (unsigned char *)temp, strlen(temp));
	md5_finish(&Md5Ctx, HA1);

	if (strcasecmp(algorithm, "md5-sess") == 0) {
		md5_init(&Md5Ctx);
		md5_append(&Md5Ctx, HA1, 16);

		Ml_snprintf(temp, sizeof(temp), ":%s:%s", nonce, cnonce);
		md5_append(&Md5Ctx, (unsigned char *)temp, strlen(temp));

		md5_finish(&Md5Ctx, HA1);
	}

	LIBMONETRA_proxy_bin2hex(HA1, 16, HA1Hex);

	/* Calculate H(A2) */
	md5_init(&Md5Ctx);
	Ml_snprintf(temp, sizeof(temp), "%s:%s", method, URL ? URL : "/");
	md5_append(&Md5Ctx, (unsigned char *)temp, strlen(temp));

	if (strcasecmp(qop, "auth-int") == 0) {
		md5_append(&Md5Ctx, (const unsigned char *)":", 1);
		md5_append(&Md5Ctx, (const unsigned char *)HEntity, 32);
	}

	md5_finish(&Md5Ctx, HA2);
	LIBMONETRA_proxy_bin2hex(HA2, 16, HA2Hex);

	/* Calculate response */
	md5_init(&Md5Ctx);
	md5_append(&Md5Ctx, (unsigned char *)HA1Hex, 32);

	Ml_snprintf(temp, sizeof(temp), ":%s:", nonce);
	md5_append(&Md5Ctx, (unsigned char *)temp, strlen(temp));

	if (*qop) {
		Ml_snprintf(temp, sizeof(temp), "%08d:%s:%s:", NonceCount, cnonce, qop);
		md5_append(&Md5Ctx, (unsigned char *)temp, strlen(temp));
	}

	md5_append(&Md5Ctx, (unsigned char *)HA2Hex, 32);
	md5_finish(&Md5Ctx, RespHash);
	LIBMONETRA_proxy_bin2hex(RespHash, 16, digest_response);
}

#define PROXY_ALLOC 1024
#define PROXY_APPEND(dest, destlen, destalloc, src, srclen) \
        do { \
		while (destlen + srclen + 1 >= destalloc) { \
			dest = Ml_realloc(dest, destalloc + PROXY_ALLOC); \
			memset(dest + destalloc, 0, PROXY_ALLOC); \
			destalloc += PROXY_ALLOC; \
		} \
		if (srclen) {memcpy(dest + destlen, src, srclen); } \
		destlen += srclen; \
	} while (0)

/** proxy_authorization_response_header
 *  Description: Generates a header to return to a proxy for digest authorization
 *  Returns    : Ml_malloc()'d header to pass on, NULL on error
 */
char *LIBMONETRA_proxy_authorization_response_header(
	proxy_http_auth_st *auth,
	const char *username,
	const char *password,
	const char *http_method
) {
	char response_digest[33];
	char HEntity[33];
	char temp[1024];
	char *ret = NULL, *temp_str = NULL;
	int retlen = 0, retalloc = 0, len = 0, i;

	memset(response_digest, 0, sizeof(response_digest));

	/* If we ever want to support auth-int, we need to support this */
	memset(HEntity, 0, sizeof(HEntity));

	if (auth->type == PROXY_AUTH_BASIC) {
		Ml_snprintf(temp, sizeof(temp), "%s:%s", username ? username : "", password ? password : "");
		len = strlen(temp);
		temp_str = Ml_malloc(B64_TARGSIZE(len) + 1);
		memset(temp_str, 0, B64_TARGSIZE(len) + 1);
		base64_encode(temp, len, temp_str, B64_TARGSIZE(len));

		/* strip trailing return chars */
		len = strlen(temp_str);
		for (i = len - 1; i >= 0; i--) {
			if ((temp_str[i] != '\r') && (temp_str[i] != '\n')) {
				break;
			} else {
				temp_str[i] = 0;
			}
		}

		PROXY_APPEND(ret, retlen, retalloc, "Basic ", 6);
		len = strlen(temp_str);
		PROXY_APPEND(ret, retlen, retalloc, temp_str, len);

		Ml_free(temp_str); temp_str = NULL;
		return ret;
	}

	if (auth->type != PROXY_AUTH_DIGEST) {
		return NULL;
	}

	LIBMONETRA_proxy_authentication_digest(
		auth->algorithm,
		username,
		auth->realm,
		password,
		auth->nonce,
		1,
		"0a4f113b",
		auth->qop,
		http_method,
		auth->domain,
		HEntity,
		response_digest
	);

	PROXY_APPEND(ret, retlen, retalloc, "Digest ", 7);

	/* Username */
	Ml_snprintf(temp, sizeof(temp), "username=\"%s\"", username ? username : "");
	len = strlen(temp);
	PROXY_APPEND(ret, retlen, retalloc, temp, len);

	/* Realm */
	Ml_snprintf(temp, sizeof(temp), ", realm=\"%s\"", auth->realm ? auth->realm : "");
	len = strlen(temp);
	PROXY_APPEND(ret, retlen, retalloc, temp, len);

	/* Nonce */
	if (auth->nonce != NULL) {
		Ml_snprintf(temp, sizeof(temp), ", nonce=\"%s\"", auth->nonce);
		len = strlen(temp);
		PROXY_APPEND(ret, retlen, retalloc, temp, len);
	}

	/* URI */
	Ml_snprintf(temp, sizeof(temp), ", uri=\"%s\"", auth->domain ? auth->domain : "/");
	len = strlen(temp);
	PROXY_APPEND(ret, retlen, retalloc, temp, len);

	/* Algorithm */
	if (auth->algorithm != NULL) {
		Ml_snprintf(temp, sizeof(temp), ", algorithm=\"%s\"", auth->algorithm);
		len = strlen(temp);
		PROXY_APPEND(ret, retlen, retalloc, temp, len);
	}

	/* Parameters for qop */
	if (auth->qop) {
		Ml_snprintf(temp, sizeof(temp), ", qop=\"%s\", nc=%08x, cnonce=\"%s\"", auth->qop, 1, "0a4f113b");
		len = strlen(temp);
		PROXY_APPEND(ret, retlen, retalloc, temp, len);
	}

	/* Response */
	Ml_snprintf(temp, sizeof(temp), ", response=\"%s\"", response_digest);
	len = strlen(temp);
	PROXY_APPEND(ret, retlen, retalloc, temp, len);

	/* Opaque */
	if (auth->opaque != NULL) {
		Ml_snprintf(temp, sizeof(temp), ", opaque=\"%s\"", auth->opaque);
		len = strlen(temp);
		PROXY_APPEND(ret, retlen, retalloc, temp, len);
	}

	return ret;
}
