/*
 * Copyright (c) 1983, 1995, 1996 Eric P. Allman
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *	  must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "main.h"
#include "mstdlib.h"
#include "buf.h"

#ifdef _WIN32
#  include <windows.h>
#  include <io.h>
#  ifndef _WIN32_WCE
#    define write(a,b,c) _write(a,b,c)
#  endif
#  ifdef _WIN64
#    define ssize_t __int64
#  else
#    define ssize_t long
#  endif
#endif

#include <stdarg.h>

#ifndef _WIN32
#    include <sys/ioctl.h>
#    include <sys/param.h>
#endif

/*
**	SNPRINTF, VSNPRINT -- counted versions of printf
**
**	These versions have been grabbed off the net.  They have been
**	cleaned up to compile properly and support for .precision and
**	%lx has been added.
*/

/**************************************************************
 * Original:
 * Patrick Powell Tue Apr 11 09:48:21 PDT 1995
 * A bombproof version of doprnt (dopr) included.
 * Sigh.  This sort of thing is always nasty do deal with.	Note that
 * the version here does not include floating point. (now it does ... tgl)
 *
 * snprintf() is used instead of sprintf() as it does limit checks
 * for string length.  This covers a nasty loophole.
 *
 * The other functions are there to prevent NULL pointers from
 * causing nast effects.
 **************************************************************/

/*static char _id[] = "$Id: snprintf.c 28195 2012-10-02 20:26:09Z bhouse $";*/

enum snprintf_output_type {
	VSNPRINTF_OUT_NONE  = 0,
	VSNPRINTF_OUT_BUF   = 1,
	VSNPRINTF_OUT_FD    = 2,
	VSNPRINTF_OUT_FP    = 3,
	VSNPRINTF_OUT_ALLOC = 4
};

#define VSNPRINTF_CACHE_SIZE 512

typedef struct snprintf_state_t {
	enum snprintf_output_type out_type;

	/* VSNPRINTF_OUT_FD */
	int                       fd;

	/* VSNPRINTF_OUT_FP */
	FILE                     *fp;

	/* VSNPRINTF_OUT_FD _and_ VSNPRINTF_OUT_FP */
	char                      cache[VSNPRINTF_CACHE_SIZE];
	size_t                    cache_cnt;
	int                       error;  /* 1 if error occured */

	/* VSNPRINTF_OUT_BUF */
	char                     *outbuf;
	size_t                    outbuf_size;

	/* VSNPRINTF_OUT_ALLOC */
	Ml_buf_t                   buf;

	/* ALL */
	int                       cnt;
} snprintf_state_t;


static void dopr(snprintf_state_t *st, const char *format, va_list args);
static void fmtstr(const char *value, int ljust, int len, int zpad, int maxwidth, snprintf_state_t *st);
static void fmtnum(M_int64 value, int base, int dosign, int ljust, int len, int zpad, snprintf_state_t *st);
static void fmtfloat(double value, char type, int ljust, int len, int precision, int pointflag, int zpad, snprintf_state_t *st);
static void dopr_flush_file_cache(snprintf_state_t *st);
static void dopr_out(const char *data, int len, snprintf_state_t *st);
static void dopr_outch(char ch, snprintf_state_t *st);
static void dopr_outch_multi(char c, int cnt, snprintf_state_t *st);
static void dostr(const char *str, int cut, snprintf_state_t *st);


static int snprintf_init_state(snprintf_state_t *st)
{
	memset(st, 0, sizeof(snprintf_state_t));
	st->out_type = VSNPRINTF_OUT_NONE;
	st->fd       = -1;

	return 1;
}


int Ml_vsnprintf(char *str, size_t count, const char *fmt, va_list args)
{
	snprintf_state_t st;

	snprintf_init_state(&st);
	st.out_type    = VSNPRINTF_OUT_BUF;
	st.outbuf      = str;
	st.outbuf_size = count;

	/* Init first character so the buffer is always NULL
	 * terminated even if nothing is appended */
	if (st.outbuf != NULL && st.outbuf_size)
		st.outbuf[0]   = '\0';

	dopr(&st, fmt, args);

	/* Make sure valgrind warns us if the buffer isn't big enough by always
	 * writing the last byte */
	if (st.outbuf != NULL && st.outbuf_size)
		st.outbuf[(st.outbuf_size-1)] = '\0';

	return st.cnt;
}


int Ml_snprintf(char *str, size_t count, const char *fmt,...)
{
	int     len;
	va_list args;

	va_start(args, fmt);
	len = Ml_vsnprintf(str, count, fmt, args);
	va_end(args);
	return len;
}

int Ml_vasprintf(char **ret, const char *fmt, va_list args)
{
	char             *out;
	size_t            outlen;
	snprintf_state_t  st;

	snprintf_init_state(&st);
	st.out_type    = VSNPRINTF_OUT_ALLOC;
	st.buf         = Ml_buf_create();
	dopr(&st, fmt, args);

	Ml_buf_finish(st.buf, &out, &outlen);
	*ret = out;

	return st.cnt;
}


int Ml_asprintf(char **ret, const char *fmt, ...)
{
	int     len;
	va_list ap;

	va_start(ap, fmt);
	len = Ml_vasprintf(ret, fmt, ap);
	va_end(ap);
	return len;
}


int Ml_vfprintf(FILE *fp, const char *fmt, va_list args)
{
	snprintf_state_t st;

	snprintf_init_state(&st);
	st.out_type = VSNPRINTF_OUT_FP;
	st.fp       = fp;

	dopr(&st, fmt, args);

	if (st.error)
		return -1;

	return st.cnt;
}

int Ml_fprintf(FILE *fp, const char *fmt, ...)
{
	int     len;
	va_list ap;

	va_start(ap, fmt);
	len = Ml_vfprintf(fp, fmt, ap);
	va_end(ap);

	return len;
}

int Ml_vdprintf(int fd, const char *fmt, va_list args)
{
	snprintf_state_t st;

	snprintf_init_state(&st);
	st.out_type = VSNPRINTF_OUT_FD;
	st.fd       = fd;
	dopr(&st, fmt, args);

	if (st.error)
		return -1;

	return st.cnt;
}

int Ml_dprintf(int fd, const char *fmt, ...)
{
	int     len;
	va_list ap;

	va_start(ap, fmt);
	len = Ml_vdprintf(fd, fmt, ap);
	va_end(ap);

	return len;
}

int Ml_vprintf(const char *fmt, va_list args)
{
	snprintf_state_t st;

	snprintf_init_state(&st);
	st.out_type = VSNPRINTF_OUT_FD;
	st.fd       = 0;

	dopr(&st, fmt, args);

	if (st.error)
		return -1;

	return st.cnt;
}

int Ml_printf(const char *fmt, ...)
{
	int     len;
	va_list ap;

	va_start(ap, fmt);
	len = Ml_vprintf(fmt, ap);
	va_end(ap);

	return len;
}

#define get_intval(args, longflag, longlongflag, sizetflag, is_unsigned) \
	(M_int64)(is_unsigned?( \
		sizetflag?va_arg(args, size_t):( \
			longlongflag?va_arg(args, M_uint64):( \
				longflag?va_arg(args, unsigned long):( \
					va_arg(args, unsigned int) \
				) \
			) \
		) \
	):( \
		sizetflag?va_arg(args, ssize_t):( \
			longlongflag?va_arg(args, M_int64):( \
				longflag?va_arg(args, long):( \
					va_arg(args, int) \
				) \
			) \
		) \
	))


static void dopr(snprintf_state_t *st, const char *format, va_list args)
{
	char        ch;
	M_int64     value;
	double      fvalue;
	int         longlongflag = 0;
	int         sizetflag    = 0;
	int         longflag     = 0;
	int         pointflag    = 0;
	int         maxwidth     = 0;
	char       *strvalue;
	int         ljust;
	int         len;
	int         zpad;
	const char *str_start    = NULL;

	while ((ch = *format++))
	{
		switch (ch)
		{
			case '%':
				if (str_start) {
					dostr(str_start, (format-str_start)-1, st);
					str_start=NULL;
				}
				ljust = len = zpad = maxwidth = 0;
				longflag = longlongflag = pointflag = sizetflag = 0;
		nextch:
				ch = *format++;
				switch (ch)
				{
					case 0:
						dostr("**end of format**", 0, st);
						dopr_flush_file_cache(st);
						return;
					case '-':
						ljust = 1;
						goto nextch;
					case '0':	/* set zero padding if len not set */
						if (len == 0 && !pointflag)
							zpad = '0';
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
						if (pointflag) {
							maxwidth = maxwidth * 10 + ch - '0';
						} else {
							len = len * 10 + ch - '0';
						}
						goto nextch;
					case '*':
						if (pointflag) {
							maxwidth = va_arg(args, int);
						} else {
							len = va_arg(args, int);
						}
						goto nextch;
					case '.':
						pointflag = 1;
						goto nextch;
					case 'l':
						if (longflag) {
							longlongflag = 1;
						} else {
							longflag = 1;
						}
						goto nextch;
					case 'z':
						sizetflag = 1;
						goto nextch;
					case 'u':
					case 'U':
						/* fmtnum(value,base,dosign,ljust,len,zpad) */
						value = get_intval(args, longflag, longlongflag, sizetflag, 1);
						fmtnum(value, 10, 0, ljust, len, zpad, st);
						break;
					case 'o':
					case 'O':
						/* fmtnum(value,base,dosign,ljust,len,zpad) */
						value = get_intval(args, longflag, longlongflag, sizetflag, 1);
						fmtnum(value, 8, 0, ljust, len, zpad, st);
						break;
					case 'i':
					case 'd':
					case 'D':
						value = get_intval(args, longflag, longlongflag, sizetflag, 0);
						fmtnum(value, 10, 1, ljust, len, zpad, st);
						break;
					case 'x':
						value = get_intval(args, longflag, longlongflag, sizetflag, 1);
						fmtnum(value, 16, 0, ljust, len, zpad, st);
						break;
					case 'X':
						value = get_intval(args, longflag, longlongflag, sizetflag, 1);
						fmtnum(value, -16, 0, ljust, len, zpad, st);
						break;
					case 'p':
						value = (M_uint64)va_arg(args, void *);
						fmtnum(value, 16, 0, ljust, len, zpad, st);
						break;
					case 's':
						strvalue = va_arg(args, char *);
						if (maxwidth > 0 || !pointflag) {
							if (pointflag && len > maxwidth)
								len = maxwidth; /* Adjust padding */
							fmtstr(strvalue, ljust, len, zpad, maxwidth, st);
						}
						break;
					case 'c':
						ch = va_arg(args, int);
						dopr_outch(ch, st);
						break;
					case 'e':
					case 'E':
					case 'f':
					case 'g':
					case 'G':
						fvalue = va_arg(args, double);
						fmtfloat(fvalue, (char)ch, ljust, len, maxwidth, pointflag, zpad, st);
						break;
					case '%':
						dopr_outch(ch, st);
						continue;
					default:
						dostr("???????", 0, st);
						break;
				}
				break;
			default:
				if (str_start == NULL) {
					str_start=format-1;
				}
				break;
		}
	}
	if (str_start) {
		dostr(str_start, (format-str_start)-1, st);
		str_start=NULL;
	}

	dopr_flush_file_cache(st);
}


/*! Do a manual string length calculation as we need to cap it at
 *  maxwidth.
 * \param[in] str      string to get length of
 * \param[in] maxwidth maximum width of string, or 0 if not specified
 * \return length of string
 */
static size_t strlen_to(const char *str, int maxwidth)
{
	size_t i;
	size_t len = 0;

	if (maxwidth <= 0)
		return strlen(str);

	for (i=0; i < (size_t)maxwidth && str[i] != '\0'; i++)
		len++;

	return len;
}

static void fmtstr(const char *value, int ljust, int len, int zpad, int maxwidth, snprintf_state_t *st)
{
	int padlen;
	int str_len; /* amount to pad */
	int padch;

	if (zpad)
		padch = '0';
	else
		padch = ' ';

	if (value == 0)
		value = "<NULL>";

	str_len = strlen_to(value, maxwidth);

	padlen = len - str_len;

	if (padlen < 0)
		padlen = 0;

	if (ljust)
		padlen = -padlen;

	if (padlen > 0) {
		dopr_outch_multi(padch, padlen, st);
		padlen=0;
	}

	dostr(value, maxwidth, st);
	dopr_outch_multi(padch, -padlen, st);
}

static void fmtnum(M_int64 value, int base, int dosign, int ljust, int len, int zpad, snprintf_state_t *st)
{
	char     signvalue    = 0;
	M_uint64 uvalue;
	char     convert[64];
	int      place        = 0;
	int      padlen       = 0;  /* amount to pad */
	int      caps         = 0;
	int      i;
	int      c;

	/*
	 * DEBUGP(("value 0x%x, base %d, dosign %d, ljust %d, len %d, zpad
	 * %d\n", value, base, dosign, ljust, len, zpad ));
	 */
	uvalue = value;
	if (dosign) {
		if (value < 0) {
			signvalue = '-';
			uvalue = -value;
		}
	}

	if (base < 0) {
		caps = 1;
		base = -base;
	}

	do {
		convert[place++] = (caps ? "0123456789ABCDEF" : "0123456789abcdef")
			[uvalue % (unsigned) base];
		uvalue = (uvalue / (unsigned) base);
	} while (uvalue);
	convert[place] = 0;

	if (len < 0) {
		/* this could happen with a "*" width spec */
		ljust = 1;
		len = -len;
	}

	padlen = len - place;
	if (padlen < 0)
		padlen = 0;

	if (ljust)
		padlen = -padlen;

	/*
	 * DEBUGP(( "str '%s', place %d, sign %c, padlen %d\n",
	 * convert,place,signvalue,padlen));
	 */
	if (zpad && padlen > 0) {
		if (signvalue) {
			dopr_outch(signvalue, st);
			--padlen;
			signvalue = 0;
		}

		if (padlen > 0) {
			dopr_outch_multi(zpad, padlen, st);
			padlen=0;
		}
	}

	if (padlen > 0) {
		dopr_outch_multi(' ', padlen, st);
		padlen=0;
	}

	if (signvalue)
		dopr_outch(signvalue, st);

	/* Must reverse convert, as it is in reverse order */
	for (i=0; i<place-(i+1); i++) {
		c=convert[i];
		convert[i]=convert[place-i-1];
		convert[place-i-1]=c;
	}
	dostr(convert, place, st);

	if (padlen < 0) {
		dopr_outch_multi(' ', -padlen, st);
	}
}

static void fmtfloat(double value, char type, int ljust, int len, int precision, int pointflag, int zpad, snprintf_state_t *st)
{
	char convert[512];
	int  padlen       = 0; /* amount to pad */
	int  isneg        = 0;
	char fmt[32];

	/* We don't really want to support any sort of floating point, but in case there
	 * is some ancient code printing floats, go ahead and rely on the system's
	 * sprintf() to provide it.
	 * NOTE: we cannot rely on the system itself providing snprintf(), so we
	 *       provide an insanely large buffer */
#ifdef sprintf
#  undef sprintf
#endif
	if (pointflag)
		sprintf(fmt, "%%.%d%c", precision, type);
	else
		sprintf(fmt, "%%%c", type);
#define abs_val(a) (a >= 0)?a:(-a)
	sprintf(convert, fmt, abs_val(value));

	if (len < 0) {
		/* this could happen with a "*" width spec */
		ljust = 1;
		len = -len;
	}

	if (value < 0)
		isneg=1;

	if (isneg) {
	  padlen = len - strlen(convert) - 1;
	} else {
	  padlen = len - strlen(convert);
	}

	if (padlen < 0)
		padlen = 0;

	if (ljust)
		padlen = -padlen;

	if (isneg && zpad)
		dopr_outch('-', st);


	if (padlen > 0) {
		if (zpad) {
		  dopr_outch_multi(zpad, padlen, st);
		} else {
		  dopr_outch_multi(' ', padlen, st);
		}
		padlen=0;
	}

	if (isneg && !zpad)
		dopr_outch('-', st);

	dostr(convert, 0, st);

	if (padlen < 0) {
		if (zpad) {
		  dopr_outch_multi(zpad, -padlen, st);
		} else {
		  dopr_outch_multi(' ', -padlen, st);
		}
	}
}

static void dopr_outch(char ch, snprintf_state_t *st)
{
	dopr_out(&ch, 1, st);
}


static void dostr(const char *str, int maxwidth, snprintf_state_t *st)
{
	dopr_out(str, strlen_to(str, maxwidth), st);
}


static void dopr_outch_multi(char c, int cnt, snprintf_state_t *st)
{
	char  buf[256];
	char *ptr;
	char *temp     = NULL;

	if (cnt <= 0)
		return;

	if (cnt > (int)sizeof(buf)) {
		temp = Ml_malloc(cnt);
		ptr  = temp;
	} else {
		ptr = buf;
	}

	memset(ptr, c, cnt);
	dopr_out(ptr, cnt, st);

	Ml_free(temp);
}


static void dopr_write_file(const char *data, size_t len, snprintf_state_t *st)
{
	if (st->error)
		return;

	if (st->out_type == VSNPRINTF_OUT_FD) {
		if (write(st->fd, data, len) != (int)len)
			st->error = 1;
	} else if (st->out_type == VSNPRINTF_OUT_FP) {
		if (fwrite(data, 1, len, st->fp) != (size_t)len)
			st->error = 1;
	}
}


static void dopr_flush_file_cache(snprintf_state_t *st)
{
	dopr_write_file(st->cache, st->cache_cnt, st);
	st->cache_cnt = 0;
}

static void dopr_out(const char *data, int len, snprintf_state_t *st)
{
	int write_bytes;

	if (len <= 0)
		return;

	switch (st->out_type) {
		case VSNPRINTF_OUT_NONE:
			/* Shouldn't really happen, but we could just be getting a count
			 * of possible bytes */
			break;
		case VSNPRINTF_OUT_BUF:
			write_bytes = len;
			if (st->cnt + write_bytes >= (int)st->outbuf_size) {
				write_bytes = (int)st->outbuf_size - st->cnt - 1;
			}
			if (write_bytes > 0) {
				memcpy(st->outbuf + st->cnt, data, write_bytes);
				st->outbuf[st->cnt + write_bytes] = 0;
			}
			break;
		case VSNPRINTF_OUT_FD:
		case VSNPRINTF_OUT_FP:
			/* Should only loop at most twice */
			while (len) {
				/* Write as many bytes as we can to the cache */
				write_bytes = len;
				if (st->cache_cnt + write_bytes > (int)sizeof(st->cache)) {
					write_bytes = (int)sizeof(st->cache) - st->cache_cnt;
				}
				memcpy(st->cache + st->cache_cnt, data, write_bytes);
				st->cache_cnt += write_bytes;
				len  -= write_bytes;
				if (len)
					data += write_bytes;

				/* Flush cache if cache is full */
				if (st->cache_cnt == sizeof(st->cache))
					dopr_flush_file_cache(st);

				/* If there are _still_ more than sizeof(st->cache) bytes left to
				 * write, go ahead and flush them all out too */
				if ((size_t)len >= sizeof(st->cache)) {
					dopr_write_file(data, len, st);
					len = 0;
				}
			}

			break;
		case VSNPRINTF_OUT_ALLOC:
			Ml_buf_add_bytes(st->buf, data, len);
			break;
	}

	/* According to the snprintf() manpage, cnt is the number of bytes that
	 * would have been written, even if they couldn't all be written */
	st->cnt += len;
}
