#include "main.h"
#include "monetra.h"
#include "mstring.h"
#include "mstdlib.h"
#include "proxy.h"

#ifdef ENABLE_SSL
#    include <openssl/crypto.h>
#    include <openssl/x509.h>
#    include <openssl/pem.h>
#    include <openssl/ssl.h>
#    include <openssl/err.h>
#    include <openssl/rand.h>
#endif

#ifndef _WIN32
#    include <netdb.h>
#    include <sys/socket.h>
#    include <netinet/in.h>
#    include <netinet/tcp.h>
#    include <arpa/inet.h>
#    ifdef __sun__
#        include <xti.h>
#    endif
#else
#    include <windows.h>
#    include <ws2tcpip.h>
#    include <mstcpip.h>
#endif

int M_InitSockets()
{
#ifdef _WIN32
	WSADATA wsaData;

	if (WSAStartup(0x202, &wsaData) == SOCKET_ERROR) {
		WSACleanup();
		return 0;
	}

#endif /* _WIN32 */
	return 1;
}

void M_closesocket(int fd, int call_shutdown)
{
	if (call_shutdown) {
		M_SetNonBlock(fd, 0); /* Make sure we're in blocking for shutdown() */
#ifdef _WIN32
		shutdown(fd, SD_BOTH);
#else
#    if !defined (SHUT_RDWR) && defined (SHUT_RD)
#        define SHUT_RDWR SHUT_RD
#    elif !defined (SHUT_RDWR)
#        define SHUT_RDWR 2
#    endif
		shutdown(fd, SHUT_RDWR);
#endif
		M_uwait(10000); /* Wait arbitrary time for shutdown */
	}

#ifdef _WIN32
	closesocket(fd);
#else
	close(fd);
#endif
}

int M_SetNonBlock(int fd, int tf)
{
#ifdef O_NONBLOCK
	int flags;

	if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
		return 0;
	}

	if (tf) {
		flags |= O_NONBLOCK;
	} else {
		flags &= ~(O_NONBLOCK);
	}

	if (fcntl(fd, F_SETFL, flags) == -1) {
		return 0;
	}

#else /* ifdef O_NONBLOCK */
#    ifndef __WATCOMC__
	if (ioctl(fd, FIONBIO, &tf) == -1)
#    else
	if (ioctl(fd, FIONBIO, (char *)&tf) == -1)
#    endif
	{ return 0; }

#endif /* ifdef O_NONBLOCK */
	return 1;
}

long M_read_socket(const char *dbg_head, int dbg_level, char *buf, int buflen, int fd)
{
	UNUSED(dbg_head);
	UNUSED(dbg_level);
#ifdef _WIN32
	return recv(fd, buf, buflen, 0);
#else
	return read(fd, buf, buflen);
#endif
}

long M_write_socket(const char *dbg_head, int dbg_level, const char *buf, int buflen, int fd)
{
	UNUSED(dbg_head);
	UNUSED(dbg_level);
#ifdef _WIN32
	return send(fd, buf, buflen, 0);
#else
	return write(fd, buf, buflen);
#endif
}


/*! Section 2.7 and 2.8 of the Unix Socket FAQ explain these
 * http://www.unl.csi.cuny.edu/faqs/sock-faq/html/unix-socket-faq-2.html#time_wait
 * Need to explore: SOL_TCP options : TCP_KEEPIDLE/TCP_KEEPCNT/TCP_KEEPINTVL 
 *     man 7 tcp
 *
 * Windows seems to have something here SIO_KEEPALIVE_VALS (milliseconds)
 *    http://msdn2.microsoft.com/en-us/library/ms741621.aspx
 * 
 * \param fd         File descriptor to set keepalive values on
 * \param idle_s     Time in seconds that the connection is idle before keepalives
 *                   begin.
 * \param interval_s Time in seconds between retries.
 * \param maxretry   Maximum number of retry attempts (at interval_s apart) without
 *                   response until the connection is dropped.
 * \return 1 on success, 0 on failure
 */
static int M_ip_set_keepalives(int fd, int idle_s, int interval_s, int maxretry)
{
#if defined(SO_KEEPALIVE) || defined(TCP_KEEPIDLE) || defined(TCP_KEEPCNT) || defined(TCP_KEEPINTVL)
	int on=1;
#endif
#if defined(TCP_KEEPALIVE) && defined(__sun__)
	struct t_kpalive ka;
#endif
#ifdef _WIN32
	struct tcp_keepalive ka = { 0 }, kar = { 0 };
	DWORD dwBytes = 0;
#endif

#ifdef SO_KEEPALIVE
	on = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
#  ifdef _WIN32
        (const char *)
#  endif
        &on, sizeof(on)) == -1) {
		return 0;
	}
#endif

#ifndef _WIN32
#  ifdef TCP_KEEPIDLE
	on = idle_s; /* how long (seconds) the connection is idle before sending keepalive probes */
	if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPIDLE, &on, sizeof(on)) == -1) {
		return 0;
	}
#  endif

#  ifdef TCP_KEEPCNT
	on = maxretry; /* max number of probes until connection is considered dead */
	if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPCNT, &on, sizeof(on)) == -1) {
		return 0;
	}
#  endif

#  ifdef TCP_NKEEP
	/* SCO's version of TCP_KEEPCNT */
	on = maxretry; /* max number of probes until connection is considered dead */
	if (setsockopt(fd, IPPROTO_TCP, TCP_NKEEP, &on, sizeof(on)) == -1) {
		return 0;
	}
#  endif

#  ifdef TCP_KEEPINTVL
	on = interval_s; /* time in seconds between individual keepalive probes */
	if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPINTVL, &on, sizeof(on)) == -1) {
		return 0;
	}
#  endif

#  if defined(TCP_KEEPALIVE)
#    if defined(__APPLE__)
	/* Should we only use this on MacOSX?  Some platforms seem to
	 * say they'll take a structure instead of an integer */
	on = idle_s; /* time in seconds idle before keepalive probes */
	if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPALIVE, &on, sizeof(on)) == -1) {
		return 0;
	}
#    elif defined(__sun__)
	ka.kp_onoff   = 1;
	ka.kp_timeout = interval_s / 60; /* interval timeout in minutes */
	if (ka.kp_timeout == 0)
		ka.kp_timeout = 1;
	if (setsockopt(fd, IPPROTO_TCP, TCP_KEEPALIVE, &ka, sizeof(ka)) == -1) {
		return 0;
	}
#    endif
#  endif

#else /* _WIN32 */
	ka.onoff = 1;
	ka.keepalivetime = idle_s * 1000;
	/* Keepalive retries is hardcoded to 10, so we need to come up with a ratio
	 * interval such that (maxretry * interval_s) == (10 * interval_s2) */
	ka.keepaliveinterval = maxretry * interval_s * 1000 / 10;
	if (WSAIoctl(fd, SIO_KEEPALIVE_VALS, &ka, sizeof(ka), &kar, sizeof(kar), &dwBytes, NULL, NULL) != 0) {
		return 0;
	}
#endif /* _WIN32 */

	return 1;
}


#ifndef INADDR_NONE
#    define INADDR_NONE 0xFFFFFFFF
#endif

int M_ip_connect(M_CONN *conn)
{
	struct sockaddr_in peer;
	int soct;
	int len;
	char *addr = NULL;
	int ip_connect_timeout = 5;
	char ipaddr[50];
	char error[255];
	char temperr[255];
	int is_true = 1;
#ifdef _WIN32
	const char *w32_addr = NULL;
#endif
	proxy_config_st *proxy = NULL;
	time_t starttime, t;

	proxy = &((*conn)->proxy);
	ip_connect_timeout = (*conn)->max_conn_time;

	time(&starttime);
	if (LIBMONETRA_proxy_tunnel_check(proxy, (*conn)->location, (*conn)->port)) {
		soct = LIBMONETRA_proxy_tunnel_connect(proxy, (*conn)->location, (*conn)->port, error, sizeof(error));
		if (soct == -1) {
			Ml_snprintf(temperr, sizeof(temperr), "Proxy Connect Failed: %s", error);
			M_Set_Conn_Error(conn, temperr);
			return -1;
		}
	} else {
		memset(&peer, 0, sizeof(peer));
		peer.sin_family = AF_INET;
		peer.sin_port = htons((*conn)->port);

		memset(temperr, 0, sizeof(temperr));
		len = LIBMONETRA_ghbn((*conn)->location, &addr, error, sizeof(error));
		if (len != -1) {
			memcpy((char *)&peer.sin_addr, addr, len);
			Ml_free(addr);
		} else if (strcasecmp((*conn)->location, "localhost") == 0) {
			Ml_free(addr);
			peer.sin_addr.s_addr = inet_addr("127.0.0.1");
		} else {
			Ml_free(addr);
			Ml_snprintf(temperr, sizeof(temperr), "DNS Lookup Failed: %s", error);
			M_Set_Conn_Error(conn, temperr);
			return -1;
		}

		memset(ipaddr, 0, sizeof(ipaddr));
#ifdef _WIN32
		w32_addr = inet_ntoa(peer.sin_addr);
		if (w32_addr != NULL) {
			M_strlcpy(ipaddr, w32_addr, sizeof(ipaddr));
		}

#else
		inet_ntop(AF_INET, (const void *)&peer.sin_addr, ipaddr, sizeof(ipaddr));
#endif

		time(&t);
		starttime = t;
		soct = socket(AF_INET, SOCK_STREAM, 0);
		if (soct < 0) {
			Ml_snprintf(error, sizeof(error), "socket() failed, could not get fd: %s", strerror(errno));
			M_Set_Conn_Error(conn, error);
			return -1;
		}

		if (!LIBMONETRA_connect_nonblock(soct, (struct sockaddr *)&peer, sizeof(peer), ip_connect_timeout,
			    error, sizeof(error))) {
			M_closesocket(soct, 0);
			Ml_snprintf(temperr, sizeof(temperr), "IP Connection failed: %s", error);
			M_Set_Conn_Error(conn, temperr);
			return -1;
		}
	}

	time(&t);

	/* Should detect a disconnect in 5 * 4 = 20s */
	M_ip_set_keepalives(soct, 10 /* idle_s */, 5 /* interval_s */, 4 /* maxretry */);

	/* Disable Nagle algorithm.  Monetra handles packets appropriately to not need it, just
	 * adds latency otherwise */
	setsockopt(soct, IPPROTO_TCP, TCP_NODELAY, &is_true, sizeof(is_true));

	return soct;
}

int M_VerifyPing(M_CONN *conn)
{
	M_uintptr identifier = -1;
	time_t t, lastt;
	int status = 0;
	int blocking = 0;

	blocking = (*conn)->blocking;

	M_SetBlocking(conn, 0);

	identifier = M_Ping(conn);
	time(&lastt);
	t = lastt;
	status = 1;
	while (M_CheckStatus(conn, identifier) != M_DONE) {
		if (!M_Monitor_ex(conn, 20000)) {
			status = 0;
			break;
		}

		time(&t);
		if ((t - lastt) > 5) {  /* Response Timeout */
			status = 0;
			break;
		}
	}

	M_SetBlocking(conn, blocking);

	M_DeleteResponse(conn, identifier);
	if (!status) {
		return 0;
	}

	return 1;
}

/* Let's see if there's anything in the buffer to read! */
int M_CheckRead_select(int ptr, long delay)
{
	fd_set readfs;
	struct timeval timeout;
	int res;

	FD_ZERO(&readfs);
	FD_SET(ptr, &readfs);
	timeout.tv_sec = 0;
	timeout.tv_usec = delay;
	res = select(ptr + 1, &readfs, NULL, NULL, &timeout);
	if (res > 0) {
		if (FD_ISSET(ptr, &readfs)) {
			return 1;
		}
	}

	return 0;
}

/* Let's make sure that the buffer isn't too full to write to
 *  I'd assume this will only happen in dial-up scenerios */
int M_CheckWrite_select(int ptr, long delay)
{
	fd_set sendfs;
	struct timeval timeout;
	int res;

	FD_ZERO(&sendfs);
	FD_SET(ptr, &sendfs);
	timeout.tv_sec = 0;
	timeout.tv_usec = delay;

	res = select(ptr + 1, NULL, &sendfs, NULL, &timeout);
	if (res > 0) {
		if (FD_ISSET(ptr, &sendfs)) {
			return 1;
		}
	}

	return 0;
}

#ifndef _WIN32
static FILE *M_openlog(void)
{
	char debug_file[255];
	FILE *fp = NULL;

	Ml_snprintf(debug_file, sizeof(debug_file), "/tmp/libmonetra-%d.log", getpid());
	fp = fopen(debug_file, "ab");
	return fp;
}

static void M_output_buffer(FILE *fp, const char *prefix, const char *buf, int len)
{
	int i,c;

	if (fp == NULL) {
		return;
	}

	for (i = 0; i < len; i++) {
		c = buf[i];
		if ((c < 32) || (c > 126)) {
			c = 32;
		}

		fprintf(fp, "%s: %05d : %c -- HEX : 0x%0x DEC: %02d\n", prefix, len, c, buf[i], buf[i]);
	}
}

#endif /* ifndef _WIN32 */


int M_Monitor_IP(M_CONN *conn, unsigned long delay)
{
	char temp_buffer[IP_READ_SIZE];

#ifndef _WIN32
	FILE *fp = NULL;
#endif
	int bytes_read = 0;
	int bytes_written = 0;
	int status = 1;
	size_t outlen;
	int loopcnt;

#ifndef _WIN32
	if ((*conn)->do_debug) {
		fp = M_openlog();
	}

#endif

#ifndef _WIN32
	if ((*conn)->do_debug && (fp != NULL)) {
		fprintf(fp, "Looking to write\n");
	}

#endif
	M_lock(conn);
	if (status && (*conn)->outbuf != NULL) {
		if (M_CheckWrite_select((*conn)->ptr, 0)) {

			/* Currently this tries to write everything at once */
			outlen = Ml_buf_len((*conn)->outbuf);
			if ((*conn)->conn_method == M_SOCKETS) {
				bytes_written = M_write_socket(NULL, 0, Ml_buf_peek((*conn)->outbuf), outlen, (*conn)->ptr);
			} else if ((*conn)->conn_method == M_SSL) {
#ifdef ENABLE_SSL
				bytes_written = SSL_write((SSL *)(*conn)->ssl, Ml_buf_peek((*conn)->outbuf), outlen);
#endif
			}

			if (bytes_written <= 0) {
				/* connection closed */
				status = 0;
			}

			if (status) {
#ifndef _WIN32
				if ((*conn)->do_debug && (fp != NULL)) {
					fprintf(fp, "Wrote %d bytes\n", bytes_written);
					M_output_buffer(fp, "WROTE", Ml_buf_peek((*conn)->outbuf), bytes_written);
				}
#endif
				if ((size_t)bytes_written < Ml_buf_len((*conn)->outbuf)) {
					Ml_buf_drop((*conn)->outbuf, bytes_written);
				} else {
					Ml_buf_cancel((*conn)->outbuf);
					(*conn)->outbuf = NULL;
				}
			}
		}
	}

	M_unlock(conn);

#ifndef _WIN32
	if ((*conn)->do_debug && (fp != NULL)) {
		fprintf(fp, "Looking to read\n");
	}

#endif
	M_lock(conn);
	loopcnt = 0;
	while (status && (
#ifdef ENABLE_SSL
	               ((*conn)->conn_method == M_SSL && SSL_pending((SSL *)(*conn)->ssl)) ||
#endif
	               M_CheckRead_select((*conn)->ptr, delay))) {
		if ((*conn)->conn_method == M_SOCKETS) {
			bytes_read = M_read_socket(NULL, 0, temp_buffer, IP_READ_SIZE, (*conn)->ptr);
		} else if ((*conn)->conn_method == M_SSL) {
#ifdef ENABLE_SSL
			bytes_read = SSL_read((SSL *)(*conn)->ssl, temp_buffer, IP_READ_SIZE);
#endif
		}

#ifndef _WIN32
		if ((*conn)->do_debug && (fp != NULL)) {
			fprintf(fp, "Read %d bytes\n", bytes_read);
			M_output_buffer(fp, "READ", temp_buffer, bytes_read);
		}
#endif

		if (bytes_read <= 0) {
			/* connection closed */
			status = 0;
			break;
		}

		if ((*conn)->inbuf == NULL)
			(*conn)->inbuf = Ml_buf_create();
		Ml_buf_add_bytes((*conn)->inbuf, temp_buffer, bytes_read);

#ifndef _WIN32
		if ((*conn)->do_debug && (fp != NULL)) {
			fprintf(fp, "Full inbuffer size: %d\n", (int)Ml_buf_len((*conn)->inbuf));
			M_output_buffer(fp, "FULL INBUF", Ml_buf_peek((*conn)->inbuf), Ml_buf_len((*conn)->inbuf));
		}
#endif

		loopcnt++;

		/* Don't loop if we didn't fill the buffer */
		if ((bytes_read < IP_READ_SIZE) || (loopcnt > 256)) {
			break;
		}
	}

	M_unlock(conn);

#ifndef _WIN32
	if ((*conn)->do_debug) {
		fclose(fp);
	}

#endif
	return status;
}

int M_SendTransaction_IP(M_CONN *conn, M_uint64 identifier, char *transaction)
{

	M_lock(conn);

	if ((*conn)->outbuf == NULL)
		(*conn)->outbuf = Ml_buf_create();
	Ml_buf_add_byte((*conn)->outbuf, 0x02);
	Ml_buf_add_uint((*conn)->outbuf, identifier);
	Ml_buf_add_byte((*conn)->outbuf, 0x1c);
	Ml_buf_add_str((*conn)->outbuf, transaction);
	Ml_buf_add_byte((*conn)->outbuf, 0x03);

	M_unlock(conn);

	return 1;
}
