#include "main.h"
#include "atouint64.h"
#include "mstdlib.h"
#include "buf.h"
#include <ctype.h>

#ifndef _WIN32
#    include <strings.h>
#endif

static char *M_encode_quoted(const char *str, int len);
static char *M_StructureTransaction(M_CONN *myconn, M_QUEUE *queue);

void UnescapeQuotes(char *quoted_data)
{
	int i, cnt;

	for (i = 0, cnt = 0; quoted_data != NULL && quoted_data[i] != 0; i++) {
		if ((quoted_data[i] == '"') && (quoted_data[i + 1] == '"')) {
			/* Skip one of the quotes */
			i++;
		}

		quoted_data[cnt++] = quoted_data[i];
	}

	if (quoted_data) {
		quoted_data[cnt] = 0;
	}
}


static char *M_StructureTransaction(M_CONN *conn, M_QUEUE *queue)
{
	char  *out = NULL;
	size_t out_len = 0;
	char *temp = NULL;
	int i;
	Ml_buf_t buf;

	UNUSED(conn);

	buf = Ml_buf_create();

	for (i = 0; i < queue->transaction_fields; i++) {
		if (
			!strcasecmp(queue->transaction[i].key,   "action") &&
			!strcasecmp(queue->transaction[i].value, "ping")
		) {
			Ml_buf_add_str(buf, "PING\r\n");
			Ml_buf_finish(buf, &out, &out_len);
			return out;
		}
	}

	for (i = 0; i < queue->transaction_fields; i++) {
		Ml_buf_add_str(buf, queue->transaction[i].key);
		Ml_buf_add_byte(buf, '=');
		if (queue->transaction[i].value) {
			temp    = M_encode_quoted(queue->transaction[i].value, strlen(queue->transaction[i].value));
			if (temp == NULL) {
				Ml_buf_add_str(buf, queue->transaction[i].value);
			} else {
				Ml_buf_add_str(buf, temp);
				Ml_free(temp);
			}
		}

		Ml_buf_add_str(buf, "\r\n");
	}
	Ml_buf_finish(buf, &out, &out_len);
	return out;
}

int M_SendTransaction(M_CONN *conn, M_QUEUE *queue)
{
	char *transaction = NULL;
	int ret = 0;

	transaction = M_StructureTransaction(conn, queue);

	if (((*conn)->conn_method == M_SOCKETS) || ((*conn)->conn_method == M_SSL)) {
		ret = M_SendTransaction_IP(conn, queue->identifier, transaction);
	}

	Ml_free(transaction);

	/* pos=M_AddOutstandingAuth(conn, identifier); */
	M_lock(conn);
	queue->status = M_PENDING;
	M_unlock(conn);

	return ret;
}

int Ml_verify_trans_in_queue(M_CONN *conn, M_uintptr identifier)
{
	void *trans = (void *)identifier;
	int ret = 0;
	M_QUEUE *queue = NULL, *ptr = NULL;

	if (conn == NULL) {
		return 0;
	}

	M_lock(conn);
	if ((*conn)->validate_identifier) {
		queue = (*conn)->queue;
		ptr = queue;
		while (ptr != NULL) {
			if (ptr == trans) {
				ret = 1;
				break;
			}

			ptr = ptr->next;
			/* Don't remember if it's circular .. but just in case */
			if (ptr == queue) {
				break;
			}
		}
	} else {
		ret = 1;
	}

	M_unlock(conn);
	return ret;
}

/* Returns NULL if no quoting needed, otherwise allocated
 * NULL-terminated buffer of quoted data */
static char *M_encode_quoted(const char *str, int len)
{
	int alloc_len = 0;
	int cnt = 0, i;
	char *ret = NULL;
	int needs_quoting = 0; /* 0 = no, 1 = yes, 2 = yes and has embedded quotes */

	if (str == NULL)
		return 0;

	/* If starting or trailing whitespace, needs to be quoted */
	if (len && (str[0] == ' ' || str[0] == '\t' || str[len-1] == ' ' || str[len-1] == '\t'))
		needs_quoting = 1;

	for (i = 0; i < len; i++) {
		switch (str[i]) {
			case '\r':
			case '\n':
			case ',':
				needs_quoting = 1;
				break;
			case '"':
				needs_quoting = 2;
				break;
		}

		if (needs_quoting == 2) {
			break;
		}
	}

	if (!needs_quoting) {
		return NULL;
	}

	if (needs_quoting == 2) {
		alloc_len = (len * 2) + 2;
	} else if (needs_quoting == 1) {
		alloc_len = len + 2;
	}

	ret = Ml_malloc(alloc_len + 1);

	if (needs_quoting == 1) {
		ret[0] = '"';
		memcpy(ret + 1, str, len);
		ret[len + 1] = '"';
		ret[len + 2] = 0;
	} else {
		ret[cnt++] = '"';
		for (i = 0; i < len; i++) {
			if (str[i] == '"') {
				ret[cnt++] = '"';
			}

			ret[cnt++] = str[i];
		}

		ret[cnt++] = '"';
		ret[cnt] = 0;
	}

	return ret;
}


/* Find closing bracket for pattern range, honor escaping, and 
 * ignore closing bracket as first character or second character
 * if after '!' */
static const char *pattern_range_close(const char *pat)
{
	int i, len;
	char escaped = 0;
	if (pat == NULL) {
		len = 0;
	} else {
		len = strlen(pat);
	}
	for (i=0; i<len; i++) {
		switch(pat[i]) {
			case '\\':
				if (escaped) escaped = 0;
				else escaped = 1;
			break;
			case ']':
				if (!(i == 0 || (i == 1 && pat[0] == '!')) && !escaped)
					return(pat+i);
				escaped = 0;
			break;
			default:
				escaped = 0;
			break;
		}
	}
	return(NULL);
}

/* Match pattern (pat) starting at pat_start to string (str) starting at str_start */
static int pattern_matches(const char *pat, const char *str, int pat_start, int str_start)
{
	int c, c2, c3, i;
	const char *rb = NULL;
	int pat_len, str_len;
	int not_pat = 0;

	if (pat == NULL || str == NULL) return(0);
	pat_len = strlen(pat);
	str_len = strlen(str);

	/* If done with both pattern and string, then it's a match, but
	 * if pattern is done, but string is not, fail */
	if (pat_start == pat_len) return(str_len == str_start);

	c=pat[pat_start];
	switch(c) {
		case '\\':
			if (pat_len == pat_start+1) return(0); /* Error */
			c2=pat[pat_start+1];
			if (c2 == '[' || c2 == ']' || c2 == '?' || c2 == '*' || c2 == '\\') {
				if (str_len == str_start) return(0); /* String used up */
				if (str[str_start] == c2)
					return(pattern_matches(pat, str, pat_start+2, str_start+1));
			} else
				return(0); /* Error */
		break;
		case '[':
			if (str_len == str_start) return(0); /* Nothing to match */
			rb = pattern_range_close(pat+(pat_start+1)); /* Closing right bracket */
			if (rb == NULL) return(0); /* Error in pattern */
			/* If first character is '!' match everything but */
			if (pat[pat_start+1] == '!') {
				pat_start++;
				not_pat = 1;
			} else
				not_pat = 0;
			for (i=pat_start+1; i<rb-pat; i++) {
				c2=pat[i];
				switch(c2) {
					case '\\':
						c3=pat[i+1];
						if (c3 == '[' || c3 == ']' || c3 == '?' || c3 == '*' || c3 == '\\') {
							if (str[str_start] == c3) {
								if (not_pat) return(0);
								return(pattern_matches(pat, str, rb-pat+1, str_start+1));
							}
						} else
							return(0); /* Error in pattern */
					break;
					case '-':
						if (i == pat_start+1 || i == rb-pat-1) {
							if (str[str_start] == '-') { /* First char or last in brackets */
								if (not_pat) return(0);
								return(pattern_matches(pat, str, rb-pat+1, str_start+1));
							}
						} else {
							/* Check Range */
							if (pat[i-1] < str[str_start] && str[str_start] < pat[i+1]) {
								if (not_pat) return(0);
								return(pattern_matches(pat, str, rb-pat+1, str_start+1));
							}
						}
					break;
					case ']':
						if (i == pat_start+1) {
							if (str[str_start] == c2) {
								if (not_pat) return(0);
								return(pattern_matches(pat, str, rb-pat+1, str_start+1));
							}
						} else {
							return(0);
						}
					break;
					
					default:
						if (str[str_start] == c2) {
							if (not_pat) return(0);
							return (pattern_matches(pat, str, rb-pat+1, str_start+1));
						}
					break;
				}
			}
			/* Nothing matched */
			if (not_pat) return (pattern_matches(pat, str, rb-pat+1, str_start+1));
			return(0);
		break;
		case ']':
			return(0); /* Error in pattern */
		break;
		case '?':
			if (str_len == str_start) return(0); /* String used up */
			/* ? always matches 1 char */
			return(pattern_matches(pat, str, pat_start+1, str_start+1));
		break;
		case '*':
			if (pat_start == pat_len-1) /* * matches anything */
				return(1);
			for (i=str_start; i<=str_len; i++) {
				/* Cycle through characters until we get a full pattern match */
				if (pattern_matches(pat, str, pat_start+1, i))
					return(1);
			}
			/* No pattern match */
			return(0);
		break;
		default:
			if (str_len == str_start) return(0); /* String used up */
			else if (str[str_start] == c)
				return(pattern_matches(pat, str, pat_start+1, str_start+1));
			else
				return(0);
		break;
	}
	return(0);
}

/* Match pattern against string as per  'man 7 glob' though we
 * don't support newer POSIX functions like named character classes
 * (e.g. [:lower:]), collating symbols, or equivalence class expressions */
int Ml_pattern_match(const char *pattern, const char *str)
{
	return(pattern_matches(pattern, str, 0, 0));
}


char **Ml_str_explode(char delim, const char *s, size_t s_len, int *num, int **len_array)
{
	size_t          i;
	size_t          num_strs = 1;
	unsigned char **out      = NULL;
	unsigned char  *dupstr   = NULL;

	*num = 0;
	if (s == NULL)
		return NULL;

	dupstr = Ml_malloc(s_len + 1);
	if (s_len > 0)
		memcpy(dupstr, s, s_len);
	dupstr[s_len] = 0;

	for (i=0; i<s_len; i++)
		if (dupstr[i] == delim)
			num_strs++;

	out = Ml_malloc(num_strs * sizeof(*out));
	if (len_array != NULL)
		*len_array = Ml_malloc(num_strs * sizeof(**len_array));
	*num = (int)num_strs;

	out[0] = dupstr;
	num_strs = 1;
	for (i=0; i<s_len; i++) {
		if (dupstr[i] == delim) {
			dupstr[i] = 0;
			out[num_strs] = dupstr+(i+1);
			if (len_array != NULL)
				(*len_array)[num_strs-1] = (int)((dupstr+i) - out[num_strs-1]);
			num_strs++;
		}
	}
	if (len_array != NULL)
		(*len_array)[num_strs-1] = (int)((dupstr+s_len) - out[num_strs-1]);

	return (char **)out;
}

void Ml_str_explode_free(char **strs, size_t num)
{
	if (strs != NULL) {
		/* First pointer contains entire buffer */
		if (num > 0) Ml_free(strs[0]);
		Ml_free(strs);
	}
}

char **Ml_str_explode_str(char delim, const char *s, int *num)
{
	if (s == NULL)
		return NULL;
	return Ml_str_explode(delim, s, strlen(s), num, NULL);
}


#define Ml_MIN(a,b)     ((a)<(b)?(a):(b))
int Ml_str_cpy(char *dest, const char *src, size_t dest_len)
{
	size_t copy_len;

	if (dest == NULL)
		return 0;

	copy_len = Ml_MIN(strlen(M_Safe(src)), dest_len-1);

	if (src != NULL)
		memmove(dest, src, copy_len);

	dest[copy_len] = 0;

	return 1;
}

void Ml_str_lower(char *str)
{
	for ( ; *str != '\0'; str++) {
		*str = tolower(*str);
		str++;
	}
}


static int Ml_str_implode_has_restricted_chars(const char *s, char delim_char, char enclose_char, char escape_char)
{
	size_t i;
	size_t len;

	len = strlen(M_Safe(s));

	for (i=0; i<len; i++) {
		if (s[i] == delim_char || s[i] == enclose_char || s[i] == escape_char)
			return 1;
	}
	return 0;
}


static void Ml_str_implode_escape(Ml_buf_t buf, const char *s, char enclose_char, char escape_char)
{
	size_t len;
	size_t i;

	if (s == NULL) {
		len = 0;
	} else {
		len = strlen(s);
	}

	for (i=0; i<len; i++) {
		if (s[i] == enclose_char || s[i] == escape_char)
			Ml_buf_add_byte(buf, escape_char);

		Ml_buf_add_byte(buf, s[i]);
	}
}


char *Ml_str_implode(char delim_char, char enclose_char, char escape_char, char **strs, size_t num_strs, int always_enclose)
{
	size_t   i;
	char    *ret    = NULL;
	size_t   outlen = 0;
	int      has_restricted;
	Ml_buf_t buf;

	buf = Ml_buf_create();

	for (i=0; i<num_strs; i++) {
		if (i != 0)
			Ml_buf_add_byte(buf, delim_char);

		has_restricted = (strs[i] == NULL)?0:Ml_str_implode_has_restricted_chars(strs[i], delim_char, enclose_char, escape_char);
		if (always_enclose || has_restricted) {
			Ml_buf_add_byte(buf, enclose_char);
			if (has_restricted) {
				Ml_str_implode_escape(buf, strs[i], enclose_char, escape_char);
			} else if (strs[i] != NULL) {
				Ml_buf_add_str(buf, strs[i]);
			}
			Ml_buf_add_byte(buf, enclose_char);
		} else if (strs[i] != NULL) {
			Ml_buf_add_str(buf, strs[i]);
		}
	}

	Ml_buf_finish(buf, &ret, &outlen);
	return ret;
}
